/*
 * convert.c - SATURN assembler, loader
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* gotta check input parameters in callers code
 * for possible use of ssprinf on big endian systems !! */

void    set_8( unsigned char *destp, unsigned char src )
{
	*destp = src;
}

void    set_16( unsigned char *destp, unsigned int src )
{
	unsigned char *d = destp + 2, *s = (unsigned char *)( &src );

	*(--d) = *s++;
	*(--d) = *s;

/* Original code is here:
 * register unsigned int s = *src;
 * 
 * *(--d) = s;
 * s >>= 8;
 *      *(--d) = s;
 *
 * this original works ONLY on little endian systems !!
 */
}

void    set_32( unsigned char *destp, unsigned long src )
{
	unsigned char *d = destp + 4, *s = (unsigned char *)( &src );

	*(--d) = *s++;
	*(--d) = *s++;
	*(--d) = *s++;
	*(--d) = *s;

/* Original code is here:
 * unsigned long s = *src;
 * 
 * *(--d) = s;
 * s >>= 8;
 * *(--d) = s;
 * s >>= 8;
 * *(--d) = s;
 * s >>= 8;
 *      *(--d) = s;
 *
 * this original works ONLY on little endian systems !!
 */
}

int     get_16( unsigned char *src )
{
	unsigned char *s = src;
	register int v;

	v = (int)( *s++ );
	v <<= 8;
	v |= (int)( *s );
	return v;
}

long    get_32( unsigned char *src )
{
	unsigned char *s = src;
	long v;

	v = (long)( *s++ );
	v <<= 8;
	v |= (long)( *s++ );
	v <<= 8;
	v |= (long)( *s++ );
	v <<= 8;
	v |= (long)( *s );
	return v;
}

unsigned int    get_u16( unsigned char *src )
{
	unsigned char *s = src;
	register unsigned int v;

	v = (unsigned int)( *s++ );
	v <<= 8;
	v |= (unsigned int)( *s );
	return v;
}

unsigned long   get_u32( unsigned char *src )
{
	unsigned char *s = src;
	unsigned long v;

	v = (unsigned long)( *s++ );
	v <<= 8;
	v |= (unsigned long)( *s++ );
	v <<= 8;
	v |= (unsigned long)( *s++ );
	v <<= 8;
	v |= (unsigned long)( *s );
	return v;
}
