/*
 * envfile.c - file open routine with enviroment search
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#ifndef	macintosh
# include <malloc.h>
#endif

#include "envfile.h"

char	*nameptr;

void	fix_dirc( char *pathname )
{
#if	defined(__MSDOS__)
	while( ( pathname = strchr( pathname, '/' ) ) != NULL )
		*pathname++ = '\\';
#else
	while( ( pathname = strchr( pathname, '\\' ) ) != NULL )
		*pathname++ = '/';
#endif
}

void	setname( char **nameaddr, char *name )
{
	if( *nameaddr != NULL ) {
		FREE( *nameaddr );
		DEBUG_PRN( "Address %ld: free'd memory\n", *nameaddr );
	}
	if( ( *nameaddr = calloc( 1, strlen( name ) + 1 ) ) == NULL )
#if	defined(__RPLCOMP__)
		error( FALSE, "error getting memory for %s", name );
#else
		error( "error getting memory for %s", name );
#endif
	DEBUG_PRN( "Address %ld: allocated %d bytes\n", *nameaddr, strlen( name ) + 1 );
	strcpy( *nameaddr, name );
	fix_dirc( *nameaddr );
}

FILE    *openfile( char *filename, char *envname,
	char *defaultpath, char *mode, char **actpath )
{
	FILE *fp;
	char *cp, *path, *base = NULL, *filebase = filename;
	char drive_letter = '\0', path_letter = '\0';
	int len, namelen;
	int dot_done = FALSE, done_default = FALSE, got_term = TRUE;

	namelen = strlen( filebase ) + 2;
	if( namelen >= 4 && filename[1] == ':' ) {
		drive_letter = *filename;
		filebase += 2;
	}
	if( ( cp = strchr( filebase, '\\' ) ) == NULL )
		if( ( cp = strchr( filebase, '/' ) ) == NULL )
			cp = filebase - 1;
	len = cp - filebase;
	if( len < 0 || ( len > 0 && strncmp( filebase, "..", len ) ) ) {
		if( *envname == '\0' || ( path = getenv( envname ) ) == NULL ) {
			path = defaultpath;
			got_term = done_default = TRUE;
		}
		len = 0;
		while( len >= 0 ) {
			path_letter = drive_letter;
			if( *path == '\0' )
				len = ( got_term && !dot_done ) ? 0 : -1;
			else
				if( *path != ';' ) {
					base = path;
					while( *path != '\0' && *path != ';' )
						path++;
					if( ( len = path - base ) >= 2 && base[1] == ':' ) {
						path_letter = *base;
						base += 2;
						len -= 2;
					}
					got_term = FALSE;
					if( *path == ';' ) {
						path++;
						got_term = TRUE;
					}
					if( len == 1 && *base == '.' )
						len = 0;
				} else {
					while( *path == ';' )
						path++;
					len = 0;
				}
			if( drive_letter != '\0' && drive_letter != path_letter ) {
				len = 0;
				continue;
			}
			if( len < 0 ) {
				if( *path == '\0' && !done_default ) {
					path = defaultpath;
					done_default = TRUE;
					got_term = FALSE;
					len = 0;
				}
				continue;
			}
			if( len == 0 && dot_done )
				continue;
			if( ( nameptr = calloc( 1, len + namelen + 2 ) ) == NULL )
#if	defined(__RPLCOMP__)
				error( FALSE, "error getting memory for %s", filename );
#else
				error( "error getting memory for %s", filename );
#endif
			DEBUG_PRN( "Address %ld: allocated %d bytes\n", filename, len + namelen );
			if( path_letter != '\0' ) {
				*nameptr = path_letter;
				nameptr[1] = ':';
			}
			if( len != 0 ) {
				strncat( nameptr, base, len );
				len = strlen( nameptr );
				if( nameptr[len - 1] != '\\' )
					nameptr[len] = '\\';
			} else
				dot_done = TRUE;
			strcat( nameptr, filebase );
			fix_dirc( nameptr );
			if( ( fp = fopen( nameptr, mode ) ) != NULL ) {
				if( actpath != NULL )
					*actpath = nameptr;
				return fp;
			}
			FREE( nameptr );
			DEBUG_PRN( "Address %ld: free'd memory (openfile rel)\n", nameptr );
			continue;
		}
	}
	if( ( nameptr = calloc( 1, namelen ) ) == NULL )
#if	defined(__RPLCOMP__)
		error( FALSE, "error getting memory for %s", filename );
#else
		error( "error getting memory for %s", filename );
#endif
	DEBUG_PRN( "Address %ld: allocated %d bytes\n", nameptr, namelen );
	strcpy( nameptr, filename );
	fix_dirc( nameptr );
	if( ( fp = fopen( nameptr, mode ) ) != NULL ) {
		if( actpath != NULL )
			*actpath = nameptr;
		return fp;
	}
	FREE( nameptr );
	DEBUG_PRN( "Address %ld: free'd memory (openfile abs)\n", nameptr );
	if( actpath != NULL )
		*actpath = NULL;
	return NULL;
}
