#ifndef _SASM_H_
#define _SASM_H_

struct imacro {
	int	im_type;
	long	im_line;				/* Line inside macro */
	char	im_buffer[257];	/* Macro arguments?? */
	long	im_addr;
	char	*im_filename;		/* Macro line definition */
	int	im_flagptr;
	int	im_listflags;
};

struct loop {
	char	*lp_filename;
	long	lp_fpos;
	char	lp_label[SYM_NAME_LEN + 1];
	long	lp_value;
	long	lp_end;
	long	lp_step;
};

struct lineref {
	unsigned long   l_line;
	struct lineref  *l_next;
};

struct fillref {
	long    f_address;
	long    f_bias;
	unsigned char   f_class;
	unsigned char   f_subclass;
	int     f_length;
	struct fillref  *f_next;
};

struct SATURN {
	char    id[6];
	char    length[2];
	char    nibbles[4];
	char    symbols[2];
	char    references[2];
	char    start[4];
	char    absolute[1];
	char    filler[1];
	char    date[26];
	char    title[40];
	char    softkeys[20];
	char    version[26];
	char    romid[4];
	char    reserved[118]; /* 256 - all above */
};

struct symbol {
	char    s_name[SYM_NAME_LEN + 1];
	long    s_value;
	char    s_red;
	unsigned char   s_flags;
	struct lineref  *s_lineref;
	unsigned int    s_fillcount;
	struct fillref  *s_fillref;
	struct symbol   *s_left;
	struct symbol   *s_right;
};

struct expression {
	int	e_extcount;
	int	e_relcount;
	char	e_name[SYM_NAME_LEN + 2]; /* ':' or '=' included */
	long	e_value;
};

struct OPCODE {
	char    value[MAXOPLINE + 1];
	unsigned long   length;
};

struct o_symbol {
	char	os_name[SYM_NAME_LEN];
	char	os_fillcount[2];
	char	os_value[4];
};

struct o_fillref {
	char	of_class[1];
	char	of_subclass[1];
	char	of_address[4];
	char	of_bias[4];
	char	of_length[2];
	char	of_reserved[6 + SYM_NAME_LEN - (1 + 1 + 4 + 4 + 2)];
};

union SFC {
	struct	o_symbol	s;
	struct	o_fillref	f;
	char	c[SYM_NAME_LEN + 6];
};

void    putlist(char *);

extern void sys_initialize(void), s_initialize(void), cleanmacros(void);
extern void readoptable(void);
extern unsigned long    opcodelen(char *);
extern struct OPCODE *getopcode(char *);

extern struct expression        *doexpr(void);

extern void setname(char **, char *);
extern FILE     *openfile(char *, char *, char *, char *, char **);

extern void error(char *, char *);

extern void DEBUG_PRN( char *, ... );

extern void
	allsymbols(void(*)(), struct symbol *),
	addlineref(struct symbol *),
	addlabel(char *, long, int);
extern struct symbol
	*sys_addsymbol(char *, long, int),
	*addsymbol(char *, long, int);

extern int getopt(int, char * const [], const char *);

extern void
	set_8(unsigned char *, unsigned char),
	set_16(unsigned char *, unsigned int),
	set_32(unsigned char *, unsigned long);
extern int      get_16(unsigned char *);
extern long     get_32(unsigned char *);
extern unsigned int     get_u16(unsigned char *);
extern unsigned long    get_u32(unsigned char *);

#endif
