/*
 * makerom.c - SATURN preprocessor utility for building libraries
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif

#include "makerom.h"

FILE	*infp, *outfp;
char
	*in_name, *out_name = NULL, *head_name,
	*hash_name, *end_name, *lib_name,
	*progname, *mesg_label, *conf_label, rom_id[4],
	hex[] = "0123456789ABCDEF0123456789abcdef";
int	name_len = 0;

int
	title_flag		=	FALSE,
	llist_flag		=	FALSE,
	outp_flag		=	FALSE,
	conf_flag		=	FALSE,
	mesg_flag		=	FALSE,
	name_flag		=	FALSE,
	head_flag		=	FALSE,
	hash_flag		=	FALSE,
	end_flag			=	FALSE,
	parse_end_flag	=	FALSE;

struct {
	char	*name;
	void	(*function)(char *line);
} parse_fncs[] = {
	{	"TITLE",			parse_title     },
	{	"CONFIGURE",	parse_configure },
	{	"ROMPHEAD",		parse_head      },
/*	{	"IF",				parse_if        }, dunno what to do w/ it */
	{	"LLIST",			parse_llist     },
	{	"MESSAGES",		parse_mesgs     },
	{	"NAME",			parse_name      },
	{	"TABLE",			parse_table     },
	{	"OUTPUT",		parse_output    },
	{	"SEARCH",		parse_search    },
	{	"RELOCATE",		parse_rel       },
	{	"FINISH",		parse_finish    },
	{	"END",			parse_end       },
	{	NULL,				null(void(*)()) }
};

struct exts_def {
	int	len, ofs, num, flag, tname;
	char	*name1, *name2;
	struct exts_def	*next, *sort;
}	*ext_first, *ext_sort;

int	hashes[16];

extern	int	optind;
extern	char	*optarg, *version;

void  usage( int flag )
{
/*	fprintf( stderr, "MAKEROM %s, Copyright (C) 1995 Mario Mikocevic Mozgy\n", version );
	fprintf( stderr, "MAKEROM comes with ABSOLUTELY NO WARRANTY\n" );
 */
	fprintf( stderr, "makerom %s\n", version );
	if( flag ) {
		fprintf( stderr, "Usage: %s [options] ... file.mn [file.m]\n", progname );
		fprintf( stderr, "options:\n" );
/*		fprintf( stderr, "-d debugfile\tdebug mode, specified output file\n" );
 */
		fprintf( stderr, "-v\t\tprint version\n" );
		exit( 1 );
	} else
		exit( 0 );
}

int	chtoi( char ch )
{
	if( isxdigit( ch ) )
		return ( ( ch > '9' ) ? ( toupper( ch ) - '7' ) : ( ch - '0' ) );
	else
		return -1;
}

int	myatoi( char **ch )
{
	register int i = 0;

	while( isxdigit( **ch ) ) {
		i *= 16;
		i += ( ( **ch > '9' ) ? ( toupper( *(*ch)++ ) - '7') : ( *(*ch)++ - '0' ) );
	}
	return i;
}

char	*mystrcpy( char *str )
{
	char	*tmp;

	tmp = (char *)calloc( strlen( str ) + 1, sizeof( char ) );
	CHECK( tmp );
	return ( strcpy( tmp, str ) );
}

int	mystrlen( char *tmp )
{
	char ch, *cp = tmp;
	int len = 0;

	while( ( ch = *cp++ ) != '\0' ) {
		if( ch == '\\' ) {
			if( strspn( cp, hex ) != 1 )
				++cp;
			++cp;
		}
		++len;
	}
	return len;
}

#ifdef	macintosh

int	strcasecmp( char *s1, char *s2 )
{
	register int	i;
	char	*t1, *t2, *tp;

	t1 = (char *)malloc( strlen( s1 ) + 1 );
	CHECK( t1 );
	t2 = (char *)malloc( strlen( s2 ) + 1 );
	CHECK( t2 );
	tp = t1;
	while( *s1 )
		*tp++ = toupper( *s1++ );
	tp = t2;
	while( *s2 )
		*tp++ = toupper( *s2++ );
	i = strcmp( t1, t2 );
	free( t1 );
	free( t2 );
	return i;
}

#endif

int	mystrcmp( char *str1, char *str2 )
{
/* stricmp == strcasecmp */
#if	defined(__MSDOS__)
	if( !stricmp( str1, str2 ) )
#else
	if( !strcasecmp( str1, str2 ) )
#endif
		return TRUE;
	if( toupper( *str1++ ) != toupper( *str2++ ) ||
		toupper( *str1++ ) != toupper( *str2++ ) )
		return FALSE;
	while( *str2 )
		if( toupper( *str1++ ) != toupper( *str2++ ) )
			return FALSE;
	return TRUE;
}

int	mystrcmplen( char *str1, char *str2 )
{
	register int i;

	i = strlen( str1 ) - strlen( str2 );
	if( i )
		return i;
	return ( strcmp( str1, str2 ) );
}

char	*conc_ext( char *filename, char *ext, int replace )
{
	char *buf, *ch;

	buf = (char *)calloc( strlen( filename ) + strlen( ext ) + 1, sizeof( char) );
	CHECK( buf );
	strcpy( buf, filename );
	ch = strrchr( buf, '.' );
	if( ch )
#if	defined(__MSDOS__)
		if( replace || !stricmp( ch + 1, ext ) )
#else
		if( replace || !strcmp( ch + 1, ext ) )
#endif
			*ch = '\0';
	strcat( buf, "." );
	strcat( buf, ext );
#if	defined(__MSDOS__)
	ch = buf;
	while( *ch )
		*ch++ = (char)toupper( *ch );
#endif
	return( buf );
}

char	*get_token( char **src, char *token_sep )
/*
 * Just a little more convenient than strtok()
 * This function returns a pointer to the first token
 * in src, and makes src point to the "new string".
 */
{
	char    *tok;

	tok = *src;
	if( !( src && *src && **src ) )
		return NULL;

	/* Make *src point after token */
	*src = strpbrk( *src, token_sep );
	if( *src ) {
		**src = '\0';
		++(*src);
		while( **src && strchr( token_sep, **src ) )
			++(*src);
	} else
		*src = "";
	return tok;
}

void	out_nibasc( FILE *fp, char *name )
{
	char	ch, *cp = name;
	int	value, done = 0;

	while( ( ch = *cp++ ) != '\0' ) {
		if( ch != '\\' ) {
			if( done++ == 0 )
				fprintf( fp, "\tNIBASC\t\\" );
			fprintf( fp, "%c", ch );
			done = done % 20;       /* max 20 characters in one line definition */
			if( done == 0 )
				fprintf( fp, "\\\n" );
		} else {
			if( done != 0 )
				fprintf( fp, "\\\n" );
			done = 0;
			if( strspn( cp, hex ) == 1 )
				value = ( strchr( hex, *cp++ ) - hex ) & 0x0F;
			else {
				value = 16 * ( ( strchr( hex, *cp++ ) - hex ) & 0x0F);
				value += ( strchr( hex, *cp++ ) - hex ) & 0x0F;
			}
			fprintf( fp, "\tCON(2)\t#%X\n", value );
		}
	}
	if( done != 0 )
		fprintf( fp, "\\\n" );
}

void	parse_title( char *line )
{
	char	*rest;

	if( title_flag ) {
		fprintf( outfp, "**ERROR! DUPLICATE TITLE: %s\n", line );
		return;
	}
	rest = strpbrk( line, " ,\t\n" );
	SKIPSPC( rest );
	KILLNEWLINE( rest )
	fprintf( outfp, "TITLE %s\n", rest );
	title_flag = TRUE;
}

void	parse_llist( char *line )
{
	char	*rest;

	if( llist_flag ) {
		fprintf( outfp, "**ERROR! DUPLICATE LLIST: %s\n", line );
		return;
	}
	rest = strpbrk( line, " ,\t\n" );
	SKIPSPC( rest );
	KILLNEWLINE( rest )
	fprintf( outfp, "LLIST %s\n", rest );
	llist_flag = TRUE;
}

void	parse_output( char *line )
{
	char	*rest;

	if( outp_flag ) {
		fprintf( outfp, "**ERROR! DUPLICATE OUTPUT: %s\n", line );
		return;
	}
	rest = strpbrk( line, " ,\t\n" );
	SKIPSPC( rest );
	KILLNEWLINE( rest )
	fprintf( outfp, "OUTPUT %s\n", rest );
	outp_flag = TRUE;
}

void	parse_search( char *line )
{
	char	*rest;

	rest = strpbrk( line, " ,\t\n" );
	SKIPSPC( rest );
	KILLNEWLINE( rest )
	fprintf( outfp, "SEARCH %s\n", rest );
}

void	parse_mesgs( char *line )
{
	char	*tmp;

	if( mesg_flag ) {
		fprintf( outfp, "** ERROR! DUPLICATE MESSAGE: %s\n", line );
		return;
	}
	get_token( &line, " ,\t\n" );
	tmp = get_token( &line, " ,\t\n" );
	mesg_label = mystrcpy( tmp );
	fprintf( outfp, "**MESSAGES %s\n", mesg_label );
	mesg_flag = TRUE;
}

void	parse_name( char *line )
{
	char	*tmp;

	if( name_flag ) {
		fprintf( outfp, "** ERROR! DUPLICATE NAME: %s\n", line );
		return;
	}
	get_token( &line, " ,\t\n" );
	tmp = get_token( &line, "\n" );
	name_len = mystrlen( tmp );
	lib_name = mystrcpy( tmp );
	fprintf( outfp, "**NAME %s\n", lib_name );
	name_flag = TRUE;
}

void	parse_configure( char *line )
{
	char	*tmp;

	if( conf_flag ) {
		fprintf( outfp, "** ERROR! DUPLICATE CONFIGURE: %s\n", line );
		return;
	}
	get_token( &line, " ,\t\n" );
	tmp = get_token( &line, " ,\t\n" );
	conf_label = mystrcpy( tmp );
	fprintf( outfp, "**CONFIGURE %s\n", conf_label );
	conf_flag = TRUE;
}

void	parse_rel( char *line )
{
	char *ext_name;

	get_token( &line, " ,\t\n" );
	ext_name = get_token( &line, " ,\t\n" );
	if( end_flag )
		fprintf( outfp, "** WARNING: CODE AFTER FINISH %s\n", ext_name );
	fprintf( outfp, "REL %s\n", ext_name );
	read_ext( ext_name );
}

void	parse_head( char *line )
{
	char	*tmp, *base;

	if( head_flag ) {
		fprintf( outfp, "** ERROR! DUPLICATE ROMPHEAD: %s\n", line );
		return;
	}
	get_token( &line, " ,\t\n" );
	tmp = get_token( &line, " ,\t\n" );
	head_name = mystrcpy( tmp );
	fprintf( outfp, "**ROMPHEAD %s\n", head_name );
	base = mystrcpy( head_name );
	tmp = conc_ext( base, "o", TRUE );
	fprintf( outfp, "REL %s\n", tmp );
	head_flag = TRUE;
}

void	parse_table( char *line )
{
	char	*tmp, *base;

	if( hash_flag ) {
		fprintf( outfp, "**ERROR! DUPLICATE TABLE: %s\n", line );
		return;
	}
	get_token( &line, " ,\t\n" );
	tmp = get_token( &line, " ,\t\n" );
	hash_name = mystrcpy( tmp );
	fprintf( outfp, "**TABLE %s\n", hash_name );
	base = mystrcpy( hash_name );
	tmp = conc_ext( base, "o", TRUE );
	fprintf( outfp, "REL %s\n", tmp );
	hash_flag = TRUE;
}

void	parse_finish( char *line )
{
	char	*tmp, *base;

	get_token( &line, " ,\t\n" );
	tmp = get_token( &line, " ,\t\n" );
	end_name = mystrcpy( tmp );
	fprintf( outfp, "**FINISH %s\n", end_name );
	base = mystrcpy( end_name );
	tmp = conc_ext( base, "o", TRUE );
	fprintf( outfp, "REL %s\n", tmp );
	end_flag = TRUE;
}


void	parse_end( char *line )
{
	fprintf( outfp, "END\n" );
	parse_end_flag = TRUE;
}

void	parse_line( char *line )
{
	char *command;
	register int i, flag = FALSE;

	if( *line == '*' )
		return;
	SKIPSPC( line );
	command = mystrcpy( line );
	if( ( command = strtok( command, " ,\t\n" ) ) == NULL )
		return;
	for( i = 0; parse_fncs[i].name; i++ )
		if( mystrcmp( parse_fncs[i].name, command ) ) {
			parse_fncs[i].function( line );
			flag = TRUE;
			return;
		}
	if( !flag )
		printf( "Unknown command: %s\n", command );
}

void	parse_file( void )
{
	char	buf[MAXBUFSIZE], *line;

	while( ( ( line = fgets( buf, MAXBUFSIZE, infp ) ) != NULL ) &&
			 !parse_end_flag )
		parse_line( line );
}

void    make_head(void)
{
	FILE    *fp;

	if((fp = fopen(head_name, "w")) == NULL) {
		printf("Could not open %s\n", head_name);
		exit(0);
	}
	fprintf(fp, "* ROMPART %s header\n\n", rom_id);
	fprintf(fp, "=ROMID%s\tEQU\t#%s\n", rom_id, rom_id);
	fprintf(fp, "\tCON(5)\t=DOLIB\n");
	fprintf(fp, "=LIB%s\n", rom_id);
	fprintf(fp, "\tREL(5)\t=SYSEND%s\n", rom_id);
	fprintf(fp, "\tCON(2)\t%d\n", name_len);
	if(name_len) {
		out_nibasc(fp, lib_name);
/*              fprintf(fp, "\tNIBASC\t\\%s\\\n", lib_name); */
		fprintf(fp, "\tCON(2)\t%d\n", name_len);
	}
	fprintf(fp, "=SYS%s\tCON(3)\t#%s\n", rom_id, rom_id);
	fprintf(fp, "\tREL(5)\t=HasH%s\n", rom_id);
	if(mesg_flag)
		fprintf(fp, "\tREL(5)\t=%s\n", mesg_label);
	else
		fprintf(fp, "\tCON(5)\t0\t\tNo Messages\n");
	fprintf(fp, "\tREL(5)\t=LinK%s\n", rom_id);
	fprintf(fp, "\tREL(5)\t=%s\n", conf_label);
	fprintf(fp, "\tEND\n");
	fclose(fp);
}

void    make_hash( void )
{
	struct exts_def *dummy;
	register int    i, ofs;
#if     0
	register char   *ch;
#endif
	FILE    *fp;

	if( ( fp = fopen( hash_name, "w" ) ) == NULL ) {
		printf( "Could not open %s\n", hash_name );
		exit( 0 );
	}
	fprintf( fp, "* ROMPART %s hashtable\n\n", rom_id );
	fprintf( fp, "=HasH%s\n", rom_id );
	fprintf( fp, "\tCON(5)\t=DOHSTR\n" );
	fprintf( fp, "\tREL(5)\tTaBlEnD\n" );
	for( i = 0; i < 16; i++ )
		if( hashes[i] )
			fprintf( fp, "\tREL(5)\toBleN%d\n", i + 1 );
		else
			fprintf( fp, "\tCON(5)\t0\n" );
	fprintf( fp, "\tREL(5)\tnaMEend\n" );
	dummy = ext_sort;
	for( i = 0; i < 16; i++ )
		if( hashes[i] ) {
			fprintf( fp, "oBleN%d\n", i + 1 );
			ofs = 0;
			while( dummy->len == ( i + 1 ) ) {
				fprintf( fp, "\tCON(2)\t#%X\n", i + 1 );
				out_nibasc( fp, dummy->name2 );

#if     0
/* Old code */
	/* We can add here as second parameter of strpbrk any
		suspicious character that NIBASC can't handle */
				if(!(strpbrk(dummy->name2, "\\\t\n")))
					fprintf(fp, "\tNIBASC\t\\%s\\\n", dummy->name2);
				else {
					ch = dummy->name2;
					while(*ch)
						fprintf(fp, "\tCON(2)\t#%X\n", *ch++);
				}
#endif

				if( dummy->tname )
					fprintf( fp, "\tCON(3)\t(~%s)/4096\n", dummy->name1 );
				else
					fprintf( fp, "\tCON(3)\t%d\n", dummy->num );
				dummy->ofs = ofs;
				ofs += ( 5 + 2 * ( i + 1 ) );
				if( ( dummy = dummy->sort ) == NULL )
					break;
			}
		}
	fprintf( fp, "naMEend\n" );
	dummy = ext_first;
	while( dummy != NULL ) {
		if( !dummy->tname && ( dummy->len != 0 ) ) {
			fprintf( fp, "=~x%s\tEQU\t#%s+4096*%d\n", dummy->name2, rom_id, dummy->num );
			fprintf( fp, "\tCON(5)\t(*)-(oBleN%d)-%d\n", dummy->len, dummy->ofs );
		}
		dummy = dummy->next;
	}
	dummy = ext_first;
	while( dummy != NULL ) {
		if( !dummy->tname && ( dummy->len == 0 ) )
			fprintf( fp, "=~%s\tEQU\t#%s+4096*%d\n", dummy->name2, rom_id, dummy->num );
		dummy = dummy->next;
	}
	fprintf( fp, "TaBlEnD\n" );
	fprintf( fp, "=LinK%s\n", rom_id );
	fprintf( fp, "\tCON(5)\t=DOHSTR\n" );
	fprintf( fp, "\tREL(5)\tenDLink\n" );
	dummy = ext_first;
	while( dummy != NULL ) {
		if( !dummy->tname && ( dummy->len != 0 ) )
			fprintf( fp, "\tREL(5)\t=x%s\n", dummy->name2 );
		dummy = dummy->next;
	}
	dummy = ext_first;
	while( dummy != NULL ) {
		if( !dummy->tname && ( dummy->len == 0 ) )
			fprintf( fp, "\tREL(5)\t=%s\n", dummy->name2 );
		dummy = dummy->next;
	}
	fprintf( fp, "enDLink\n\tEND\n" );
	fclose( fp );
}

void    make_end(void)
{
	FILE    *fp;

	if((fp = fopen(end_name, "w")) == NULL) {
		printf("Could not open %s\n", end_name);
		exit(0);
	}
	fprintf(fp, "* ROMPART %s end\n\n", rom_id);
	fprintf(fp, "\tCON(4)\t0\n");
	fprintf(fp, "=SYSEND%s\n\tEND\n", rom_id);
	fclose(fp);
}

void    parse_external(char *line)
{
	register int    i = 0, len, test = TRUE, tname = FALSE;
	char    *ch1, *ch2, *tmp;
	struct exts_def *dummy, *blah = NULL, *foo;

	tmp = mystrcpy(line);
	ch2 = strpbrk(tmp, " \n");
	if(*ch2++ == '\n') {
		free(tmp);
		return;
	}
	ch1 = strtok(tmp, " \t\n");
	KILLNEWLINE(ch2)

/* hm, in MAKEROM.DOC is stated that max is 11 chars in hash
 *      name, but hash table provides space even for 16 char hash names
 *      !and! original MAKEROM still chokes on 10 & 11 char names */

/* Maybe we should do changes in RPLCOMP as well to do it better */

	while(isxdigit(*ch1)) {
		i *= 16;
		i += chtoi(*ch1++);
		len = strlen(ch1);
		if(*ch1 == 'x') {
			len--;
			break;
		}
		if(i == len || i == 0)
			break;
	}

	if(i > 0 && i < 17)
		hashes[i - 1] = TRUE;
	else
		if(i == 0)
			len = i;
		else {
			KILLNEWLINE(line)
			printf("Invalid External entry: %s\n", line);
			return;
		}

	dummy = (struct exts_def *)calloc(1, sizeof(struct exts_def));
	CHECK(dummy);
	if( ext_first != NULL ) {
		blah = ext_first;
		while( blah->next ) {
			if( !strcmp( blah->name1, ch1 ) )
				tname = TRUE;
			blah = blah->next;
		}
		if( !strcmp( blah->name1, ch1 ) )
			tname = TRUE;
	}
	/* dummy->len = len; */
	dummy->tname = tname;
	dummy->len = i;
	dummy->ofs = 0;
	dummy->num = ((ext_first != NULL) ? blah->num + 1 : 0 );
	dummy->flag = ((*ch1 == 'x') ? TRUE : FALSE );
	dummy->name1 = mystrcpy(ch1);
	dummy->name2 = mystrcpy(ch2);
	dummy->next = NULL;
	if(ext_first != NULL)
		blah->next = dummy;
	else
		ext_first = dummy;
	free(tmp);
	if(dummy->len == 0)
		return;
	if(ext_sort != NULL) {
		blah = ext_sort;
		while(((test = (blah->len < dummy->len) ? -1 :
				(blah->len > dummy->len) ? 1 :
				mystrcmplen(blah->name2, dummy->name2)) < 0) &&
				blah->sort != NULL)
			blah = blah->sort;
		if(blah == ext_sort && test > 0) {
			ext_sort = dummy;
			ext_sort->sort = blah;
		} else {
			if(test > 0 || blah->sort) {
				foo = ext_sort;
				while(foo->sort != blah)
					foo = foo->sort;
				foo->sort = dummy;
				dummy->sort = blah;
			} else
				blah->sort = dummy;
		}
	}
	else
		ext_sort = dummy;
}

void    read_ext(char *file_name)
{
	char    *tmp_name, buf[MAXBUFSIZE];
	FILE    *tmpfp;

	tmp_name = conc_ext(file_name, "ext", TRUE );
	if((tmpfp = fopen(tmp_name, "r")) != NULL) {
		while((fgets(buf, MAXBUFSIZE, tmpfp)) != NULL)
			switch(buf[0]) {
				case 'Y':
				case 'y':
					strncpy(rom_id, buf + 1, 3);
					rom_id[3] = '\0';
					break;
				case 'X':
				case 'x': /* don't know what I am supposed to do */
					break;
				default :
					parse_external(buf);
					break;
			}
		fclose(tmpfp);
	/* Kludge!!! stupid fopen() on some stations garble input string
		fprintf(outfp, "** Parsing %s\n", tmp_name);            *blah*  */
		fprintf(outfp, "** Parsing %s\n", conc_ext(file_name, "ext", TRUE ) );
	} else {
		fprintf(outfp, "** Could not open %s\n", tmp_name);
	}
}

void    make_files(void)
{
	if(!head_flag) {
		puts("Missing ROMPHEAD command");
		exit(0);
	}
	make_head();

	if(!hash_flag) {
		puts("Missing TABLE command");
		exit(0);
	}
	make_hash();

	if(!end_flag) {
		puts("Missing FINISH command");
		exit(0);
	}
	make_end();
}

void    sort_xnames( void )
{
	struct exts_def *dummy;
	int     xnames = 0, xindex, nullindex;

	dummy = ext_first;
	while( dummy != NULL ) {
		if( !dummy->tname && ( dummy->len != 0 ) )
			xnames++;
		dummy = dummy->next;
	}
	/* Now keep a running index of what the next xNAME index should be
	 * and what the next nullname should be and fix the indexes.
	 */
	xindex = 0;
	nullindex = xnames;
	dummy = ext_first;
	while( dummy != NULL ) {
		if( !dummy->tname )
			if( dummy->len != 0 )
				dummy->num = xindex++;
			else
				dummy->num = nullindex++;
		dummy = dummy->next;
	}
}

void get_options( int argc, char **argv )
{
	char	i;

	while( ( i = getopt( argc, argv, "v" ) ) != (char)( -1 ) )
		switch( i ) {
/*			case 'd':
				debug = TRUE;
				dbgname = optarg;
				DEBUG1( "Debug file is '%s'", dbgname );
				break;
 */
			case 'v':
				usage( FALSE );
				break;
			default:
				usage( TRUE );
				break;
		}
}

void	main( int argc, char **argv )
{
	register int    i;
	char *std_sign = "-";

	progname = *argv;
	get_options( argc, argv );
	if( ( argc - optind ) > 3 )
		usage( TRUE );
	switch( argc - optind + 1 ) {
		case 3:
			if( !strcmp( argv[optind + 1], std_sign ) )
				out_name = std_sign;
			else
				out_name = conc_ext( argv[optind + 1], "m", FALSE );
		case 2:
			if( !strcmp( argv[optind], std_sign ) )
				in_name = std_sign;
			else
				in_name = conc_ext( argv[optind], "mn", FALSE );
			if( out_name == NULL)
				out_name = conc_ext( argv[optind], "m", TRUE );
			break;
/*		case 1:
			out_name = in_name = std_sign;
			break;
 */
		default:
			usage( TRUE );
			break;
	}

	if( strcmp( in_name, std_sign ) )
		if( ( infp = fopen( in_name, "r" ) ) == NULL ) {
			perror( in_name );
			exit( 1 );
		}
	if( strcmp( out_name, std_sign ) )
		if( ( outfp = fopen( out_name, "w" ) ) == NULL ) {
			perror( out_name );
			exit( 1 );
		}
	for( i = 0; i < 16; i++ )
		hashes[i] = FALSE;

/* here goes parsing of inname file */
	parse_file();
/* sorting xNAMEs */
	sort_xnames();
/* now we make tables and loader file */
	make_files();

	fclose( infp );
	fclose( outfp );
}
