/*
 * hash.c - hashing routines
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif

#include "hash.h"

void	hash_err( char *errstr )
{
	fprintf( stderr, "hash: %s", errstr );
	exit( 1 );
}

struct HASHROOT	*hashmake( int num )
{
	struct HASHROOT	*dummy;

	if( ( dummy = (struct HASHROOT *)calloc( 1, sizeof( struct HASHROOT ) ) ) == NULL )
		hash_err( "No memory for hashmake\n" );
	if( ( dummy->hashptr = (struct HASH **)calloc( num, sizeof( struct HASH ) ) ) == NULL )
		hash_err( "No memory for hashmake\n" );
	dummy->max = num;
	dummy->count = 0;
	return dummy;
}

int	hashchck( char *hash, struct HASHROOT *hash_root )
{
	register unsigned int	sum = 0;

	while( *hash != '\0' )
		sum += 2 * sum + *hash++;
	return sum % hash_root->max;
}

struct HASH	*hashadd( struct HASHROOT *hash_root, char *hash, int size )
{
	struct HASH	*hashptr, **dummy;

	if( ( hashptr = (struct HASH *)calloc( 1, sizeof( struct HASH ) ) ) == NULL )
		hash_err( "No memory for hashadd\n" );
	if( ( hashptr->define = (char *)calloc( 1, size ) ) == NULL )
		hash_err( "No memory for hashadd\n" );
	strncpy( hashptr->hash, hash, 12 );
	dummy = &hash_root->hashptr[hashchck( hash, hash_root )];
	hashptr->prev = *dummy;
	*dummy = hashptr;
	++hash_root->count;
	return hashptr;
}

struct HASH	*hashfind( struct HASHROOT *hash_root, char *hash )
{
	struct HASH	*hashptr;

	hashptr = hash_root->hashptr[hashchck( hash, hash_root )];
	while( hashptr != NULL && strncmp( hash, hashptr->hash, 12 ) )
		hashptr = hashptr->prev;
	return hashptr;
}
