/*
 * rplcomp.c - SATURN SysRPL compiler
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#ifndef	macintosh
# include <malloc.h>
#endif
#include <ctype.h>

#include "hash.h"
#include "rplcomp.h"

unsigned short	ctrl_stack[1000];
unsigned int	ctrl_lines[1000];
int	cur_token, err_count, ctrl_stack_cnt, label_cnt;
int
	debug	=	FALSE,
	force_comment	=	FALSE,
	commented	=	FALSE,
	in_next_file	=	FALSE,
	romid	=	FALSE;
FILE	*outfp = stdout, *extfp = NULL, *dbgfile = stdout;
char	*in_name, *out_name = NULL, *ext_name = NULL, *dbgname = NULL;
char	*progname;
char	*token_list[MAXTOKEN], token_buf[MAXBUFSIZE], line_buf[MAXBUFSIZE];
char	str_buf[1024], name_buf[50], *romid_num;
unsigned	char	romid_flag;
struct	HASHROOT	*hash_root;
struct	ARRY	*arryp;
struct	FILES	infiles[25], *infp;

extern	int	optind;
extern	char	*optarg, *version;

struct {
	char	names[6];
	char	values[6];
} chr_aliases = {	"ntr\"\\", "\n\t\r\"\\"	};

struct {
	char	hex[33];
	char	dec[11];
} str_table = {	"0123456789ABCDEF0123456789abcdef",	"0123456789"	};

struct {
	char	*token;
	int	flag;
	void	(*function)(char *);
} parse_ftab1[] = {
	{	";",	TRUE,	parse_semi	},
	{	"(;)",	TRUE,	parse_semi	},
	{	"[;]",	TRUE,	parse_semi	},
	{	"THEN",	TRUE,	parse_then	},
	{	"(THEN)",	TRUE,	parse_then	},
	{	"[THEN]",	TRUE,	parse_then	},
	{	"CODE",	TRUE,	parse_code	},
	{	"TITLE",	FALSE,	parse_title	},
	{	"xTITLE",	FALSE,	parse_title	},
	{	"STITLE",	FALSE,	parse_title	},
	{	"EJECT",	FALSE,	parse_title	},
	{	"ASSEMBLE",	FALSE,	parse_assemble	},
	{	"SETFLAG",  FALSE,	parse_title	},
	{	"CLRFLAG",  FALSE,	parse_title	},
/*	{	"#IF",	FALSE,	parse_prep	},
	{	"#IFNOT",	FALSE,	parse_prep	},
	{	"#ELSE",	FALSE,	parse_prep	},
	{	"#ENDIF",	FALSE,	parse_prep	},
 */
	{	NULL,	FALSE,	null(void(*)())	}
};

struct {
	char	*token;
	int	flag;
	void	(*function)(void);
} parse_ftab2[] = {
	{	"ENDCODE",	TRUE,	parse_endcode	},
	{	"TAG",	TRUE,	parse_tag	},
	{	"CHR",	TRUE,	parse_chr	},
	{	"%",	TRUE,	parse_real,	},
	{	"%%",	TRUE,	parse_e_real	},
	{	"C%",	TRUE,	parse_complex	},
	{	"C%%",	TRUE,	parse_e_complex	},
	{	"RPL",	FALSE,	parse_rpl	},
	{	"(",	FALSE,	parse_lparen	},
	{	"EXTERNAL",	FALSE,	parse_external	},
	{	"LOCAL",	FALSE,	parse_local	},
	{	"DEFINE",	FALSE,	parse_define	},
	{	"INCLUDE",	FALSE,	parse_include	},
	{	"{{",		TRUE,	parse_bind	},
	{	"}}",		TRUE,	parse_endbind	},
	{	NULL,	FALSE,	null(void(*)())	}
};

struct {
	char	*token;
	int	flag;
	void	(*function)(int);
	int	parameter;
} parse_ftab3[] = {
	{	"ROMID",	FALSE,	parse_romid,	0	},
	{	"xROMID",	FALSE,	parse_romid,	1	},
	{	"LOCALNAME",	FALSE,	parse_label,	FALSE	},
	{	"LOCALLABEL",	FALSE,	parse_label,	FALSE	},
	{	"NAMELESS",	FALSE,	parse_label,	TRUE	},
	{	"LABEL",	FALSE,	parse_label,	TRUE	},
	{	"ARRY",	TRUE,	parse_arry,	0	},
	{	"LNKARRY",	TRUE,	parse_arry,	1	},
	{	"LNKOTHER",	TRUE,	parse_arry,	2	},
	{	NULL,	FALSE,	null(void(*)()),	0	}
};

struct {
	char	*token;
	int	flag;
	void	(*function)(int, char *);
	int	par1;
	char	*par2;
} parse_ftab4[] = {
	{	"NAME",	TRUE,	parse_name,	1,	"NAME"	},
	{	"NULLNAME",	FALSE,	parse_name,	0,	"NULLNAME"	},
	{	"xNAME",	TRUE,	parse_name,	2,	"xNAME"	},
	{	"sNAME",	TRUE,	parse_name,	3,	"sNAME"	},
	{	"tNAME",	FALSE,	parse_name,	4,	"NULLNAME"	},
	{	"ALIAS",	FALSE,	parse_name,	4,	"ALIAS"	},
	{	NULL,	FALSE,	null(void(*)()),	0,	NULL	}
};

struct {
	char	*mnem;
} ctrl_table[14] = {
	{	"::"	},
	{	"BEGIN"	},
	{	"WHILE"	},
	{	"{"	},
	{	"DO"	},
	{	"IF"	},
	{	"ELSE"	},
	{	"TAG"	},
	{	"ARRY"	},
	{	"LNKARRY"	},
	{	"LNKOTHER"	},
	{	"SYMBOL"	},
	{	"UNIT"	},
	{	"FCN"	}
};

struct {
	char  *ntoken;
	char  *getname;
	char  *putname;
	char  *getaddr;
	char  *putaddr;
} lamtable[22] = {
	{ "ONE",        "1GETLAM",  "1PUTLAM", "#613B6", "#615E0" },
	{ "TWO",        "2GETLAM",  "2PUTLAM", "#613E7", "#615F0" },
	{ "THREE",      "3GETLAM",  "3PUTLAM", "#6140E", "#61600" },
	{ "FOUR",       "4GETLAM",  "4PUTLAM", "#61438", "#61615" },
	{ "FIVE",       "5GETLAM",  "5PUTLAM", "#6145C", "#61625" },
	{ "SIX",        "6GETLAM",  "6PUTLAM", "#6146C", "#61635" },
	{ "SEVEN",      "7GETLAM",  "7PUTLAM", "#6147C", "#61645" },
	{ "EIGHT",      "8GETLAM",  "8PUTLAM", "#6148C", "#61655" },
	{ "NINE",       "9GETLAM",  "9PUTLAM", "#6149C", "#61665" },
	{ "TEN",       "10GETLAM", "10PUTLAM", "#614AC", "#61675" },
	{ "ELEVEN",    "11GETLAM", "11PUTLAM", "#614BC", "#61685" },
	{ "TWELVE",    "12GETLAM", "12PUTLAM", "#614CC", "#61695" },
	{ "THIRTEEN",  "13GETLAM", "13PUTLAM", "#614DC", "#616A5" },
	{ "FOURTEEN",  "14GETLAM", "14PUTLAM", "#614EC", "#616B5" },
	{ "FIFTEEN",   "15GETLAM", "15PUTLAM", "#614FC", "#616C5" },
	{ "SIXTEEN",   "16GETLAM", "16PUTLAM", "#6150C", "#616D5" },
	{ "SEVENTEEN", "17GETLAM", "17PUTLAM", "#6151C", "#616E5" },
	{ "EIGHTEEN",  "18GETLAM", "18PUTLAM", "#6152C", "#616F5" },
	{ "NINETEEN",  "19GETLAM", "19PUTLAM", "#6153C", "#61705" },
	{ "TWENTY",    "20GETLAM", "20PUTLAM", "#6154C", "#61715" },
	{ "TWENTYONE", "21GETLAM", "21PUTLAM", "#6155C", "#61725" },
	{ "TWENTYTWO", "22GETLAM", "22PUTLAM", "#6156C", "#61735" }
};


struct {
	char	*mnem;
	int	prev;
	int	next;
	char	*table[5];
} tok_table[15] = {
	{	"::",	-1,	0,	{	"DOCOL",	NULL,	NULL,	NULL,	NULL	}	},
	{	"BEGIN",	-1,	1,	{	"BEGIN",	NULL,	NULL,	NULL,	NULL	}	},
	{	"WHILE",	1,	2,	{	"WHILE",	"DOCOL",	NULL,	NULL,	NULL	}	},
	{	"REPEAT",	2,	-1,	{	"SEMI",	"REPEAT",	NULL,	NULL,	NULL	}	},
	{	"UNTIL",	1,	-1,	{	"UNTIL",	NULL,	NULL,	NULL,	NULL	}	},
	{	"AGAIN",	1,	-1,	{	"AGAIN",	NULL,	NULL,	NULL,	NULL	}	},
	{	"{",	-1,	3,	{	"DOLIST",	NULL,	NULL,	NULL,	NULL	}	},
	{	"}",	3,	-1,	{	"SEMI",	NULL,	NULL,	NULL,	NULL	}	},
	{	"DO",	-1,	4,	{	"DO",	NULL,	NULL,	NULL,	NULL	}	},
	{	"LOOP",	4,	-1,	{	"LOOP",	NULL,	NULL,	NULL,	NULL	}	},
	{	"+LOOP",	4,	-1,	{	"+LOOP",	NULL,	NULL,	NULL,	NULL	}	},
	{	"UNIT",	-1,	12,	{	"DOEXT",	NULL,	NULL,	NULL,	NULL	}	},
	{	"SYMBOL",	-1,	11,	{	"DOSYMB",	NULL,	NULL,	NULL,	NULL	}	},
	{	"IF",	-1,	5,	{	"*** IF ***",	"'",	"DOCOL",	NULL,	NULL	}	},
	{	"ELSE",	5,	6,	{	"SEMI",	"*** ELSE ***",	"'",	"DOCOL",	NULL	}	}
};

void  usage( int flag )
{
	fprintf( stderr, "rplcomp %s\n", version );
	if( flag )
		return;
	fprintf( stderr, "Usage: %s [options] ... [sourcefile [outfile [extfile]]]\n", progname );
	fprintf( stderr, "options:\n" );
	fprintf( stderr, "-C\t\tforce commented mode\n" );
	fprintf( stderr, "-d debugfile\tdebug mode, specified output file\n" );
	fprintf( stderr, "-v\t\tprint version\n" );
}

char	*mystrcpy(char *str)
{
	char	*tmp;

	tmp = (char *)calloc(strlen(str) + 1, sizeof(char));
	CHECK(tmp);
	return (strcpy(tmp, str));
}

int	mystrcmp( char *str1, char *str2 )
{
/* stricmp == strcasecmp */
#if	defined(__MSDOS__)
	if( !stricmp( str1, str2 ) )
#else
	if( !strcasecmp( str1, str2 ) )
#endif
		return TRUE;
	if( toupper( *str1++ ) != toupper( *str2++ ) ||
		toupper( *str1++ ) != toupper( *str2++ ) )
		return FALSE;
	while( *str2 )
		if( toupper( *str1++ ) != toupper( *str2++ ) )
			return FALSE;
	return TRUE;
}

char	*conc_ext( char *filename, char *ext )
{
	char *buf, *ch;

	buf = (char *)calloc( strlen( filename ) + strlen( ext ) + 1, sizeof( char ) );
	CHECK( buf );
	strcpy( buf, filename );
	ch = strrchr( buf, '.' );
	if( ch )
#if	defined(__MSDOS__)
		if( !stricmp( ch + 1, ext ) )
#else
		if( !strcmp( ch + 1, ext ) )
#endif
			*ch = '\0';
	strcat( buf, "." );
	strcat( buf, ext );
#if	defined(__MSDOS__)
	ch = buf;
	while( *ch )
		*ch++ = (char)toupper( *ch );
#endif
	return( buf );
}

int	ctrl_stack_mm( int ofs )
{
	if( ctrl_stack_cnt == 0 )
		return FALSE;
	if( (int)ctrl_stack[ ctrl_stack_cnt - 1 ] != ofs )
		return FALSE;
	ctrl_stack_cnt--;
	return TRUE;
}

void	ctrl_stack_pp( int ofs )
{
	if( ctrl_stack_cnt == 1000 ) {
		fprintf( stderr, "Fatal error: control stack overflow\n" );
		exit( 1 );
	}
	ctrl_stack[ ctrl_stack_cnt ] = (unsigned short)ofs;
	ctrl_lines[ ctrl_stack_cnt++ ] = infp->line;
}

void	err_std( char *err )
{
	fprintf( stderr, err );
	exit( 1 );
}

struct LINE	*make_line(void)
{
	struct LINE	*dummy;

	if((dummy = (struct LINE *)calloc(1, sizeof(struct LINE))) == NULL)
		err_std("Cannot malloc LINE BLK\n");
	return dummy;
}

struct ARRY	*make_arry(void)
{
	struct ARRY	*dummy;

	if((dummy = (struct ARRY *)calloc(1, sizeof(struct ARRY))) == NULL)
		err_std("Cannot malloc ARRY BLK\n");
	return dummy;
}

void	make_item(struct ARRY *arryp)
{
	struct ITEM	*dummy;

	if((dummy = (struct ITEM *)calloc(1, sizeof(struct ITEM))) == NULL)
		err_std("Cannot malloc ITEM BLK\n");
	if(arryp->itm_last == NULL)
		arryp->items = dummy;
	else
		arryp->itm_last->next = dummy;
	arryp->itm_last = dummy;
}

void	output(char *str)
{
	struct LINE *dummy;

	if(arryp == NULL) {
		fputs(str, outfp);
		return;
	}
	dummy = make_line();
	dummy->line = mystrcpy(str);
	if(arryp->type == NULL) {
		if(arryp->dimens == NULL)
			arryp->dimens = dummy;
		else
			arryp->dim_last->next = dummy;
		arryp->dim_last = dummy;
	} else {
		if(arryp->itm_last->item == NULL)
			arryp->itm_last->item = dummy;
		else
			arryp->itm_last->last->next = dummy;
		arryp->itm_last->last = dummy;
	}
}

void	out_ch(char c)
{
	char ch[2];

	if(arryp != NULL) {
		ch[0] = c;
		ch[1] = '\0';
		output(ch);
		return;
	}
	putc(c, outfp);
}

void	error(int err, char *fmt_str, ...)
{
	char	buffer[MAXBUFSIZE];
	va_list	arg_ptr;
	struct	FILES	*fp;

	va_start(arg_ptr, fmt_str);
	if(err)
		err_count++;
	vsprintf(buffer, fmt_str, arg_ptr);
	va_end(arg_ptr);
	fp = infp;
	while(fp->fp == NULL)
		fp--;
	fprintf(outfp, "*| %s: %d: %s\n", fp->name, (infp->fp != NULL) ? fp->line : -(fp->line), buffer);
	fprintf(stderr, "%s: %d: %s\n", fp->name, (infp->fp != NULL) ? fp->line : -(fp->line), buffer);
}

void	stack_err(char *str)
{
	error(FALSE, "Unmatched control word: %s\n", str);
}

void	writeline(char *fmt_str, ...)
{
	char	buffer[MAXBUFSIZE];
	va_list	arg_ptr;

	va_start(arg_ptr, fmt_str);
	vsprintf(buffer, fmt_str, arg_ptr);
	va_end(arg_ptr);
	output(buffer);
}

void	end_file(void)
{
	if(infp == infiles) {
		while(ctrl_stack_cnt-- != 0)
			error(TRUE, "Unmatched control word in line #%d: %s", ctrl_lines[ctrl_stack_cnt], ctrl_table[ctrl_stack[ctrl_stack_cnt]].mnem);
		if(!romid && ext_name != NULL)
			remove(ext_name);
		if(err_count == 0)
			exit(0);
		if(err_count == 1)
			fputs("1 error in RPL source\n", stderr);
		else
			fprintf(stderr, "%d errors in RPL source\n", err_count);
		exit(1);
	}
	if(infp->fp != NULL) {
		fclose(infp->fp);
		infp->fp = NULL;
	}
	FREE(infp->name);
	infp--;
	in_next_file = TRUE;
	commented = force_comment;
	if(infp->fp != NULL)
		fprintf(outfp, "*|| Resuming file %s at line %d\n", infp->name, infp->line);
}

void	make_tokens(void)
{
	int	count = 0;
	char	*token, *cp;

	token = strcpy(token_buf, line_buf);
	if((cp = strtok(token, "\t\n\v\f\r ")) != NULL) {
		token_list[count++] = cp;
		while((cp = strtok(NULL, "\t\n\v\f\r ")) != NULL && count < MAXTOKEN - 1)
			token_list[count++] = cp;
	}
	token_list[count] = NULL;
	cur_token = 0;
}

void	get_line(void)
{
	char	*cp, ch;

	in_next_file = FALSE;
	*token_list = NULL;
	do {
		if(infp->fp != NULL) {
			if(fgets(line_buf, MAXBUFSIZE, infp->fp) == NULL)
				end_file();
			else {
				infp->line++;
				if((cp = strchr(line_buf, '\n')) != NULL)
					*cp = '\0';
				else
					do {
						ch = getc(infp->fp);
					} while(!feof(infp->fp) && !ferror(infp->fp) && ch != '\n');
				if(line_buf[0] == '*') { /* hmm, maybe it could be done better */
					if(commented)
						fputs("*- ", outfp);
					fputs(line_buf, outfp);
					putc('\n', outfp);
				} else {
					make_tokens();
					if(commented)
						fprintf(outfp, "*- %s\n", line_buf);
					else
						if(*token_list == NULL)
							putc('\n', outfp);
				}
			}
		} else {
			strcpy(line_buf, infp->name);
			if(line_buf[0] != '\0')
				FREE(infp->name);
			infp--;
			make_tokens();
		}
	} while(*token_list == NULL);
}

char	*just_get_token(void)
{
	while(token_list[cur_token] == NULL)
		get_line();
	return token_list[cur_token++];
}

char	*get_token(void)
{
	char	*cp;
	struct HASH	*hp;

	do {
		while(token_list[cur_token] == NULL)
			get_line();
		if((hp = hashfind(hash_root, token_list[cur_token])) != NULL &&
			 hp->flag == HASH_F_DEF) {
			if((++infp - infiles) >= 25 ) {
				error(TRUE, "Macro nesting level exceeded");
				infp--;
				hp = NULL;
				return (token_list[cur_token++]);
			}
			if(token_list[++cur_token] != NULL)
				cp = mystrcpy(token_list[cur_token] - token_buf + line_buf);
			else
				cp = "";
			infp->name = cp;
			infp->fp = NULL;
			infp->line = (infp - 1)->line;
			strcpy(line_buf, hp->define);
			make_tokens();
		}
	} while(hp != NULL && hp->flag == HASH_F_DEF);
	return token_list[cur_token++];
}

void	arry_el(char *token)
{
	unsigned short	value;

	value = (ctrl_stack_cnt > 0) ? ctrl_stack[ctrl_stack_cnt - 1] : -1;
	if(value == 8 || value == 9) {
		if(arryp->type != NULL) {
			if(strcmp(arryp->type, token))
				error(TRUE, "Mixed types in array: %s, %s", arryp->type, token);
		} else
			arryp->type = token;
	} else
		writeline("\tCON(5)\t=%s\n", token);
}

void	flush_item(int flag, struct LINE *line)
{
	struct LINE *next;

	next = line;
	while((line = next) != NULL) {
		next = line->next;
		if(flag)
			fputs(line->line, outfp);
		FREE(line->line);
		FREE(line);
	}
}

void	parse_arry(int thread)
{
	int	dims, items, i, linked, count;
	char	*token, *type;
	struct HASH	*hp;
	struct ITEM *item, *next;
	struct LINE *line;

	ctrl_stack_pp(thread + 8);
	arryp = make_arry();
	dims = 0;
	items = 1;
	do {
		token = get_token();
		if(strcmp("[", token)) {
			while((hp = hashfind(hash_root, token)) != NULL && hp->flag != HASH_F_DEF)
				token = hp->define;
			if(strspn(token, str_table.dec) != strlen(token)) {
				error(TRUE, "Illegal array dimension number:  %s", token);
				dims = -1;
				ctrl_stack_cnt--;
				flush_item(FALSE, arryp->dimens);
				return;
			}
		} else {
			if(dims != -1)
				break;
			ctrl_stack_cnt--;
			flush_item(FALSE, arryp->dimens);
			return;
		}
		dims++;
		i = atoi(token);
		items *= i;
		writeline("\tCON(5)\t%d\n", i);
	} while (1);
	if(thread == 2)
		arryp->type = "DOOTHER";
	linked = 0;
	count = 0;
	make_item(arryp);
	while(strcmp("]", (token = get_token()))) {
		if(strcmp("_", token)) {
			cur_token--;
			if(parse_lines(1)) {
				linked++;
				count++;
				make_item(arryp);
			}
		} else {
			make_item(arryp);
			count++;
		}
	}
	if( arryp->type == NULL || ( thread != 0 && linked == 0 ) ) {
		error( TRUE, "Array object type not defined" );
		dims = -1;
	}
	if(dims == 0)
		items = count;
	if(items != 0 && count > items) {
		error(TRUE, "Too many objects listed for array:  %d > %d", count, items);
		items = count;
		dims = -1;
	}
	if(thread == 0 && items != 0 && (linked != count || linked != items)) {
		error(TRUE, "ARRY has %d undefined objects.", items - linked);
		dims = -1;
	}
	if(!ctrl_stack_mm(thread + 8)) {
		error(TRUE, "Improper nesting in %s", ctrl_table[thread + 8].mnem);
		dims = -1;
	}
	if(dims == -1) {
		next = arryp->items;
		while((item = next) != NULL) {
			flush_item(FALSE, item->item);
			next = item->next;
			FREE(item);
		}
		flush_item(FALSE, arryp->dimens);
		FREE(arryp);
		arryp = NULL;
		return;
	}
	type = arryp->type;
	arryp->type = NULL;
	line = arryp->dimens;
	item = arryp->items;
	FREE(arryp);
	arryp = NULL;
	arry_el((thread != 0) ? "DOLNKARRAY" : "DOARRY");
	writeline("\tREL(5)\tLBL%03d\n", label_cnt);
	writeline("\tCON(5)\t=%s\n", type);
	if(dims != 0)
		writeline("\tCON(5)\t%d\n", dims);
	else
		writeline("\tCON(5)\t1\n\tCON(5)\t%d\n", items);
	flush_item(TRUE, line);
	next = item; /* temporary storage for later */
	if(thread != 0) {
		dims = 0;
		count = 0;
		while(item != NULL && dims < items) {
			if( ( item->item != NULL && count != 0 ) || count > 3 ) {
				writeline( "\tNIBHEX\t%0*d\n", count * 5, 0 );
				count = 0;
			}
			if(item->item != NULL)
				writeline("\tREL(5)\tLBL%03dA%03d\n", label_cnt, dims);
			else
				count++;
			dims++;
			item = item->next;
		}
		count += items - dims;
		while(count > 0) {
			if(count > 3)
				writeline("\tNIBHEX\t%020d\n", 0);
			else
				writeline("\tNIBHEX\t%0*d\n", count * 5, 0);
			count -= 4;
		}
		dims = 0;
		while((item = next) != 0 && dims < items) {
			if(thread != 0 && item->item != NULL)
				writeline("LBL%03dA%03d\n", label_cnt, dims);
			flush_item(TRUE, item->item);
			next = item->next;
			FREE(item);
			dims++;
		}
	} else
		while((item = next) != NULL) {
			flush_item(TRUE, item->item);
			next = item->next;
			FREE(item);
		}
	writeline("LBL%03d\n", label_cnt++);
}

void	out_e_real(char *str)
{
	int	i;

	output("\tNIBHEX\t");
	i = 20;
	while(i > 15)
		out_ch(str[i--]);
	output("\n\tNIBHEX\t");
	i = 15;
	while(i > 0)
		out_ch(str[i--]);
	output("\n\tNIBHEX\t");
	out_ch(*str);
	out_ch('\n');
}

void	out_real(char *str)
{
	int	i;

	output("\tNIBHEX\t");
	i = 15;
	while(i > 12)
		out_ch(str[i--]);
	output("\n\tNIBHEX\t");
	i = 12;
	while(i > 0)
		out_ch(str[i--]);
	output("\n\tNIBHEX\t");
	out_ch(*str);
	out_ch('\n');
}

/* Input:
 *	token	= pointer to token
 *	mant	= mantissa size in nibbles
 *	exp	= exponent size in nibbles
 *	max	= maximum positive exponent
 */
char	*real_exp(char *token, int mant, int exp, long max)
{
	char	*cp = token, num[80];
	int	i, d = 0, j;
	long	value = 0;

	if(strlen(token) >= 80)	/* Ignore too long reals */
		return NULL;

/* I really HATE bloody huge regexp routines so here
 * is my code for parsing real numbers as regular expression */
/*	re_comp("^[+-]?([0-9]*)?(\.([0-9]*))?([eE]([+-]?[0-9]+))?$"); */

	if(!*cp) return NULL;

	num[0] = (*cp != '-') ? '0' : '9';	/* Set sign nibble */
	if(*cp == '-' || *cp == '+')	/* Skip sign nibble */
		cp++;
	if(!*cp) return NULL;

	for(i = 1; i < 80; num[i++] = '\0');
	i = 0;

	if(*cp != '.') {
		while(*cp >= '0' && *cp <= '9')	/* count integer part */
			num[++i] = *cp++;
		if(!i) return NULL;
	}
	if(*cp == '.') {
		cp++;
		while(*cp >= '0' && *cp <= '9')	/* count decimal part */
			num[i + ++d] = *cp++;
	}

	while(num[1] == '0') {
		for(j = 1; num[j] != '\0'; j++)	/* normalize */
			num[j] = num[j + 1];
		i--;
	}

	if(num[1] == '\0') {
		for(i = 0; i < mant + exp + 1; i++)
			num[i] = '0';
		num[i] = '\0';
		return(mystrcpy(num));
	}

	j = d + i + 1;
	while(j <= mant)
		num[j++] = '0';
	if(*cp == 'e' || *cp == 'E') {
		cp++;
		j = FALSE;
		if(*cp == '-' || *cp == '+') {
			j = (*cp == '-') ? TRUE : FALSE;
			cp++;
		}
		if(*cp < '0' || *cp > '9') return NULL;
		while(*cp >= '0' && *cp <= '9') {
			value *= 10;
			value += *cp++ - '0';
		}
		if(j)
			value = -value;
	}
	value += (long)i - 1L;
	if(value < 0L && -value > max) return NULL;
	if(value > max) return NULL;
	if(value < 0L)
		value = 2 * (max + 1) + value;
	while(exp) {
		num[mant + exp--] = (value % 10) + '0';
		value /= 10;
	}
	if(*cp != '\0') return NULL;
	return mystrcpy(num);
}

void	parse_real(void)
{
	char	*real, *token;

	token = get_token();
	if((real = real_exp(token, 12, 3, 499L)) == NULL) {
		error(TRUE, "Invalid real number: %s", token);
		return;
	}
	arry_el("DOREAL");
	out_real(real);
	FREE(real);
}

void	parse_complex(void)
{
	char	*re_tok, *im_tok, *re, *im;

	re_tok = mystrcpy(get_token());
	im_tok = get_token();
	if((re = real_exp(re_tok, 12, 3, 499L)) == NULL ||
		(im = real_exp(im_tok, 12, 3, 499L)) == NULL) {
		error(TRUE, "Invalid complex number: %s %s", re_tok, im_tok);
		if(re != NULL)
			FREE(re);
		FREE(re_tok);
		return;
	}
	FREE(re_tok);
	arry_el("DOCMP");
	out_real(re);
	out_real(im);
	FREE(re);
	FREE(im);
}

void	parse_e_real(void)
{
	char	*real, *token;

	token = get_token();
	if((real = real_exp(token, 15, 5, 49999L)) == NULL) {
		error(TRUE, "Invalid extended real number: %s", token);
		return;
	}
	arry_el("DOEREL");
	out_e_real(real);
	FREE(real);
}

void	parse_e_complex(void)
{
	char	*re_tok, *im_tok, *re, *im;

	re_tok = mystrcpy(get_token());
	im_tok = get_token();
	if((re = real_exp(re_tok, 15, 5, 49999L)) == NULL ||
		(im = real_exp(im_tok, 15, 5, 49999L)) == NULL) {
		error(TRUE, "Invalid extended complex number: %s %s", re_tok, im_tok);
		FREE(re_tok);
		if(re != NULL)
			FREE(re);
		return;
	}
	FREE(re_tok);
	arry_el("DOECMP");
	out_e_real(re);
	out_e_real(im);
	FREE(re);
	FREE(im);
}

int	isrealnum(char *token)
{
	char	ch = *token;

	if(ch == '-' || ch == '+')
		token++;
	if(strlen(token) != 0 && strspn(token, str_table.dec) == strlen(token))
		return TRUE;
	return FALSE;
}

void	parse_bint2(char *token)
{
	char	*tok = token;
	int	neg;

	neg = (*tok == '-') ? TRUE : FALSE;
	if(neg || *tok == '+')
		tok++;
	while(*tok == '0')
		tok++;
	if(*tok == '\0')
		tok--;
	if(strlen(tok) > 7 ||
		(!neg && atol(tok) > 0xFFFFFL) ||
		(neg && atol(tok) > 0x80000L)) {
		error(TRUE, "Numeric overflow: %s", token);
		return;
	}
	arry_el("DOBINT");
	output("\tCON(5)\t");
	if(neg)
		output("(0-");
	output(tok);
	if(neg)
		out_ch(')');
	out_ch('\n');
}

void	parse_bint(char *token)
{
	if(strspn(token, str_table.hex) != strlen(token)) {
		error(TRUE, "Invalid hex number: %s", token);
		return;
	}
	while(*token == '0')
		token++;
	if(*token == '\0')
		token--;
	if(strlen(token) > 5) {
		error(TRUE, "Numeric overflow: %s", token);
		return;
	}
	arry_el("DOBINT");
	output("\tCON(5)\t#");
	while(*token != '\0')
		out_ch((strchr("abcdef", *token)) ? *token++ - 0x20 : *token++);
	out_ch('\n');
}

void	parse_romptr(char *tok)
{
	char	*token, ch, *r_a, *r_b, *cp;

	token = (tok == NULL) ? mystrcpy(get_token()) : tok;
	r_a = token;
	if(tok != NULL || !strchr("0123456789", *r_a) ||
		strspn(token, str_table.hex) != strlen(token)) {
		arry_el("DOROMP");
		writeline("\tCON(6)\t=~%s\n", token);
		if(tok == NULL)
			FREE(token);
		return;
	}
	r_b = cp = get_token();
	if(strspn(cp, str_table.hex) != strlen(r_b)) {
		error(TRUE, "Invalid hex number: ROMPTR %s %s", token, cp);
		FREE(token);
		return;
	}
	while(*r_a == '0')
		r_a++;
	if(*r_a == '\0')
		r_a--;
	while(*r_b == '0')
		r_b++;
	if(*r_b == '\0')
		r_b--;
	if(strlen(r_a) > 3 || strlen(r_b) > 3) {
		error(TRUE, "Numeric overflow: ROMPTR %s %s", token, cp);
		FREE(token);
		return;
	}
	arry_el("DOROMP");
	output("\tCON(3)\t#");
	while((ch = *r_a) != '\0')
		out_ch(strchr("abcdef", ch) ? *r_a++ - 0x20 : *r_a++);
	out_ch('\n');
	output("\tCON(3)\t#");
	while((ch = *r_b) != '\0')
		out_ch(strchr("abcdef", ch) ? *r_b++ - 0x20 : *r_b++);
	out_ch('\n');
	FREE(token);
}

void	parse_name(int tokflag, char *tok)
{
	char	*token, *cp, *namep, ch;
	unsigned char	hashflag;
	int	len;
	struct HASH	*hp;

	name_buf[0] = 'x';
	token = get_token();
	namep = strcpy(name_buf + 1, token);
	if(tokflag == 2)
		namep = name_buf;
	if(tokflag >= 3)
		token = get_token();
	if((hp = hashfind(hash_root, token)) == NULL && romid_flag == 1) {
		hp = hashadd(hash_root, namep, 1);
		hp->flag = romid_flag;
	} else
		if(hp != NULL) {
			hashflag = hp->flag;
			if(hashflag != HASH_F_0 && hashflag != HASH_F_EXT) {
				error(TRUE, "Illegal name re-use: %s %s", tok, namep);
				return;
			}
		}
	len = strlen(token);
	cp = token;
	while((ch = *cp) != '\0') {
		if(ch != '\\') {
			cp++;
			continue;
		}
		if(cp[1] == '\\') {
			len--;
			cp += 2;
			continue;
		}
		if(strspn(cp + 1, str_table.hex) < 2) {
			error(TRUE, "Bad hex character code:  %s", token);
			return;
		}
		len -= 2;
		cp++;
	}
	if(tokflag == 0)
		len = 0;
	if(!romid && extfp != NULL)
		fprintf(extfp, "%c%s\n", romid_flag + 'X', romid_num);
	romid = TRUE;
	if(extfp != NULL)
		fprintf(extfp, "%x%s %s\n", len, namep, token);
	if(tokflag != 4) {
		if(tokflag != 0)
			writeline("\tCON(6)\t=~%s\n", namep);
		writeline("=%s\n", namep);
	}
}

void	parse_label( int flag )
{
	char	*token;
	struct HASH	*hp;

/*
 *	if( romid_flag != 0 )
 *		error( TRUE, "Meaningless in xROM:  %s %s", token_list[cur_token - 1], token_list[cur_token] );
 *
 * OK, this is commented now 'til I decide what to do with it,
 * maybe command option !?!
 *
 */

	token = get_token();
	if( ( hp = hashfind( hash_root, token ) ) != NULL )
		if( ( flag && hp->flag != HASH_F_0 ) || ( !flag && hp->flag != HASH_F_LOC ) )
			error( FALSE, "Illegal name re-use:  %s", token );
	writeline( "%c%s\n", ( flag ) ? '=' : ':', token );
	if( !flag && hp == NULL ) {
		hp = hashadd( hash_root, token, 1 );
		hp->flag = HASH_F_LOC;
	}
}

void	parse_external(void)
{
	char	*token;
	register unsigned char	flag;
	struct HASH	*hp;

	token = get_token();
	if((hp = hashfind(hash_root, token)) == NULL) {
		hp = hashadd(hash_root, token, 1);
		hp->flag = HASH_F_EXT;
		return;
	}
	flag = hp->flag;
	if(flag == HASH_F_EXT || flag == 0) {
		hp->flag = HASH_F_EXT;
		return;
	}
	error(TRUE, "Illegal symbol re-use:  EXT %s", token);
}

void	parse_local(void)
{
	char	*token;
	struct HASH	*hp;

	token = get_token();
	if((hp = hashfind(hash_root, token)) == NULL) {
		hp = hashadd(hash_root, token, 1);
		hp->flag = HASH_F_LOC;
		return;
	}
	if(hp->flag != HASH_F_LOC)
		error(TRUE, "Cannot make local:  LOCAL %s", token);
}

void	parse_define( void )
{
	char	*token, *cp;
	struct HASH	*hp;

	token = just_get_token();
	cp = token_list[cur_token];
	if( cp == NULL )
		cp = "";
	else
		cp += line_buf - token_buf;
	if( ( hp = hashfind( hash_root, token ) ) != NULL ) {
		if( ( hp->flag != HASH_F_DEF ) || ( strcmp( hp->define, cp ) ) )
			error( TRUE, "Cannot DEFINE %s", token );
	} else {
		hp = hashadd( hash_root, token, strlen( cp ) + 2 );
		hp->flag = HASH_F_DEF;
		strcpy( hp->define, cp );
	}
	while( ( cp = token_list[cur_token] ) != NULL && strcmp( "(", cp ) )
		cur_token++; /* umh, ugly way of reading anything 'til end of line */
	if( cp != NULL )
		parse_lparen();
	else
		get_line();
}

void	parse_romid( int flag )
{
	char	*token;

	token = get_token();
	if( strspn( token, str_table.hex ) != strlen( token ) || strlen( token ) > 3)
		error( TRUE, "Invalid hex number: ROMID %s", token );
	else {
		while( *token == '0' )
			token++;
		if( *token == '\0' )
			token--;
		if( romid && ( ( romid_flag != (unsigned char)flag ) || strcmp( token, romid_num ) ) )
			error( TRUE, "Too many ROM IDs: ROMID %s", token );
		else {
			romid_num = mystrcpy( token );
			romid_flag = (unsigned char)flag;
			fprintf( outfp, "* ROMID set to %s\n", token );
		}
	}
}

void	parse_title( char *token )
{
	writeline( "\t%s", token );
	if( token_list[cur_token] != NULL )
		writeline( "\t%s", token_list[cur_token] - token_buf + line_buf );
	out_ch( '\n' );
	get_line();
}

void	parse_then(char *token)
{
	int	if_f = FALSE, else_f = FALSE;

	if(*token != '[' && (if_f = ctrl_stack_mm(5)) == 0 && (else_f = ctrl_stack_mm(6)) == 0)
		stack_err(token);
	if(*token != '(')
		if(*token != '[') {
			output("\tCON(5)\t=SEMI\n");
			output("*** THEN ***\n");
			if(if_f)
				output("\tCON(5)\t=RPIT\n");
			if(else_f)
				output("\tCON(5)\t=RPITE\n");
		} else
			output("\tCON(5)\t=THEN\n");
}

void	parse_lparen(void)
{
	char	*token, ch, *dummy;

	if(infp->fp == NULL) {
		get_line();
		return;
	}
	dummy = just_get_token();
	do {
		token = dummy - token_buf + line_buf;
		if(!commented)
			out_ch('*');
		if(!commented)
			out_ch(' ');
		while((ch = *token) != '\0' && ch != ')') {
			if(!commented)
				out_ch(ch);
			token++;
		}
		if(!commented)
			out_ch('\n');
		if(*token == ')') {
			while(token_list[cur_token] != NULL &&
				(token - line_buf) >= (token_list[cur_token] - token_buf)) {
					cur_token++;
			}
			break;
		}
		get_line();
		dummy = token_list[0];
	} while(1);
	if(*++token != '\0' && !strchr("\t\n\v\f\r ", *token)) {
		--cur_token;
	 	token_list[cur_token] = token - line_buf + token_buf;
	}
}

void	parse_semi(char *token)
{
	if(*token != '[')
		if(!ctrl_stack_mm(0) && !ctrl_stack_mm(12) && !ctrl_stack_mm(11))
			stack_err(token);
	if(*token != '(')
		output("\tCON(5)\t=SEMI\n");
	return;
}

int	istoken(char *token)
{
	if(!strcmp(token_list[0], token)) {
		if(force_comment)
			writeline("*- %s\n", line_buf);
		else
			writeline("* %s\n", token_list[0] - token_buf + line_buf);
		cur_token = 0;
		return TRUE;
	}
	if(token_list[1] == NULL)
		return FALSE;
	if(strcmp(token_list[1], token))
		return FALSE;
	if(token_list[0] != token_buf && token_list[0] != token_buf + 1)
		return FALSE;
	if(force_comment)
		writeline("*- %s\n%s\n", line_buf, token_buf);
	else
		writeline("* %s\n%s\n", token_list[1] - token_buf + line_buf, token_buf);
	cur_token = 1;
	return TRUE;
}

void	parse_code(char *token)
{
	if(!force_comment)
		writeline("* %s\n", token - token_buf + line_buf);
	in_next_file = FALSE;
	commented = FALSE;
	arry_el("DOCODE");
	writeline("\tREL(5)\tLBL%03d\n", label_cnt);
	do {
		get_line();
		if(in_next_file || istoken("RPL")) {
			error(TRUE, "Unmatched directive: CODE");
			label_cnt++;
			if(in_next_file)
				return;
			break;
		}
		if(istoken("ENDCODE")) {
			writeline("LBL%03d\n", label_cnt++);
			break;
		}
		if(istoken("ASSEMBLE"))
			continue;
		if(istoken("LOOP")) {
			output("\tA=DAT0\tA\n");
			output("\tD0=D0+\t5\n");
			output("\tPC=(A)\n");
			continue;
		}
		output(line_buf);
		out_ch('\n');
	} while (1);
	cur_token++;
	commented = force_comment;
}

void	parse_endcode(void)
{
	error(TRUE, "Unmatched directive: ENDCODE");
}

void	parse_rpl(void)
{
	if(!force_comment)
		fputs("* RPL\n", outfp);
	commented = force_comment;
}

void	parse_assemble(char *token)
{
	int	i, flag = FALSE;

	if(!force_comment)
		writeline("* %s\n", token - token_buf + line_buf);
	in_next_file = FALSE;
	commented = FALSE;
	do {
		get_line();
		if(istoken("RPL")) {
			just_get_token();
			commented = force_comment;
			return;
		}
		if(istoken("ENDCODE")) {
			error(TRUE, "Unmatched directive: ENDCODE");
			just_get_token();
			commented = force_comment;
			return;
		}
		if(istoken("ASSEMBLE"))
			continue;
		for(i=0; (parse_ftab4[i].token != NULL && !flag); i++)
			if(istoken(parse_ftab4[i].token)) {
				parse_ftab4[i].function(parse_ftab4[i].par1, parse_ftab4[i].par2);
				flag = TRUE;
			}
		if(flag)
			continue;
		if(istoken("LOOP")) {
			output("\tA=DAT0\tA\n");
			output("\tD0=D0+\t5\n");
			output("\tPC=(A)\n");
		} else {
			output(line_buf);
			out_ch('\n');
		}
	} while(!in_next_file);
	return;
}

int	get_str(char *token, char sep, int *flag, char **last)
{
	char	ch, *cp, *cptr;

	cptr = str_buf;
	*last = token + 1;
	if(sep != '\0' && *flag == FALSE)
		if(*token++ != '"')
			return -1;
	*flag = FALSE;
	while((*cptr++ = ch = *token++) != '\0') {
		*last = token;
		if(sep != '\0' && ch == '"') {
			if(*token == '"') {
				token++;
				continue;
			}
			*--cptr = '\0';
			if(*token != '\0' && !strchr("\t\n\v\f\r ", *token))
				return -1;
			return cptr - str_buf;
		} else {
			if(ch != '\\')
				continue;
			if(*token == '\0') {
				*flag = TRUE;
				*--cptr = '\0';
				return cptr - str_buf;
			}
			if((cp = strchr(chr_aliases.names, *token)) != NULL) {
				cptr[-1] = chr_aliases.values[cp - chr_aliases.names];
				token++;
				continue;
			}
			if(strspn(token, str_table.hex) < 2)
				return -1;
			cptr[-1] = 16 * ((strchr(str_table.hex, *token++) - str_table.hex) & 0x0F);
			cptr[-1] += (strchr(str_table.hex, *token++) - str_table.hex) & 0x0F;
		}
	}
	if(sep != '\0')
		return -1;
	return cptr - str_buf - 1;
}

void	parse_str(char *mnem, char *token, char sep)
{
	int	flag = FALSE, first_flag = FALSE, token_flag, len, done = 0;
	char	*cptr, *next;
	unsigned char	ch;

	token_flag = (token < token_buf || (token - token_buf) >= 1024) ? FALSE : TRUE;
	if(token_flag && sep != '\0')
		token += line_buf - token_buf;
	if((len = get_str(token, sep, &flag, &next)) != -1) {
		arry_el(mnem);
		first_flag = flag;
		if(first_flag)
			if(sep != '\0')
				writeline("\tREL(5)\tLBL%03d\n", label_cnt);
			else
				writeline("\tCON(2)\t((LBL%03d)-(*))/2-1\n", label_cnt);
		else
			if(sep != '\0')
				writeline("\tCON(5)\t%d\n", 2 * len + 5);
			else
				writeline("\tCON(2)\t%d\n", len);
	}
	do {
		cptr = str_buf;
		while(len-- > 0) {
			ch = *cptr++;
			if(ch != '\\' && ch >= ' ' && ch <= '~') {
				if(done++ == 0)
					output("\tNIBASC\t\\");
				out_ch(ch);
				done = done % 20; /* max 20 characters in one line definition */
				if(done == 0)
					output("\\\n");
			} else {
				if(done !=  0)
					output("\\\n");
				done = 0;
				writeline("\tCON(2)\t#%X\n", ch);
			}
		}
		if(flag && !token_flag)
			error(TRUE, "Continued %s inside DEFINITION: %s", mnem + 2, token);
		if(!flag || !token_flag) {
			if(done != 0)
				output("\\\n");
			if(first_flag)
				writeline(" LBL%03d\n", label_cnt++);
			if(len == -2)
				error(TRUE, "Invalid string:  %s", token);
			if(next < line_buf || (next - line_buf) >= 1024)
				return;
			next += token_buf - line_buf;
			while(token_list[cur_token] != NULL && next >= token_list[cur_token])
				cur_token++;
			return;
		}
		get_line();
		len = get_str(token = line_buf, sep, &flag, &next);
	} while (1);
}

void	parse_chr(void)
{
	char	*cp, *token, ch;

	token = get_token();
	switch(strlen(token)) {
		case 1:
			if(*token == '"') {
				token += line_buf - token_buf;
				if(token[2] == '"') {
					token++;
					get_token();
				}
			}
			break;
		case 2:
			if(*token == '^' && token[1] >= '?' && token[1] <= '_') {
				token++;
				*token ^= 0x40;
				break;
			}
			if(*token == '\\')
				if((cp = strchr(chr_aliases.names, token[1])) != NULL) {
					*++token = chr_aliases.values[cp - chr_aliases.names];
					break;
				}
		case 3:
			if(*token == '"' && token[2] == '"') {
				token++;
				break;
			}
			if(*token == '\\') {
				cp = token + 1;
				if(strspn(cp, str_table.hex) == 2) {
					*token = ((strchr(str_table.hex, *cp++) - str_table.hex) & 0x0F) << 4;
					*token += (strchr(str_table.hex, *cp) - str_table.hex) & 0x0F;
					break;
				}
			}
		case 4:
			if(*token == '"' && token[1] == '^' &&
				token[2] >= '?' && token[2] <= '_' && token[3] == '"') {
				token += 2;
				*token ^= 0x40;
				break;
			}
			if(*token == '"' && token[1] == '\\')
				if((cp = strchr(chr_aliases.names, token[2])) != NULL && token[3] == '"') {
					*++token = chr_aliases.values[cp - chr_aliases.names];
					break;
				}
		case 5:
			if(*token == '"' && token[1] == '\\') {
				cp = token + 2;
				if(strspn(cp, str_table.hex) == 2 && token[4] == '"') {
					*token = ((strchr(str_table.hex, *cp++) - str_table.hex) & 0x0F) << 4;
					*token += (strchr(str_table.hex, *cp) - str_table.hex) & 0x0F;
					break;
				}
			}
		default:
			error(TRUE, "Invalid char: %s", token);
			return;
	}
	arry_el("DOCHAR");
	ch = *token;
	if(ch >= ' ' && ch <= '~' && ch != '\\')
		writeline("\tNIBASC\t\\%c\\\n", ch);
	else
		writeline("\tCON(2)\t%d\n", *token);
}

void	parse_grob(char *def, char *tok)
{
	char	*cp1, *cp2, *cp3;
	int	len, done, todo;

	done = 16;
	todo = 0;
	cp2 = cp1 = mystrcpy(get_token());
	if(strspn(cp1, str_table.hex) != strlen(cp2)) {
		error(TRUE, "Invalid hex number: %s %s", tok, cp1);
		FREE(cp1);
		return;
	}
	while(*cp1 != '\0') {
		todo <<= 4;
		/* check is this '+=' or just a '=' ?? */
		todo += ((strchr(str_table.hex, *cp1++) - str_table.hex) & 0x0F);
	}
	if(todo == 0)
		return;
	cp3 = get_token();
	if(strspn(cp3, str_table.hex) != strlen(cp3)) {
		error(TRUE, "Invalid hex number: %s %s %s", tok, cp1, cp3);
		FREE(cp2);
		return;
	}
	len = strlen(cp3);
	if(len > todo) {
		error(TRUE, "Numeric overflow: %s %s %s", tok, cp2, cp3);
		FREE(cp2);
		return;
	}
	FREE(cp2);
	arry_el(def);
	writeline("\tCON(5)\t%d", todo + 5);
	if(todo == 0) {
		out_ch('\n');
		return;
	}
	while(todo-- > len) {
		if(done == 16) {
			output("\n\tNIBHEX\t");
			done = 0;
		}
		out_ch('0');
		done++;
	}
	while(*cp3 != '\0') {
		if(done == 16) {
			output("\n\tNIBHEX\t");
			done = 0;
		}
		out_ch((strchr("abcdef", *cp3)) ? *cp3++ - 0x20 : *cp3++);
		done++;
	}
	out_ch('\n');
}

void	parse_tag(void)
{
	parse_str("DOTAG", get_token(), '\0');
	ctrl_stack_pp(7);
	while(!parse_lines(1));
	if(!ctrl_stack_mm(7))
		error(TRUE, "Improper nesting in TAG");
}

void	parse_include(void)
{
	char	*token;

	token = get_token();
	if((infp - infiles) == 24) {
		error(TRUE, "INCLUDE nesting too deep:  %s", token);
		return;
	}
	infp++;
	infp->line = 0;
	infp->name = NULL;
	if((infp->fp = openfile(token, "RPL_INCLUDE", "", "rt", &infp->name)) == NULL) {
		infp--;
		error(TRUE, "Cannot open INCLUDE file: %s", token);
		return;
	}
	fprintf(outfp, "*|| Reading from %s\n", infp->name);
	get_line();
}

void	parse_endbind(void)
{
        error(TRUE, "Unmatched control word: }}");
}

void	parse_bind(void)
{
	int	i, j, binds, flg;
	char	*token;

	static char tmpname[23][15];

	binds = 0;
	flg = FALSE;
	do {
	    token = just_get_token();
	    if(strcmp("}}",token)) {
	        strncpy(tmpname[binds], token, 14);
		tmpname[binds][15] = '\0';
		binds++;
	    }
	    else
	      flg = TRUE;
	}
	while (!flg && (binds<23));

	if(binds>22) {
	      error(TRUE, "Too many lambda bindings");
	      binds = 22;
	}
	else if(binds==0) {
	      error(TRUE, "Missing lambda bindings");
	      binds = 1;
	}

	/* Output the binding sequence */

	if(binds==1)
	       writeline("\tCON(5)\t=1LAMBIND\n");
	else if(binds==2)
	  {
	       writeline("\tCON(5)\t='\n");
	       writeline("\tCON(5)\t=NULLLAM\n");
	       writeline("\tCON(5)\t=DUPTWO\n");
	       writeline("\tCON(5)\t=DOBIND\n");
	  }
	else
	  {
	       writeline("\tCON(5)\t='\n");
	       writeline("\tCON(5)\t=NULLLAM\n");
	       writeline("\tCON(5)\t=%s\n", lamtable[binds-1].ntoken);
	       writeline("\tCON(5)\t=NDUPN\n");
	       writeline("\tCON(5)\t=DOBIND\n");
	  }

	/* Output the lambda name equates */

	for(i=0; i<binds ; i++) {
	       j = binds-i-1;
	       writeline("=%s\t=\t%s\t* %s\n", tmpname[i], lamtable[j].getaddr, lamtable[j].getname);
	       writeline("=%s@\t=\t%s\t* %s\n", tmpname[i], lamtable[j].getaddr, lamtable[j].getname);
	       writeline("=%s!\t=\t%s\t* %s\n", tmpname[i], lamtable[j].putaddr, lamtable[j].putname);
	       writeline("=!%s\t=\t%s\t* %s\n", tmpname[i], lamtable[j].putaddr, lamtable[j].putname);
	}
}


int	parse_token(char *token)
{
	int	i = 0, flag = FALSE, thread;
	char	**cp;

	if( ( *token == '(' && token[strlen( token ) - 1] == ')' ) ||
		( *token == '[' && token[strlen( token ) - 1] == ']' ) ) {
		flag = TRUE;
		++token;
		token[strlen( token ) - 1] = '\0';
	}
	while(i < 15) {
		if(!strcmp(token, tok_table[i].mnem)) {
			if(flag) {
				token--;
				if(*token == '(')
					token[strlen(token)] = ')';
				else
					if(*token == '[')
						token[strlen(token)] = ']';
			}
			if(*token == '[') {
				writeline("\tCON(5)\t=%.*s\n", strlen(token) - 2, token + 1);
				return TRUE;
			}
			cp = tok_table[i].table;
			thread = tok_table[i].next;
			if(*token != '(') {
				if(thread == 3 || thread == 0 || thread == 12 || thread == 11)
					arry_el(*cp);
				else
					while(*cp != NULL) {
						writeline((**cp == '*') ? "%s\n" : "\tCON(5)\t=%s\n", *cp);
						cp++;
					}
			}
			if(tok_table[i].prev != -1)
				if(ctrl_stack_mm(tok_table[i].prev) == FALSE)
					stack_err(token);
			if(thread != -1)
				ctrl_stack_pp(thread);
			if(*token == '(')
				return TRUE;
			if(thread != 3 && thread != 0 && thread != 12 && thread != 11)
				return TRUE;
			thread = ctrl_stack_cnt;
			while(thread <= ctrl_stack_cnt)
				parse_lines(1);
			return TRUE;
		}
		i++;
	}
	if(flag) {
		token--;
		if(*token == '(')
			token[strlen(token)] = ')';
		else
			if(*token == '[')
				token[strlen(token)] = ']';
	}
	return FALSE;
}

int	parse_lines(int count)
/* count is number of tokens to parse */
{
	int	i, step, flag, ret_flag = TRUE;
	char	*token;
	struct HASH	*hp;

	step = (count == 0) ? 0 : 1;
	while((count -= step) >= 0) {
		token = get_token();
		flag = FALSE;
		if(!parse_token(token)) {
			for(i=0; (parse_ftab1[i].token != NULL && !flag); i++)
				if(!strcmp(parse_ftab1[i].token, token)) {
					parse_ftab1[i].function(token);
					ret_flag = parse_ftab1[i].flag;
					flag = TRUE;
				}
			if(flag)
				continue;
			for(i=0; (parse_ftab2[i].token != NULL && !flag); i++)
				if(!strcmp(parse_ftab2[i].token, token)) {
					parse_ftab2[i].function();
					ret_flag = parse_ftab2[i].flag;
					flag = TRUE;
				}
			if(flag)
				continue;
			for(i=0; (parse_ftab3[i].token != NULL && !flag); i++)
				if(!strcmp(parse_ftab3[i].token, token)) {
					parse_ftab3[i].function(parse_ftab3[i].parameter);
					ret_flag = parse_ftab3[i].flag;
					flag = TRUE;
				}
			if(flag)
				continue;
			for(i=0; (parse_ftab4[i].token != NULL && !flag); i++)
				if(!strcmp(parse_ftab4[i].token, token)) {
					parse_ftab4[i].function(parse_ftab4[i].par1, parse_ftab4[i].par2);
					ret_flag = parse_ftab4[i].flag;
					flag = TRUE;
				}
			if(flag)
				continue;
			if(*token == '"') {
				parse_str("DOCSTR", token, '"');
				continue;
			}
			if(!strcmp(token, "#")) {
				parse_bint(get_token());
				continue;
			}
			if(*token == '#')
				if(strspn(token + 1, str_table.hex) == (strlen(token) - 1)) {
					parse_bint(++token);
					continue;
				}
			if(!strcmp(token, "$")) {
				parse_str("DOCSTR", get_token(), '"');
				continue;
			}
			if(!strcmp(token, "LAM")) {
				parse_str("DOLAM", get_token(), '\0');
				continue;
			}
			if(!strcmp(token, "ID")) {
				parse_str("DOIDNT", get_token(), '\0');
				continue;
			}
			if(!strcmp(token, "ROMPTR")) {
				parse_romptr(NULL);
				continue;
			}
			if(!strcmp(token, "HXS")) {
				parse_grob("DOHSTR", "HXS");
				continue;
			}
			if(!strcmp(token, "GROB")) {
				parse_grob("DOGROB", "GROB");
				continue;
			}
			if(!strcmp(token, "PTR")) {
				writeline("\tCON(5)\t=%s\n", get_token());
				continue;
			}
			if(!strcmp(token, "[#]")) {
				arry_el("DOBINT");
				ret_flag = FALSE;
				continue;
			}
			if(isrealnum(token)) {
				parse_bint2(token);
				continue;
			}
			hp = hashfind(hash_root, token);
			if(hp != NULL && hp->flag == HASH_F_EXT)
				parse_romptr(token);
			else
				writeline("\tCON(5)\t%c%s\n", (hp != NULL && hp->flag == HASH_F_LOC) ? ':' : '=', token);
		}
	}
	return ret_flag;
}

void get_options( int argc, char **argv )
{
	char	i;

	while( ( i = getopt( argc, argv, "Cd:v" ) ) != (char)( -1 ) )
		switch( i ) {
			case 'C':
				DEBUG_PRN( "Forcing comments !\n" );
				force_comment = commented = TRUE;
				break;
			case 'd':
				debug = TRUE;
				dbgname = optarg;
				DEBUG_PRN( "Debug file is '%s'\n", dbgname );
				break;
			case 'v':
				usage( TRUE );
				exit( 0 );
				break;
			default:
				usage( FALSE );
				exit( -1 );
				break;
		}
}

void	main( int argc, char **argv )
{
	char *std_sign = "-";

	progname = *argv;
	get_options( argc, argv );
	if( ( argc - optind ) > 4 ) {
		usage( FALSE );
		exit( -2 );
	}
	switch( argc - optind + 1 ) {
		case 4:
			if( !strcmp( argv[optind + 2], std_sign ) )
				ext_name = std_sign;
			else
				ext_name = conc_ext( argv[optind + 2], "ext" );
		case 3:
			if( !strcmp( argv[optind + 1], std_sign ) )
				out_name = std_sign;
			else
				out_name = conc_ext( argv[optind + 1], "a" );
		case 2:
			if( !strcmp( argv[optind], std_sign ) )
				in_name = std_sign;
			else
				in_name = conc_ext( argv[optind], "s" );
			if( out_name == NULL)
				out_name = conc_ext( argv[optind], "a" );
			break;
		case 1:
			out_name = in_name = std_sign;
			break;
		default:
			usage( FALSE );
			exit( -3 );
			break;
	}

	infp = infiles;
	infp->fp = stdin;
	infp->name = mystrcpy( in_name );
	if( strcmp( in_name, std_sign ) )
		if( !( infp->fp = fopen( in_name, "r" ) ) ) {
			perror( in_name );
			exit( 1 );
		}
	if( strcmp( out_name, std_sign ) )
		if( !( outfp = fopen( out_name, "w" ) ) ) {
			perror( out_name );
			exit( 1 );
		}
	if( ext_name != NULL )
		if( strcmp( in_name, std_sign ) ) {
			if( !( extfp = fopen( ext_name, "w" ) ) ) {
				perror( ext_name );
				exit( 1 );
			}
		} else
			extfp = stdout;
	if( dbgname != NULL && strcmp( dbgname, std_sign ) )
		if( !( dbgfile = fopen( dbgname, "wt" ) ) ) {
			perror( dbgname );
			exit( 1 );
		}

	hash_root = hashmake( 1253 );
	parse_lines( 0 );
}
