/*
 * expr.c - SATURN assembler, expression parser
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_STRING_H */

#include "opc.h"
#include "expr.h"

int	o_ptr, e_ptr;
int	o_stack[OP_STACK_MAX];
struct	expression	*left, *right, e_stack[EXPR_STACK_MAX];

/*
 * Notes: Old higher order operators are evaluated away when a new one is met
 * Thus:  - \n must be of lowest precedance so that it is never evaluated away
 *        - "(" must be of lower precedance than ")" so that "(" existance can be checked
 *        - ")" must be of lower precedance than any arithmetic operators so than all
 *              operators up to "(" will be evaluated
 */
static	char	table[] = "\n()!&~+-*/%^";
static	int	precedence[] = { 0, 1, 2, 4, 4, 5, 6, 6, 7, 7, 8, 9 };
static	struct	expression	exp, temp;

extern	int	column, absolute, expr_resolved, pass2, quote_ok;
extern	int	debug, blanks_in_expr;
extern	int	label_index_m2, label_index_m1, label_index_p1, label_index_p2;
extern	long	pc;
extern	char	*ptr, *dbgname;
extern	FILE	*dbgfile;
extern	struct	optable	*entry;

int absexpr( struct expression *dummy )
{
	if( dummy->e_extcount == 0 && dummy->e_relcount == 0 )
		return TRUE;
	return FALSE;
}

char seechar( void )
{
	if( !strchr( " \t", *ptr ) )
		return *ptr;
	return '\0';
}

char nextchar( void )
{
	if( !strchr( " \t", *ptr ) ) {
		column++;
		return *ptr++;
	}
	return '\0';
}

int getop( char c )
{
	char	*p;

	if( ( p = strchr( table + 1, c ) ) == NULL )
		return 0;
	return p - table;
}

int pushop( char c )
{
	if( ++o_ptr >= OP_STACK_MAX ) {
		errmsg( "Operator stack overflow" );
		o_ptr--;
		return FALSE;
	}
	o_stack[o_ptr] = strchr( table, c ) - table;
	return TRUE;
}

int seeop( void )
{
	if( o_ptr < 0 ) {
		errmsg( "Operator stack underflow" );
		return 0;
	}
	return o_stack[o_ptr];
}

int popop( void )
{
	if( o_ptr < 0 ) {
		errmsg( "Operator stack underflow" );
		return 0;
	}
	return o_stack[o_ptr--];
}

int pushexpr( struct expression *e )
{
	if( ++e_ptr >= EXPR_STACK_MAX ) {
		errmsg( "Expression stack overflow" );
		e_ptr--;
		return FALSE;
	}
	e_stack[e_ptr] = *e;
	return TRUE;
}

struct expression *popexpr( void )
{
	if( e_ptr < 0 ) {
		errmsg( "Expression stack underflow" );
		return NULL;
	}
	return &e_stack[e_ptr--];
}

int getsymbol( void )
{
	char	c;
	int	i = 0, equals, local;
	struct symbol	*sym;

	exp.e_extcount = 0;
	exp.e_relcount = 0;
	exp.e_name[0] = '\0';
	exp.e_value = 0;
	equals = ( seechar() == '=' ) ? TRUE : FALSE;
	local = ( seechar() == ':' ) ? TRUE : FALSE;
	while( ( c = nextchar() ) != '\0' ) {
		if( c != ')' ) {
			if( ( ( equals | local ) + SYM_NAME_LEN ) > i )
				exp.e_name[i++] = c;
			continue;
		}
		column--;
		ptr--;
		break;
	}
	exp.e_name[i] = '\0';
	if( !strcmp( exp.e_name, "*" ) ) {
		exp.e_name[0] = '\0';
		exp.e_value = pc;
		exp.e_relcount = !absolute; /* cmp ax,1 ; sbb ax,ax ; neg ax */
		return pushexpr( &exp );
	}
	if( !strcmp( exp.e_name, "--" ) && label_index_m2 > 0 )
		sprintf( exp.e_name, "lBm2%X", label_index_m2 - 1 );
	else if( !strcmp( exp.e_name, "-" ) && label_index_m1 > 0 )
		sprintf( exp.e_name, "lBm1%X", label_index_m1 - 1 );
	else if ( !strcmp( exp.e_name, "+" ) )
		sprintf( exp.e_name, "lBp1%X", label_index_p1 );
	else if ( !strcmp( exp.e_name, "++" ) )
		sprintf( exp.e_name, "lBp2%X", label_index_p2 );
	if( ( sym = findsymbol( exp.e_name ) ) == NULL ) {
		if( !pass2 ) {
			exp.e_extcount = 1;
			errmsg( "Undefined symbol" );
			return pushexpr( &exp );
		}
		if( !equals )
			errmsg( "Undefined symbol" );
		sym = addsymbol( exp.e_name, 0L, SYM_NONE );
	} else {
		i = sym->s_flags & 0x0F;
		if( !equals && ( i == 1 || i == 0 ) )
			errmsg( "Undefined symbol" );
	}
	if( !( sym->s_flags & SYM_RESOLVED ) )
		expr_resolved = FALSE;
	if( pass2 ) {
		if( equals && !( sym->s_flags & SYM_RESOLVED ) )
			sym->s_flags |= SYM_EXTERNAL;
		addlineref( sym );
	}
	if( ( exp.e_extcount = equals ) == FALSE )
		exp.e_relcount = ( ( sym->s_flags & ( SYM_RELATIVE | SYM_RESOLVED ) ) == ( SYM_RELATIVE | SYM_RESOLVED ) ) ? TRUE : FALSE;
	exp.e_value = sym->s_value;
	if( exp.e_extcount == 0 )
		exp.e_name[0] = '\0';
	return pushexpr( &exp );
}

int nextexpr( char c )
{
	char	buffer[9];
	int	i, j;
	long	value = 0;

	temp.e_extcount = 0;
	temp.e_relcount = 0;
	temp.e_name[0] = '\0';
	switch( c ) {
		case '#':
			nextchar();
			j = 0;
			while( strchr( "0123456798ABCDEFabcdef", ptr[j] ) && ptr[j] != '\0' )
				j++;
			if( j == 0 || j > 8 ) {
				if( j > 8 )
					errmsg( "Invalid HEX constant (too many digits)" );
				else
					errmsg( "Invalid HEX constant (not HEX digit)" );
				return FALSE;
			}
			i = 0;
			while( i < j ) {
				value <<= 4;
				value += fromhex( ( strchr( "abcdef", ptr[i] ) ) ? ptr[i] - 0x20 : ptr[i] );
				i++;
			}
			column += j;
			ptr += j;
			temp.e_value = value;
			return pushexpr( &temp );
			/* break; */
		case '\'':
		case '\\':
		case '"':
			if( c == '\\' || quote_ok ) {
				if( !getasc( buffer, 8 ) ) {
					if( strlen( buffer ) != 8 )
						errmsg( "Invalid ASC constant (too large)" );
					else
						errmsg( "Invalid ASC constant" );
					return FALSE;
				}
				reverse( buffer );
				breverse( buffer );
				j = strlen( buffer );
				i = 0;
				while( i < j ) {
					value <<= 4;
					value += fromhex( buffer[i++] );
				}
				temp.e_value = value;
				return pushexpr( &temp );
			} else
				errmsg( "Disallowed ASC separator" );
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			while( ( c = seechar() ) != '\0' && ( strchr( "0123456789", c ) ) ) {
				value *= 10;
				value += ( nextchar() & 0x7F ) - '0';
			}
			temp.e_value = value;
			return pushexpr( &temp );
			/* break; */
		case '%':
			nextchar();
			j = 0;
			while( ( c = seechar() ) != '\0' && ( strchr( "01", c ) ) ) {
				value *= 2;
				value += ( nextchar() & 0x7F ) - '0';
				j++;
			}
			if( j == 0 ) {
				errmsg( "Invalid BIN constant (not BIN digit)" );
				return FALSE;
			}
			temp.e_value = value;
			return pushexpr( &temp );
			/* break; */
	}
	return getsymbol();
}

int doop( int op )
{
/*	char	c = *operator; using table[op] instead */
	int	all_ok = TRUE;
	long	count;

	if( table[op] == '\n' )
		return TRUE;
	if( table[op] == '(' ) {
		errmsg( "Unmatched '(' in expression" );
		return FALSE;
	}
	if( ( right = popexpr() ) == NULL || ( left = popexpr() ) == NULL )
		return FALSE;
	switch( table[op] ) {
		case '+':
			left->e_value += right->e_value;
			left->e_relcount += right->e_relcount;
			if( left->e_extcount == 0 )
				strcpy( left->e_name, right->e_name );
			else
				if( right->e_extcount != 0 ) {
					errmsg( "Can't add two external expressions" );
					all_ok = FALSE;
				}
			left->e_extcount += right->e_extcount;
			break;
		case '-':
			left->e_value -= right->e_value;
			left->e_relcount -= right->e_relcount;
			if( right->e_extcount != 0 ) {
				errmsg( "Can't have external reference on right side of -" );
				all_ok = FALSE;
			}
			break;
		case '!':
		case '^':
		case '%':
		case '&':
		case '*':
		case '/':
			if( left->e_extcount != 0 || left->e_relcount != 0 ||
				right->e_extcount != 0 || right->e_relcount != 0 ) {
				errmsg( "Can't use !,&,*,/,%, or ^ with external reference" );
				all_ok = FALSE;
			} else
				switch( table[op] ) {
					case '!':
						left->e_value |= right->e_value;
						break;
					case '&':
						left->e_value &= right->e_value;
						break;
					case '*':
						left->e_value *= right->e_value;
						break;
					case '/':
						if( right->e_value == 0 ) {
							errmsg( "Divide by zero (result set to 0)" );
							left->e_value = 0;
						} else
							left->e_value /= right->e_value;
						break;
					case '%':
						if( right->e_value == 0 ) {
							errmsg( "Modulo zero (result set to 0)" );
							left->e_value = 0;
						} else
							left->e_value %= right->e_value;
						break;
					case '^':
						count = right->e_value;
						right->e_value = left->e_value;
						if( count < 0 ) {
							errmsg( "Exponent less than zero (result set to 0)" );
							left->e_value = 0;
						} else {
							left->e_value = 1;
							while( count-- != 0 )
								left->e_value *= right->e_value;
						}
						break;
				}
			break;
		case '~':
			if( left->e_extcount != 0 || left->e_relcount != 0 ) {
				errmsg( "Can't have external reference on left side of ~" );
				all_ok = FALSE;
			} else {
				left->e_value <<= 8;
				left->e_value += right->e_value;
				left->e_extcount = right->e_extcount;
				left->e_relcount = right->e_relcount;
				strcpy( left->e_name, right->e_name );
			}
			break;
		default:
			errmsg( "Unrecognized operator" );
			all_ok = FALSE;
			break;
	}
	if( all_ok == FALSE )
		return FALSE;
	return pushexpr( left );
}

struct expression *doexpr( void )
{
	char	c;
	int	needexpr = TRUE, result = 0, op;

	expr_resolved = FALSE;
	if( !isfield() ) {
		errmsg( "Missing expression" );
		return NULL;
	}
	o_ptr = -1;
	e_ptr = -1;
	expr_resolved = TRUE;
	pushop( '\n' );
	while( ( c = seechar() ) != '\0' ) {
		if( needexpr ) {
			DEBUG_PRN( "doexpr: expecting expression, first character is '%c'\n", c );
			if( c == '(' ) {
				DEBUG_PRN( "doexpr: got a left parenthesis\n" );
				if( !pushop( nextchar() ) ) {
					expr_resolved = FALSE;
					return NULL;
				}
			} else {
				DEBUG_PRN( "doexpr: evaluating an expression with nextexpr()\n" );
				if( !nextexpr( c ) ) {
					expr_resolved = FALSE;
					return NULL;
				}
				needexpr = FALSE;
			}
		} else {
			DEBUG_PRN( "doexpr: expecting operator, first character is '%c'\n", c );
			if( ( op = getop( nextchar() ) ) == 0 ) {
				DEBUG_PRN( "doexpr: unrecognized operator '%c'\n", table[op] );
				errmsg( "Unrecognized operator" );
			} else {
				DEBUG_PRN( "doexpr: matched an operator; checking precedence rules\n" );
				while( precedence[seeop()] >= precedence[op] ) { /* hm!? */
					DEBUG_PRN( "doexpr: precedence causes evaluation of '%c'\n", table[seeop()] );
					if( !doop( popop() ) ) {
						expr_resolved = FALSE;
						return NULL;
					}
				}
				if( c == ')' ) {
					DEBUG_PRN( "doexpr: right parenthesis; evaluating operators\n" );
					do {
						c = table[seeop()];
					} while( c != '\n' && c != '(' && ( result = doop( popop() ) ) != FALSE );
					if( c == '\n' ) {
						errmsg( "Unmatched ')' in expression" );
						expr_resolved = FALSE;
						return NULL;
					}
					if( c == '(' ) {
						DEBUG_PRN( "doexpr: found matching left parenthesis\n" );
						popop();
					} else
						if( !result ) {
							expr_resolved = FALSE;
							return NULL;
						}
				} else {
					DEBUG_PRN( "doexpr: got an operator '%c'\n", c );
					if( !pushop( c ) ) {
						expr_resolved = FALSE;
						return NULL;
					}
					needexpr = TRUE;
				}
			}
		}
		if( !blanks_in_expr || ( *ptr != ' ' && *ptr != '\t' ) )
			continue;
		DEBUG_PRN( "doexpr: skipping white space in expression\n" );
		skipwhite();
		if( !isfield() )
			break;
	}
	op = getop( '\n' );
	while( seeop() != op ) {
		DEBUG_PRN( "doexpr: evaluating operator '%c'\n", table[seeop()] );
		if( !doop( popop() ) ) {
			expr_resolved = FALSE;
			return NULL;
		}
	}
	if( e_ptr == 0 && o_ptr == 0 &&
		( e_stack[0].e_extcount + e_stack[0].e_relcount ) <= 1 ) {
		if( !( entry->op_flags & OP_MACRO ) )
			e_stack[0].e_value += (long)entry->op_bias;
/* ! Beware ! stupid MeSsyDOS bcc has bugged fprintf() */
		DEBUG_PRN( "doexpr: return value #%08X = %%%ld\n", e_stack[0].e_value, e_stack[0].e_value );
		return &e_stack[0];
	}
	if( e_ptr != 0 )
		errmsg( "More data than operators in expression" );
	if( o_ptr != 0 )
		errmsg( "Operator left on stack" );
	if( e_ptr != 0 || o_ptr != 0 ) {
		expr_resolved = FALSE;
		return NULL;
	}
	errmsg( "Too many relocatable/external references" );
	expr_resolved = FALSE;
	return NULL;
}
