/*
 * opc.c - SATURN assembler, opcode (mnemonics) definitions
 * (c) 1994 Mario Mikocevic Mozgy (Mario.Mikocevic@etf.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif

#include "opc.h"

/* Local help defines for building the mnemonics */

#define	NRML	0
#define	PSEU	( OP_PSEUDO )
#define	TST0	( OP_TEST )
#define	TSTA	( OP_TEST | OP_SHORTA )
#define	TST2	( OP_TEST | CPU_SWITCH2 )
#define	SHAA	( OP_SHORTA )
#define	SHAB	( OP_SHORTA | OP_SHORTB )
#define	SHA2	( OP_SHORTA | CPU_SWITCH2 )
#define	BRCH	( OP_BRANCH )
#define	CPU1	( CPU_SWITCH1 )
#define	CPU2	( CPU_SWITCH2 )
#define	COMB	( OP_COMBOS )

#define	CRYNO	0x0
#define	CRY0	0x1
#define	CRY1	0x2
#define	CRYANY	0x3

#define	CR_XX	0x33	/* Carry may alter */
#define	CR_00	0x00	/* Carry restarts */
#define	CR_EQ	0x30	/* Carry stays the same */
#define	CR_CC	0x11	/* Carry is cleared */
#define	CR_NC	0x10	/* Carry will be clear after this */
#define	CR_SC	0x20	/* Carry will be set after this */

extern int code_level;

struct optable *optable_root;

/* Mnem	        Oper, Flags,	Len,	Opcodes,		fs,of,sz,bi,cla,sub,A,B,Carry */
struct optable mnemonics[] = {
 { "=",		ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_EQU, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL  },
 { "?A#0",	ARG_FS, TSTA,	3,	"90C", "8AC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A#B",	ARG_FS, TSTA,	3,	"904", "8A4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A#C",	ARG_FS, TSTA,	3,	"906", "8A6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A<=B",	ARG_FS, TSTA,	3,	"98C", "8BC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A<=C",	ARG_FS, TSTA,	3,	"98A", "8BA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A<B",	ARG_FS, TSTA,	3,	"984", "8B4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A<C",	ARG_FS, TSTA,	3,	"982", "8B2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A=0",	ARG_FS, TSTA,	3,	"908", "8A8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A=B",	ARG_FS, TSTA,	3,	"900", "8A0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A=C",	ARG_FS, TSTA,	3,	"902", "8A2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A>=B",	ARG_FS, TSTA,	3,	"988", "8B8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A>=C",	ARG_FS, TSTA,	3,	"98E", "8BE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A>B",	ARG_FS, TSTA,	3,	"980", "8B0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?A>C",	ARG_FS, TSTA,	3,	"986", "8B6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ABIT#0",	ARG_NIB, TST2,	5,	"80870", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ABIT#1",	ARG_NIB, TST2,	5,	"80860", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ABIT=0",	ARG_NIB, TST2,	5,	"80860", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ABIT=1",	ARG_NIB, TST2,	5,	"80870", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B#0",	ARG_FS, TSTA,	3,	"90D", "8AD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B#A",	ARG_FS, TSTA,	3,	"904", "8A4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B#C",	ARG_FS, TSTA,	3,	"905", "8A5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B<=A",	ARG_FS, TSTA,	3,	"988", "8B8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B<=C",	ARG_FS, TSTA,	3,	"98D", "8BD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B<A",	ARG_FS, TSTA,	3,	"980", "8B0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B<C",	ARG_FS, TSTA,	3,	"985", "8B5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B=0",	ARG_FS, TSTA,	3,	"909", "8A9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B=A",	ARG_FS, TSTA,	3,	"900", "8A0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B=C",	ARG_FS, TSTA,	3,	"901", "8A1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B>=A",	ARG_FS, TSTA,	3,	"98C", "8BC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B>=C",	ARG_FS, TSTA,	3,	"989", "8B9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B>A",	ARG_FS, TSTA,	3,	"984", "8B4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?B>C",	ARG_FS, TSTA,	3,	"981", "8B1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C#0",	ARG_FS, TSTA,	3,	"90E", "8AE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C#A",	ARG_FS, TSTA,	3,	"906", "8A6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C#B",	ARG_FS, TSTA,	3,	"905", "8A5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C#D",	ARG_FS, TSTA,	3,	"907", "8A7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C<=A",	ARG_FS, TSTA,	3,	"98E", "8BE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C<=B",	ARG_FS, TSTA,	3,	"989", "8B9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C<=D",	ARG_FS, TSTA,	3,	"98B", "8BB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C<A",	ARG_FS, TSTA,	3,	"986", "8B6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C<B",	ARG_FS, TSTA,	3,	"981", "8B1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C<D",	ARG_FS, TSTA,	3,	"983", "8B3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C=0",	ARG_FS, TSTA,	3,	"90A", "8AA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C=A",	ARG_FS, TSTA,	3,	"902", "8A2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C=B",	ARG_FS, TSTA,	3,	"901", "8A1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C=D",	ARG_FS, TSTA,	3,	"903", "8A3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C>=A",	ARG_FS, TSTA,	3,	"98A", "8BA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C>=B",	ARG_FS, TSTA,	3,	"98D", "8BD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C>=D",	ARG_FS, TSTA,	3,	"98F", "8BF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C>A",	ARG_FS, TSTA,	3,	"982", "8B2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C>B",	ARG_FS, TSTA,	3,	"985", "8B5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?C>D",	ARG_FS, TSTA,	3,	"987", "8B7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?CBIT#0",	ARG_NIB, TST2,	5,	"808B0", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?CBIT#1",	ARG_NIB, TST2,	5,	"808A0", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?CBIT=0",	ARG_NIB, TST2,	5,	"808A0", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?CBIT=1",	ARG_NIB, TST2,	5,	"808B0", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D#0",	ARG_FS, TSTA,	3,	"90F", "8AF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D#C",	ARG_FS, TSTA,	3,	"907", "8A7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D<=C",	ARG_FS, TSTA,	3,	"98F", "8BF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D<C",	ARG_FS, TSTA,	3,	"987", "8B7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D=0",	ARG_FS, TSTA,	3,	"90B", "8AB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D=C",	ARG_FS, TSTA,	3,	"903", "8A3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D>=C",	ARG_FS, TSTA,	3,	"98B", "8BB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?D>C",	ARG_FS, TSTA,	3,	"983", "8B3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?HS=0",	ARG_NIB, TST0,	3,	"830", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?MP=0",	ARG_NONE, TST0,	3,	"838", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?P#",	ARG_NIB, TST0,	3,	"880", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?P=",	ARG_NIB, TST0,	3,	"890", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?SB=0",	ARG_NONE, TST0,	3,	"832", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?SR=0",	ARG_NONE, TST0,	3,	"834", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ST#0",	ARG_NIB, TST0,	3,	"870", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ST#1",	ARG_NIB, TST0,	3,	"860", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ST=0",	ARG_NIB, TST0,	3,	"860", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?ST=1",	ARG_NIB, TST0,	3,	"870", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "?XM=0",	ARG_NONE, TST0,	3,	"831", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=-A",	ARG_FS, SHAA,	-1,	"B88", "F8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=-A-1",	ARG_FS, SHAA,	-1,	"B8C", "FC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_CC,	NULL,	NULL,	NULL },
 { "A=0",	ARG_FS, SHAA,	-1,	"A80", "D0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A!B",	ARG_FS, NRML,	4,	"0E08", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A!C",	ARG_FS, NRML,	4,	"0E0E", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A&B",	ARG_FS, NRML,	4,	"0E00", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A&C",	ARG_FS, NRML,	4,	"0E06", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A+1",	ARG_FS, SHAA,	-1,	"B04", "E4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A+A",	ARG_FS, SHAA,	-1,	"A04", "C4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A+B",	ARG_FS, SHAA,	-1,	"A00", "C0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A+C",	ARG_FS, SHAA,	-1,	"A0A", "CA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A+CON",	ARG_FSEXP1, CPU2,6,	"818000", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A+D0",	ARG_NONE, COMB,	8,	"136CA136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A+D1",	ARG_NONE, COMB,	8,	"137CA137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=A-1",	ARG_FS, SHAA,	-1,	"A0C", "CC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A-B",	ARG_FS, SHAA,	-1,	"B00", "E0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A-C",	ARG_FS, SHAA,	-1,	"B0A", "EA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=A-CON",	ARG_FSEXP1, CPU2,6,	"818080", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=B",	ARG_FS, SHAA,	-1,	"A84", "D4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=B!A",	ARG_FS, NRML,	4,	"0E08", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=B&A",	ARG_FS, NRML,	4,	"0E00", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=B+A",	ARG_FS, SHAA,	-1,	"A00", "C0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=B-A",	ARG_FS, SHAA,	-1,	"B0C", "EC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=C",	ARG_FS, SHAA,	-1,	"A8A", "DA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=C!A",	ARG_FS, NRML,	4,	"0E0E", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=C&A",	ARG_FS, NRML,	4,	"0E06", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=C+A",	ARG_FS, SHAA,	-1,	"A0A", "CA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "A=D0",	ARG_NONE, COMB,	6,	"132130", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=D1",	ARG_NONE, COMB,	6,	"133131", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=DAT0",	ARG_FSN, SHAB,	-1,	"1520", "142", "14A",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "A=DAT1",	ARG_FSN, SHAB,	-1,	"1530", "143", "14B",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "A=IN",	ARG_NONE, NRML,	3,	"802", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=PC",	ARG_NONE, CPU2,	4,	"81B4", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R0",	ARG_NONE, NRML,	3,	"110", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R0.F",	ARG_FS, SHA2,	-1,	"81A010", "110", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R1",	ARG_NONE, NRML,	3,	"111", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R1.F",	ARG_FS, SHA2,	-1,	"81A011", "111", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R2",	ARG_NONE, NRML,	3,	"112", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R2.F",	ARG_FS, SHA2,	-1,	"81A012", "112", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R3",	ARG_NONE, NRML,	3,	"113", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R3.F",	ARG_FS, SHA2,	-1,	"81A013", "113", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R4",	ARG_NONE, NRML,	3,	"114", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "A=R4.F",	ARG_FS, SHA2,	-1,	"81A014", "114", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ABASE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_ABASE, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL  },
 { "ABEX",	ARG_FS, SHAA,	-1,	"A8C", "DC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ABIT=0",	ARG_NIB, CPU2,	5,	"80840", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ABIT=1",	ARG_NIB, CPU2,	5,	"80850", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ABS",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ACEX",	ARG_FS, SHAA,	-1,	"A8E", "DE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AD0EX",	ARG_NONE, NRML,	3,	"132", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AD0XS",	ARG_NONE, NRML,	3,	"13A", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AD1EX",	ARG_NONE, NRML,	3,	"133", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AD1XS",	ARG_NONE, NRML,	3,	"13B", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ALLOC",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_ALLOC, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL  },
 { "APCEX",	ARG_NONE, CPU2,	4,	"81B6", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "AR0EX",	ARG_NONE, NRML,	3,	"120", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR0EX.F",	ARG_FS, SHA2,	-1,	"81A020", "120", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR1EX",	ARG_NONE, NRML,	3,	"121", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR1EX.F",	ARG_FS, SHA2,	-1,	"81A021", "121", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR2EX",	ARG_NONE, NRML,	3,	"122", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR2EX.F",	ARG_FS, SHA2,	-1,	"81A022", "122", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR3EX",	ARG_NONE, NRML,	3,	"123", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR3EX.F",	ARG_FS, SHA2,	-1,	"81A023", "123", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR4EX",	ARG_NONE, NRML,	3,	"124", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "AR4EX.F",	ARG_FS, SHA2,	-1,	"81A024", "124", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ASC(1)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 1, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASC(2)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 2, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASC(3)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 3, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASC(4)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 4, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASC(5)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 5, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASCM(1)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 1, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASCM(2)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 2, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASCM(3)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 3, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASCM(4)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 4, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASCM(5)",	ARG_LENASC, NRML,-1,	"", "", "",		0, 5, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "ASL",	ARG_FS, SHAA,	-1,	"B80", "F0", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ASLC",	ARG_NONE, NRML,	3,	"810", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ASR",	ARG_FS, SHAA,	-1,	"B84", "F4", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ASRB",	ARG_NONE, NRML,	3,	"81C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ASRB.F",	ARG_FS, SHA2,	-1,	"81900", "81C", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ASRC",	ARG_NONE, NRML,	3,	"814", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=-B",	ARG_FS, SHAA,	-1,	"B89", "F9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=-B-1",	ARG_FS, SHAA,	-1,	"B8D", "FD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_CC,	NULL,	NULL,	NULL },
 { "B=0",	ARG_FS, SHAA,	-1,	"A81", "D1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=A",	ARG_FS, SHAA,	-1,	"A88", "D8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=A!B",	ARG_FS, NRML,	4,	"0E0C", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=A&B",	ARG_FS, NRML,	4,	"0E04", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=A+B",	ARG_FS, SHAA,	-1,	"A08", "C8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B!A",	ARG_FS, NRML,	4,	"0E0C", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=B!C",	ARG_FS, NRML,	4,	"0E09", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=B&A",	ARG_FS, NRML,	4,	"0E04", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=B&C",	ARG_FS, NRML,	4,	"0E01", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=B+1",	ARG_FS, SHAA,	-1,	"B05", "E5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B+A",	ARG_FS, SHAA,	-1,	"A08", "C8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B+B",	ARG_FS, SHAA,	-1,	"A05", "C5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B+C",	ARG_FS, SHAA,	-1,	"A01", "C1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B+CON",	ARG_FSEXP1, CPU2, 6,	"818010", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B+D0",	ARG_NONE, COMB,	8,	"136C1136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=B+D1",	ARG_NONE, COMB,	8,	"137C1137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=B-1",	ARG_FS, SHAA,	-1,	"A0D", "CD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B-A",	ARG_FS, SHAA,	-1,	"B08", "E8", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B-C",	ARG_FS, SHAA,	-1,	"B01", "E1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=B-CON",	ARG_FSEXP1, CPU2, 6,	"818090", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=C",	ARG_FS, SHAA,	-1,	"A85", "D5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=C!B",	ARG_FS, NRML,	4,	"0E09", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=C&B",	ARG_FS, NRML,	4,	"0E01", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=C+B",	ARG_FS, SHAA,	-1,	"A01", "C1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=C-B",	ARG_FS, SHAA,	-1,	"B0D", "ED", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "B=D0",	ARG_NONE, COMB,	8,	"136D5136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "B=D1",	ARG_NONE, COMB,	8,	"137D5137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BAEX",	ARG_FS, SHAA,	-1,	"A8C", "DC", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BCEX",	ARG_FS, SHAA,	-1,	"A8D", "DD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BD0EX",	ARG_NONE, COMB,	7,	"DD136DD", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BD1EX",	ARG_NONE, COMB,	7,	"DD137DD", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSL",	ARG_FS, SHAA,	-1,	"B81", "F1", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSLC",	ARG_NONE, NRML,	3,	"811", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSR",	ARG_FS, SHAA,	-1,	"B85", "F5", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSRB",	ARG_NONE, NRML,	3,	"81D", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSRB.F",	ARG_FS, SHA2,	-1,	"81901", "81D", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSRC",	ARG_NONE, NRML,	3,	"815", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BSS",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_LINK, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "BUSCB",	ARG_NONE, CPU2,	4,	"8083", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BUSCC",	ARG_NONE, NRML,	3,	"80B", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "BUSCD",	ARG_NONE, CPU2,	4,	"808D", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C+P+1",	ARG_NONE, NRML,	3,	"809", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=-C",	ARG_FS, SHAA,	-1,	"B8A", "FA", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=-C-1",	ARG_FS, SHAA,	-1,	"B8E", "FE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_CC,	NULL,	NULL,	NULL },
 { "C=0",	ARG_FS, SHAA,	-1,	"A82", "D2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=A",	ARG_FS, SHAA,	-1,	"A86", "D6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=A!C",	ARG_FS, NRML,	4,	"0E0A", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=A&C",	ARG_FS, NRML,	4,	"0E02", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=A+C",	ARG_FS, SHAA,	-1,	"A02", "C2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=A-C",	ARG_FS, SHAA,	-1,	"B0E", "EE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=B",	ARG_FS, SHAA,	-1,	"A89", "D9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=B!C",	ARG_FS, NRML,	4,	"0E0D", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=B&C",	ARG_FS, NRML,	4,	"0E05", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=B+C",	ARG_FS, SHAA,	-1,	"A09", "C9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C!A",	ARG_FS, NRML,	4,	"0E0A", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C!B",	ARG_FS, NRML,	4,	"0E0D", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C!D",	ARG_FS, NRML,	4,	"0E0F", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C&A",	ARG_FS, NRML,	4,	"0E02", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C&B",	ARG_FS, NRML,	4,	"0E05", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C&D",	ARG_FS, NRML,	4,	"0E07", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C+1",	ARG_FS, SHAA,	-1,	"B06", "E6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C+A",	ARG_FS, SHAA,	-1,	"A02", "C2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C+B",	ARG_FS, SHAA,	-1,	"A09", "C9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C+C",	ARG_FS, SHAA,	-1,	"A06", "C6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C+CON",	ARG_FSEXP1, CPU2, 6,	"818020", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C+D",	ARG_FS, SHAA,	-1,	"A0B", "CB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C+D0",	ARG_NONE, COMB,	8,	"132C2132", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C+D1",	ARG_NONE, COMB,	8,	"133C2133", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=C+P+1",	ARG_NONE, NRML,	3,	"809", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C-1",	ARG_FS, SHAA,	-1,	"A0E", "CE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C-A",	ARG_FS, SHAA,	-1,	"B02", "E2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C-B",	ARG_FS, SHAA,	-1,	"B09", "E9", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C-CON",	ARG_FSEXP1, CPU2, 6,	"8180A0", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=C-D",	ARG_FS, SHAA,	-1,	"B0B", "EB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=D",	ARG_FS, SHAA,	-1,	"A8B", "DB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=D!C",	ARG_FS, NRML,	4,	"0E0F", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=D&C",	ARG_FS, NRML,	4,	"0E07", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=D+C",	ARG_FS, SHAA,	-1,	"A0B", "CB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "C=D0",	ARG_NONE, COMB,	6,	"136134", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=D1",	ARG_NONE, COMB,	6,	"137135", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=DAT0",	ARG_FSN, SHAB,	-1,	"1560", "146", "14E",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "C=DAT1",	ARG_FSN, SHAB,	-1,	"1570", "147", "14F",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "C=ID",	ARG_NONE, NRML,	3,	"806", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=IN",	ARG_NONE, NRML,	3,	"803", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=P",	ARG_NIB, NRML,	4,	"80C0", "", "",		0, 3, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=PC",	ARG_NONE, CPU2,	4,	"81B5", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R0",	ARG_NONE, NRML,	3,	"118", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R0.F",	ARG_FS, SHA2,	-1,	"81A018", "118", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R1",	ARG_NONE, NRML,	3,	"119", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R1.F",	ARG_FS, SHA2,	-1,	"81A019", "119", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R2",	ARG_NONE, NRML,	3,	"11A", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R2.F",	ARG_FS, SHA2,	-1,	"81A01A", "11A", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R3",	ARG_NONE, NRML,	3,	"11B", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R3.F",	ARG_FS, SHA2,	-1,	"81A01B", "11B", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R4",	ARG_NONE, NRML,	3,	"11C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=R4.F",	ARG_FS, SHA2,	-1,	"81A01C", "11C", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=RSTK",	ARG_NONE, NRML,	2,	"07", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "C=ST",	ARG_NONE, NRML,	2,	"09", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CAEX",	ARG_FS, SHAA,	-1,	"A8E", "DE", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CBEX",	ARG_FS, SHAA,	-1,	"A8D", "DD", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CBIT=0",	ARG_NIB, CPU2,	5,	"80880", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CBIT=1",	ARG_NIB, CPU2,	5,	"80890", "", "",	0, 4, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CD0EX",	ARG_NONE, NRML,	3,	"136", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CD0XS",	ARG_NONE, NRML,	3,	"13E", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CD1EX",	ARG_NONE, NRML,	3,	"137", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CD1XS",	ARG_NONE, NRML,	3,	"13F", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CDEX",	ARG_FS, SHAA,	-1,	"A8F", "DF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CHARMAP",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_CHARMAP, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CLRCARRY",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_CLRCARRY, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CLRFLAG",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_CLRFLAG, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CLRHST",	ARG_NONE, NRML,	3,	"82F", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CLRLIST",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_CLRLIST, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CLRST",	ARG_NONE, NRML,	2,	"08", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CON(1)",	ARG_EXPR, NRML,	1,	"", "", "",		0, 0, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(2)",	ARG_EXPR, NRML,	2,	"", "", "",		0, 0, 2, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(3)",	ARG_EXPR, NRML,	3,	"", "", "",		0, 0, 3, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(4)",	ARG_EXPR, NRML,	4,	"", "", "",		0, 0, 4, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(5)",	ARG_EXPR, NRML,	5,	"", "", "",		0, 0, 5, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(6)",	ARG_EXPR, NRML,	6,	"", "", "",		0, 0, 6, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(7)",	ARG_EXPR, NRML,	7,	"", "", "",		0, 0, 7, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CON(8)",	ARG_EXPR, NRML,	8,	"", "", "",		0, 0, 8, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "CONFIG",	ARG_NONE, NRML,	3,	"805", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CPCEX",	ARG_NONE, CPU2,	4,	"81B7", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "CPEX",	ARG_NIB, NRML,	4,	"80F0", "", "",		0, 3, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR0EX",	ARG_NONE, NRML,	3,	"128", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR0EX.F",	ARG_FS, SHA2,	-1,	"81A028", "128", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR1EX",	ARG_NONE, NRML,	3,	"129", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR1EX.F",	ARG_FS, SHA2,	-1,	"81A029", "129", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR2EX",	ARG_NONE, NRML,	3,	"12A", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR2EX.F",	ARG_FS, SHA2,	-1,	"81A02A", "12A", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR3EX",	ARG_NONE, NRML,	3,	"12B", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR3EX.F",	ARG_FS, SHA2,	-1,	"81A02B", "12B", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR4EX",	ARG_NONE, NRML,	3,	"12C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CR4EX.F",	ARG_FS, SHA2,	-1,	"81A02C", "12C", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSL",	ARG_FS, SHAA,	-1,	"B82", "F2", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSLC",	ARG_NONE, NRML,	3,	"812", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSR",	ARG_FS, SHAA,	-1,	"B86", "F6", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSRB",	ARG_NONE, NRML,	3,	"81E", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSRB.F",	ARG_FS, SHA2,	-1,	"81902", "81E", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSRC",	ARG_NONE, NRML,	3,	"816", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSTEX",	ARG_NONE, NRML,	2,	"0B", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "CSTRING",	ARG_ASC, NRML,	-1,	"", "", "",		0, 0, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D0=(2)",	ARG_EXPR, NRML,	4,	"19", "", "",		0, 2, 2, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=(4)",	ARG_EXPR, NRML,	6,	"1A", "", "",		0, 2, 4, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=(5)",	ARG_EXPR, NRML,	7,	"1B", "", "",		0, 2, 5, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=A",	ARG_NONE, NRML,	3,	"130", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=AS",	ARG_NONE, NRML,	3,	"138", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=B",	ARG_NONE, COMB,	7,	"DD134DD", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=C",	ARG_NONE, NRML,	3,	"134", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=CS",	ARG_NONE, NRML,	3,	"13C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=D",	ARG_NONE, COMB,	7,	"DF134DF", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=D0+",	ARG_NIB, NRML,	3,	"160", "", "",		0, 2, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D0=D0+A",	ARG_NONE, COMB,	8,	"136C2136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=D0+B",	ARG_NONE, COMB,	8,	"136C9136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=D0+C",	ARG_NONE, COMB,	8,	"132CA132", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=D0+D",	ARG_NONE, COMB,	8,	"136CB136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=D0-",	ARG_NIB, NRML,	3,	"180", "", "",		0, 2, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D0=D1",	ARG_NONE, COMB,	9,	"137134137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0=HEX",	ARG_DHEX, NRML,	-1,	"19", "", "",		0, 2, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D0D1EX",	ARG_NONE, COMB,	9,	"136137136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=(2)",	ARG_EXPR, NRML,	4,	"1D", "", "",		0, 2, 2, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=(4)",	ARG_EXPR, NRML,	6,	"1E", "", "",		0, 2, 4, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=(5)",	ARG_EXPR, NRML,	7,	"1F", "", "",		0, 2, 5, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=A",	ARG_NONE, NRML,	3,	"131", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=AS",	ARG_NONE, NRML,	3,	"139", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=B",	ARG_NONE, COMB,	7,	"DD135DD", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=C",	ARG_NONE, NRML,	3,	"135", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=CS",	ARG_NONE, NRML,	3,	"13D", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D",	ARG_NONE, COMB,	7,	"DF135DF", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D0",	ARG_NONE, COMB,	9,	"136135136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D1+",	ARG_NIB, NRML,	3,	"170", "", "",		0, 2, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D1=D1+A",	ARG_NONE, COMB,	8,	"137C2137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D1+B",	ARG_NONE, COMB,	8,	"137C9137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D1+C",	ARG_NONE, COMB,	8,	"133CA133", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D1+D",	ARG_NONE, COMB,	8,	"137CB137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1=D1-",	ARG_NIB, NRML,	3,	"1C0", "", "",		0, 2, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D1=HEX",	ARG_DHEX, NRML,	-1,	"1D", "", "",		0, 2, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D1D0EX",	ARG_NONE, COMB,	9,	"137136137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=-D",	ARG_FS, SHAA,	-1,	"B8B", "FB", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=-D-1",	ARG_FS, SHAA,	-1,	"B8F", "FF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_CC,	NULL,	NULL,	NULL },
 { "D=0",	ARG_FS, SHAA,	-1,	"A83", "D3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=C",	ARG_FS, SHAA,	-1,	"A87", "D7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=C!D",	ARG_FS, NRML,	4,	"0E0B", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=C&D",	ARG_FS, NRML,	4,	"0E03", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=C+D",	ARG_FS, SHAA,	-1,	"A03", "C3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=C-D",	ARG_FS, SHAA,	-1,	"B0F", "EF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D!C",	ARG_FS, NRML,	4,	"0E0B", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=D&C",	ARG_FS, NRML,	4,	"0E03", "", "",		2, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=D+1",	ARG_FS, SHAA,	-1,	"B07", "E7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D+C",	ARG_FS, SHAA,	-1,	"A03", "C3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D+CON",	ARG_FSEXP1, CPU2, 6,	"818030", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D+D",	ARG_FS, SHAA,	-1,	"A07", "C7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D+D0",	ARG_NONE, COMB,	8,	"136C3136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=D+D1",	ARG_NONE, COMB,	8,	"137C3137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=D-1",	ARG_FS, SHAA,	-1,	"A0F", "CF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D-C",	ARG_FS, SHAA,	-1,	"B03", "E3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D-CON",	ARG_FSEXP1, CPU2, 6,	"8180B0", "", "",	3, 5, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "D=D0",	ARG_NONE, COMB,	8,	"136D7136", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "D=D1",	ARG_NONE, COMB,	8,	"137D7137", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DAT0=A",	ARG_FSN, SHAB,	-1,	"1500", "140", "148",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "DAT0=C",	ARG_FSN, SHAB,	-1,	"1540", "144", "14C",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "DAT1=A",	ARG_FSN, SHAB,	-1,	"1510", "141", "149",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "DAT1=C",	ARG_FSN, SHAB,	-1,	"1550", "145", "14D",	3, 3, 1, -1, CLASS_NORMAL, 0, 15, 6, CR_EQ,	NULL,	NULL,	NULL },
 { "DCEX",	ARG_FS, SHAA,	-1,	"A8F", "DF", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DD0EX",	ARG_NONE, COMB,	7,	"DF136DF", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DD1EX",	ARG_NONE, COMB,	7,	"DF137DF", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DSL",	ARG_FS, SHAA,	-1,	"B83", "F3", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DSLC",	ARG_NONE, NRML,	3,	"813", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DSR",	ARG_FS, SHAA,	-1,	"B87", "F7", "",	1, 0, 0, 0, CLASS_NORMAL, 0, 15, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DSRB",	ARG_NONE, NRML,	3,	"81F", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DSRB.F",	ARG_FS, SHA2,	-1,	"81903", "81F", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "DSRC",	ARG_NONE, NRML,	3,	"817", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "EJECT",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_EJECT, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ELSE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_ELSE, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "END",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_END, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ENDIF",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_ENDIF, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ENDM",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_ENDM, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "EQU",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_INC, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "EXITF",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_EXITF, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "EXITM",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_EXITM, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "FOR",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_FOR, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "GOC",	ARG_LABEL, NRML, 3,	"4", "", "",		0, 1, 2, 0, CLASS_LABEL, 0, 0, 0, CR_NC,	NULL,	NULL,	NULL },
 { "GOLONG",	ARG_LABEL, NRML, 6,	"8C", "", "",		0, 2, 4, 0, CLASS_LABEL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "GONC",	ARG_LABEL, NRML, 3,	"5", "", "",		0, 1, 2, 0, CLASS_LABEL, 0, 0, 0, CR_SC,	NULL,	NULL,	NULL },
 { "GOSBVL",	ARG_EXPR2, NRML, 7,	"8F", "", "",		0, 2, 5, 0, CLASS_NORMAL, 1, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "GOSHORT",	ARG_LABEL, PSEU, 0,	"GOTO", "GONC", "GOC",	0, 0, 0, 0, CLASS_JUMP, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "GOSUB",	ARG_LABEL, NRML, 4,	"7", "", "",		1, 1, 3, 0, CLASS_LABEL, 1, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "GOSUBL",	ARG_LABEL, NRML, 6,	"8E", "", "",		1, 2, 4, 0, CLASS_LABEL, 1, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "GOTO",	ARG_LABEL, NRML, 4,	"6", "", "",		0, 1, 3, 0, CLASS_LABEL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "GOVLNG",	ARG_EXPR2, NRML, 7,	"8D", "", "",		0, 2, 5, 0, CLASS_NORMAL, 1, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "GOYES",	ARG_LABEL, BRCH, 2,	"", "", "",		0, 0, 2, 0, CLASS_LABEL, 0, 0, 0, CR_CC,	NULL,	NULL,	NULL },
 { "HEX(1)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 1, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEX(2)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 2, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEX(3)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 3, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEX(4)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 4, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEX(5)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 5, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEXM(1)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 1, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEXM(2)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 2, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEXM(3)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 3, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEXM(4)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 4, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HEXM(5)",	ARG_LENHEX, NRML,-1,	"", "", "",		0, 5, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX, NULL,   NULL,   NULL },
 { "HS=0",	ARG_NIB, NRML,	3,	"820", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IF",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IF, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFANYCARRY", ARG_NONE, PSEU, 0,	"", "", "",		0, 0, 0, 0, CLASS_IFCRYANY, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFCARRYCLR", ARG_NONE, PSEU, 0,	"", "", "",		0, 0, 0, 0, CLASS_IFCRY0, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFCARRYSET", ARG_NONE, PSEU, 0,	"", "", "",		0, 0, 0, 0, CLASS_IFCRY1, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFDEF",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFDEF, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFEQ",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 2, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFGE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 6, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFGT",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 4, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFLE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 3, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFLT",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 1, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFNDEF",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFNDEF, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFNE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 5, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFNEG",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 1, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFNOPC",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFNOPC, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFNZ",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 5, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFOPC",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFOPC, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFPASS1",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFPASS1, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFPASS2",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFPASS2, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFPOS",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 4, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFREACHED", ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFREACHED, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFSTREQ",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFSTR, 2, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFSTRGE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFSTR, 6, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFSTRGT",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFSTR, 4, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFSTRLE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFSTR, 3, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFSTRLT",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFSTR, 1, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFSTRNE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFSTR, 5, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "IFZER",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_IFEXPR, 2, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "INC(1)",	ARG_LINK, NRML,	1,	"", "", "",		0, 0, 1, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(2)",	ARG_LINK, NRML,	2,	"", "", "",		0, 0, 2, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(3)",	ARG_LINK, NRML,	3,	"", "", "",		0, 0, 3, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(4)",	ARG_LINK, NRML,	4,	"", "", "",		0, 0, 4, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(5)",	ARG_LINK, NRML,	5,	"", "", "",		0, 0, 5, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(6)",	ARG_LINK, NRML,	6,	"", "", "",		0, 0, 6, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(7)",	ARG_LINK, NRML,	7,	"", "", "",		0, 0, 7, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INC(8)",	ARG_LINK, NRML,	8,	"", "", "",		0, 0, 8, 0, CLASS_INC, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "INCLUDE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_INCLUDE, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "INTOFF",	ARG_NONE, NRML,	4,	"808F", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "INTON",	ARG_NONE, NRML,	4,	"8080", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "JUMP",	ARG_LABEL, PSEU, 0,	"GOTO", "GONC", "GOC",	0, 0, 0, 0, CLASS_JUMP, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "LA(1)",	ARG_EXPR, CPU2,	6,	"80820", "", "",	0, 5, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(2)",	ARG_EXPR, CPU2,	7,	"80821", "", "",	0, 5, 2, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(3)",	ARG_EXPR, CPU2,	8,	"80822", "", "",	0, 5, 3, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(4)",	ARG_EXPR, CPU2,	9,	"80823", "", "",	0, 5, 4, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(5)",	ARG_EXPR, CPU2,	10,	"80824", "", "",	0, 5, 5, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(6)",	ARG_EXPR, CPU2,	11,	"80825", "", "",	0, 5, 6, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(7)",	ARG_EXPR, CPU2,	12,	"80826", "", "",	0, 5, 7, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(8)",	ARG_EXPR, CPU2,	13,	"80827", "", "",	0, 5, 8, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LA(N)",	ARG_NIB, CPU2,	5,	"80820", "", "",	0, 4, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LAASC",	ARG_LASC, CPU2,	-1,	"80820", "", "",	0, 5, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LAHEX",	ARG_LHEX, CPU2,	-1,	"80820", "", "",	0, 5, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LASTR",	ARG_LSTR, CPU2,	-1,	"80820", "", "",	0, 5, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(1)",	ARG_EXPR, NRML,	3,	"30", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(2)",	ARG_EXPR, NRML,	4,	"31", "", "",		0, 2, 2, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(3)",	ARG_EXPR, NRML,	5,	"32", "", "",		0, 2, 3, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(4)",	ARG_EXPR, NRML,	6,	"33", "", "",		0, 2, 4, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(5)",	ARG_EXPR, NRML,	7,	"34", "", "",		0, 2, 5, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(6)",	ARG_EXPR, NRML,	8,	"35", "", "",		0, 2, 6, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(7)",	ARG_EXPR, NRML,	9,	"36", "", "",		0, 2, 7, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(8)",	ARG_EXPR, NRML,	10,	"37", "", "",		0, 2, 8, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LC(N)",	ARG_NIB, NRML,	2,	"30", "", "",		0, 1, 1, -1, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LCASC",	ARG_LASC, NRML,	-1,	"30", "", "",		0, 2, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LCHEX",	ARG_LHEX, NRML,	-1,	"30", "", "",		0, 2, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LCSTR",	ARG_LSTR, NRML,	-1,	"30", "", "",		0, 2, -1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LINK",	ARG_LINK, NRML,	5,	"", "", "",		0, 0, 5, 0, CLASS_LINK, 1, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "LIST",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_LIST, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LISTALL",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_LISTALL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "LISTM",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_LISTM, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "MACRO",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_MACRO, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "MESSAGE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_MSG, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "MP=0",	ARG_NONE, NRML,	3,	"828", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "NEXT",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_NEXT, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "NIBASC",	ARG_ASC, NRML,	-1,	"", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "NIBBIN",	ARG_BIN, NRML, -1,	"", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "NIBFS",	ARG_FS, NRML,	1,	"0", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "NIBGRB",	ARG_BIN, NRML, -1,	"", "", "",		0, 0, 0, -1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "NIBHEX",	ARG_HEX, NRML,	-1,	"", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "NOP3",	ARG_NONE, NRML,	3,	"820", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "NOP4",	ARG_NONE, NRML,	4,	"6300", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "NOP5",	ARG_NONE, NRML,	5,	"64000", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "NOTREACHED", ARG_NONE, PSEU, 0,	"", "", "",		0, 0, 0, 0, CLASS_NOTREACH, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "OUT=C",	ARG_NONE, NRML,	3,	"801", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "OUT=CS",	ARG_NONE, NRML,	3,	"800", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "P=",	ARG_NIB, NRML,	2,	"20", "", "",		0, 1, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "P=C",	ARG_NIB, NRML,	4,	"80D0", "", "",		0, 3, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "P=P+1",	ARG_NONE, NRML,	2,	"0C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "P=P-1",	ARG_NONE, NRML,	2,	"0D", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "PC=(A)",	ARG_NONE, CPU1,	4,	"808C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "PC=(C)",	ARG_NONE, CPU2,	4,	"808E", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "PC=A",	ARG_NONE, CPU2,	4,	"81B2", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "PC=C",	ARG_NONE, CPU2,	4,	"81B3", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "R0=A",	ARG_NONE, NRML,	3,	"100", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R0=A.F",	ARG_FS, SHA2,	-1,	"81A000", "100", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R0=C",	ARG_NONE, NRML,	3,	"108", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R0=C.F",	ARG_FS, SHA2,	-1,	"81A008", "108", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R1=A",	ARG_NONE, NRML,	3,	"101", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R1=A.F",	ARG_FS, SHA2,	-1,	"81A001", "101", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R1=C",	ARG_NONE, NRML,	3,	"109", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R1=C.F",	ARG_FS, SHA2,	-1,	"81A009", "109", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R2=A",	ARG_NONE, NRML,	3,	"102", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R2=A.F",	ARG_FS, SHA2,	-1,	"81A002", "102", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R2=C",	ARG_NONE, NRML,	3,	"10A", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R2=C.F",	ARG_FS, SHA2,	-1,	"81A00A", "10A", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R3=A",	ARG_NONE, NRML,	3,	"103", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R3=A.F",	ARG_FS, SHA2,	-1,	"81A003", "103", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R3=C",	ARG_NONE, NRML,	3,	"10B", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R3=C.F",	ARG_FS, SHA2,	-1,	"81A00B", "10B", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R4=A",	ARG_NONE, NRML,	3,	"104", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R4=A.F",	ARG_FS, SHA2,	-1,	"81A004", "104", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R4=C",	ARG_NONE, NRML,	3,	"10C", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "R4=C.F",	ARG_FS, SHA2,	-1,	"81A00C", "10C", "",	3, 0, 0, 0, CLASS_NORMAL, 0, 7, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "RDSYMB",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_RDSYMB, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "REL",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_LABEL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "REL(1)",	ARG_LABEL, NRML, 1,	"", "", "",		0, 0, 1, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(2)",	ARG_LABEL, NRML, 2,	"", "", "",		0, 0, 2, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(3)",	ARG_LABEL, NRML, 3,	"", "", "",		0, 0, 3, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(4)",	ARG_LABEL, NRML, 4,	"", "", "",		0, 0, 4, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(5)",	ARG_LABEL, NRML, 5,	"", "", "",		0, 0, 5, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(6)",	ARG_LABEL, NRML, 6,	"", "", "",		0, 0, 6, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(7)",	ARG_LABEL, NRML, 7,	"", "", "",		0, 0, 7, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "REL(8)",	ARG_LABEL, NRML, 8,	"", "", "",		0, 0, 8, 0, CLASS_LABEL, 2, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "RESET",	ARG_NONE, NRML,	3,	"80A", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "RSI",	ARG_NONE, CPU1,	5,	"80810", "", "",	0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "RSTK=C",	ARG_NONE, NRML,	2,	"06", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "RTI",	ARG_NONE, NRML,	2,	"0F", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "RTN",	ARG_NONE, NRML,	2,	"01", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "RTNC",	ARG_NONE, NRML,	3,	"400", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_NC,	NULL,	NULL,	NULL },
 { "RTNCC",	ARG_NONE, NRML,	2,	"03", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "RTNNC",	ARG_NONE, NRML,	3,	"500", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_SC,	NULL,	NULL,	NULL },
 { "RTNSC",	ARG_NONE, NRML,	2,	"02", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "RTNSXM",	ARG_NONE, NRML,	2,	"00", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_00,	NULL,	NULL,	NULL },
 { "RTNYES",	ARG_NONE, BRCH,	2,	"00", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_CC,	NULL,	NULL,	NULL },
 { "SB=0",	ARG_NONE, NRML,	3,	"822", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SETCARRY",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_SETCARRY, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SETDEC",	ARG_NONE, NRML,	2,	"05", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SETFLAG",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_SETFLAG, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SETHEX",	ARG_NONE, NRML,	2,	"04", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SETLIST",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_SETLIST, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SHUTDN",	ARG_NONE, NRML,	3,	"807", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SLINK",	ARG_LINK, NRML,	5,	"", "", "",		0, 0, 5, 0, CLASS_LINK, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "SR=0",	ARG_NONE, NRML,	3,	"824", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "SREQ?",	ARG_NONE, NRML,	3,	"80E", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ST=0",	ARG_NIB, NRML,	3,	"840", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ST=1",	ARG_NIB, NRML,	3,	"850", "", "",		0, 2, 1, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "ST=C",	ARG_NONE, NRML,	2,	"0A", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "STITLE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_STITLE, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "STRING",	ARG_ASC, NRML,	-1,	"", "", "",		0, 0, 0, 1, CLASS_NORMAL, 0, 0, 0, CR_XX,	NULL,	NULL,	NULL },
 { "TITLE",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_TITLE, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "UNCNFG",	ARG_NONE, NRML,	3,	"804", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
/* { "UNDEF",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_UNDEF, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL }, */
 { "UNLIST",	ARG_NONE, PSEU,	0,	"", "", "",		0, 0, 0, 0, CLASS_UNLIST, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "XM=0",	ARG_NONE, NRML,	3,	"821", "", "",		0, 0, 0, 0, CLASS_NORMAL, 0, 0, 0, CR_EQ,	NULL,	NULL,	NULL },
 { "",		0, 0,		0,	"", "", "",		0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL }
};

/* mnem_left is the smallest accepted left child
 * mnem_right-1 is the biggest accepted right child
 */

/* Sort order: Smaller mnemonics in left child!! */

void buildoptable(int mnem_left, int mnem_middle, int mnem_right)
{
	int ind_left, ind_right;

	ind_left = mnem_left + (mnem_middle - mnem_left) / 2;		/* Mathematical left child */
	ind_right = mnem_middle + (mnem_right - mnem_middle) / 2;	/* Mathematical right child */

	if(ind_left < mnem_middle) {
		(&mnemonics[mnem_middle])->op_left = &mnemonics[ind_left];
		buildoptable(mnem_left, ind_left, mnem_middle);
	}
	if((ind_right > mnem_middle) && (ind_right < mnem_right)) {
		(&mnemonics[mnem_middle])->op_right = &mnemonics[ind_right];
		buildoptable(mnem_middle + 1, ind_right, mnem_right);
	}
}

void readoptable(void)
{
	int mnem_count = 0, mnem_root;
	struct optable	*mnemp = mnemonics;

	/* First remove new opcodes acording to code_level */
	while(strlen(mnemp->op_mnem) != 0)
		if((mnemp->op_flags & (CPU_SWITCH1 | CPU_SWITCH2)) <= code_level)
			mnemonics[mnem_count++] = *mnemp++;
		else
			mnemonics[mnem_count] = *mnemp++;
	/* Choose root mnemonics */
	mnem_root = mnem_count / 2;
	optable_root = &mnemonics[mnem_root];
	/* Now recursively assign children for the accepted mnemonics */
	buildoptable(0, mnem_root, mnem_count);
}
