/*
 * opcode.c - SATURN assembler, mnemonics parser
 * (c) 1994 Mario Mikocevic Mozgy (Mario.Mikocevic@etf.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#ifndef	macintosh
#  include <malloc.h>
#endif

#include "opc.h"
#include "opcode.h"

unsigned	int	flags;
int	expr_resolved, offset, carry, f_ptr = -1;
unsigned	char	class;
char	f_stack[IF_STACK_MAX][SYM_NAME_LEN + 1], f_elsedone[IF_STACK_MAX];
char	symname[CHAR_BUF_LEN + 1];
FILE	*symfile;
struct	optable	*entry, *macroentry, *macro_root;
struct	macro	**lastmacro = NULL;
struct	loop	*fnentry = NULL;
struct	OPCODE	null_opcode = {	"",	0L	}, this_opcode;

static	int	was_test;
/* static	FILE	*opfile; */
static 	struct	expression	abs_expr;
static	union {
	char	c[256];
	struct {
		char	nam[O_SYM_MOD];
		struct o_symbol	s[O_SYM_CNT];
	} sym;
} in;
static struct fs fs_table[10] = {
	{	"P",	0	},
	{	"WP",	1	},
	{	"XS",	2	},
	{	"X",	3	},
	{	"S",	4	},
	{	"M",	5	},
	{	"B",	6	},
	{	"W",	7	},
	{	"A",	15	},
	{	"",	-1	},
};

extern	int	debug, code_level, assembling, addingmacro, pass1, pass2;
extern	int	absolute, begin_pseudos_ok, imtype, impending, column;
extern	int	listflags, listthisline, listaddress, codeonly, combos;
extern	int	lineofpage, pagelength, o_ptr, e_ptr;
extern	int	have_start_addr, reading, rel_expressions_ok, stdinread;
extern	int	insideloop;
extern	long	pc, codestart, listcount, abase;
extern	long	loops[LOOP_STACK_MAX][3];
/* obsolete with opc.c */
/* extern 	char	*opname; */
extern	char	*dbgname, *inname, *nameptr, label[], *ptr, charset[];
extern	char	*linestart, prog_symbol[], flag[], title[], subtitle[];
extern	char	dectab[], hextab[];
extern	FILE	*dbgfile, *infile, *symfile;
extern	struct optable *optable_root;

void	zerrmsg( char	*str )
{
	this_opcode.value[0] = '\0';
	this_opcode.length = 0;
	errmsg( str );
}

char	*uppercase( char *str )
{
	char	*sp = str - 1, c;

	while( ( c = *++sp ) != '\0' ) {
		if( strchr( "abcdefghijklmnopqrstuvwxyz", c ) )
			*sp -= 0x20;
	}
	return str;
}

struct optable	*getopmem( int num_entries )
{
	struct optable *opt;

	if( ( opt = (struct optable *)calloc( num_entries, sizeof( struct optable ) ) ) == NULL )
		error( "error getting memory for optable of %d entries", num_entries );
	DEBUG_PRN( "Address %ld: allocated %d bytes\n", opt, num_entries * sizeof( struct optable ) );
	return opt;
}

struct optable	*newmacro( char *name )
{
	struct optable	*dummy;

	if( ( dummy = getopmem( 1 ) ) == NULL )
		error( "%s", "error getting memory for macro entry" );
	strncpy( dummy->op_mnem, name, SYM_NAME_LEN );
	dummy->op_operand = ARG_NONE;
	dummy->op_flags = ( OP_MACRO | OP_PSEUDO );
	dummy->op_length = -1;
	dummy->op_string[0] = '\0';
	dummy->alt1_op_string[0] = '\0';
	dummy->alt2_op_string[0] = '\0';
	dummy->op_fs_nib = 0;
	dummy->op_offset = 0;
	dummy->op_size = 0;
	dummy->op_class = CLASS_NORMAL;
	dummy->op_subclass = CLASS_NORMAL;
	dummy->op_Afield = 0;
	dummy->op_Bfield = 0;
	dummy->op_macroptr = NULL;
	lastmacro = &dummy->op_macroptr;
	add_optable( dummy, &macro_root );
	return dummy;
}

struct loop	*newloop( void )
{
	struct loop *dummy;

	if( ( dummy = (struct loop *)calloc( 1, sizeof( struct loop ) ) ) == NULL )
		error( "error getting memory for for-next loop" );
	DEBUG_PRN( "Address %ld: allocated %d bytes\n", dummy, sizeof( struct loop ) );
	return dummy;
}

struct optable	*add_optable( struct optable *tbl_pointer, struct optable **lastptr )
{
	register int	compare, done = FALSE;

	do {
		if( *lastptr != NULL )
			if( ( compare = strcmp( (*lastptr)->op_mnem, tbl_pointer->op_mnem ) ) != 0 )
				lastptr = ( compare < 0 ) ? &(*lastptr)->op_right : &(*lastptr)->op_left;
			else
				done = TRUE;
		else {
			*lastptr = tbl_pointer;
			tbl_pointer->op_left = NULL;
			tbl_pointer->op_right = NULL;
			done = TRUE;
		}
	} while( !done );
	return *lastptr;
}

int	pushcond( void )
{
	if( ++f_ptr >= IF_STACK_MAX ) {
		errmsg( "Conditional assembly stack overflow" );
		f_ptr--;
		return FALSE;
	}
	strncpy( f_stack[f_ptr], label, SYM_NAME_LEN );
	f_elsedone[f_ptr] = '\0';
	return TRUE;
}

int	checkcond( void )
{
	if( f_ptr != -1 )
		if( !strncmp( f_stack[f_ptr], label, SYM_NAME_LEN ) )
			return TRUE;
	return FALSE;
}

/* obsolete with opc.c */
/*
void	readoptable( void )
{
	int i, optablesize;
	struct optable *opptr;

	if( ( opfile = openfile( opname, "SASM_LIB", "\\lib\\sasm;\\lib", "rb", NULL ) ) == NULL )
		error( "error opening %s for reading", opname );
	i = fscanf( opfile, "Opcodes-4:%d\n", &optablesize );
	if( ferror( opfile ) )
		error( "error reading from opcode table file (%s)", nameptr );
	if( i != 1 )
		error( "%s is not an opcode table file", nameptr );
	if( ( opptr = getopmem( optablesize ) ) == NULL )
		error( "%s", "error getting memory for opcode table" );
	if( ( i = fread( opptr, sizeof( struct optable ), optablesize, opfile ) ) != optablesize )
		error( "error reading from opcode table file, item %ld", i + 1 );
	while( optablesize-- > 0 ) {
		if( ( opptr->op_flags & ( CPU_SWITCH2 | CPU_SWITCH1 ) ) <= code_level )
			add_optable( opptr, &optable_root );
		opptr++;
	}
	fclose( opfile );
	FREE( nameptr );
	DEBUG_PRN( "Address %ld: free'd memory\n", nameptr );
}
 */

void	alloptable( void doit(struct optable *), struct optable *root )
{
	struct optable	*dummy;

	if( root != NULL ) {
		if( root->op_left != NULL )
			alloptable( doit, root->op_left );
		dummy = root->op_right;
		doit( root );
		if( dummy != NULL )
			alloptable( doit, dummy );
	}
}

struct optable	*findoptable( char *mnem_in, struct optable *node )
{
	int compare;

	do {
		if( node == NULL )
			return node;
		if( ( compare = strcmp( node->op_mnem, mnem_in ) ) == 0 )
			return node;
		node = ( compare < 0 ) ? node->op_right : node->op_left;
	} while ( 1 );
}

struct optable	*all_findoptable( char *mnem_in )
{
	struct optable	*dummy;

	if( ( dummy = findoptable( mnem_in, optable_root ) ) == NULL )
		if( ( dummy = findoptable( mnem_in, macro_root ) ) == NULL ) {
			mnem_in = uppercase( mnem_in );
			if( ( dummy = findoptable( mnem_in, optable_root ) ) == NULL )
				dummy = findoptable( mnem_in, macro_root );
		}
	return dummy;
}

int	fromhex( char c )
{
	return ( c > '9' ) ? c - '7' : c - '0';
}

char	*getterm( void )
{
	if( *ptr == '"' ) {
		ptr++;
		return "\"";
	}
	if( *ptr == '\'' ) {
		ptr++;
		return "'";
	}
	if( *ptr == '<' ) {
		ptr++;
		return ">";
	}
	return " \t";
}

int	getfs( void )
{
	int	i;
	char	fs[CHAR_BUF_LEN + 1];

	if( !getfield( fs, CHAR_BUF_LEN, " \t," ) )
		return -1;
	uppercase( fs );
	i = 0;
	while( i < 9 && strcmp( fs_table[i].string, fs ) )
		i++;
	return fs_table[i].value;
}

struct expression	*getdigitexpr( void )
{
	struct expression	*dummy;

	if( ( dummy = doexpr_fill() ) != NULL ) {
		if( dummy->e_relcount != 0 ) {
			errmsg( "Relocatable offset not allowed here" );
			return NULL;
		}
		if( expr_resolved != 0 ) {
			if( dummy->e_value > 15L || dummy->e_value < 0L )
				errmsg( "Expression out of range" );
		}
	}
	return dummy;
}

int	getasc( char *dest, int len )
{
	int value, valid = FALSE;
	char *bp, buffer[CHAR_BUF_LEN + 1];
	char termchr[2];

	termchr[1] = '\0';
	if( ( termchr[0] = *ptr ) != '\0' ) {
		ptr++;
		if( getfield( buffer, CHAR_BUF_LEN, termchr ) && *ptr == termchr[0] ) {
			ptr++;
			bp = buffer;
			while( ( value = decode( &bp ) ) != -1 && len-- > 1 ) {
				value = charset[value];
				*dest++ = hextab[value & 0x0F];
				len--;
				*dest++ = hextab[( value >> 4 ) & 0x0F];
				valid = TRUE;
			}
		}
	}
	*dest = '\0';
	if( valid && *bp == '\0' )
		return TRUE;
	else
		return FALSE;
}

int	gethex( char *dest, int len )
{
	int	valid = FALSE;
	char	*bp = NULL, buffer[MAXOPLINE + 1];

	if( getfield( buffer, MAXOPLINE, " \t" ) ) {
		bp = buffer;
		if( *bp == '#' )
			bp++;
		while( *bp != '\0' && strchr( hextab, *bp  ) && len-- > 0 ) {
			*dest++ = ( strchr( "abcdef", *bp ) ) ? *bp - 0x20 : *bp;
			bp++;
			valid = TRUE;
		}
	}
	*dest = '\0';
	if( valid && *bp == '\0' )
		return TRUE;
	else
		return FALSE;
}

int	getbin( char *dest, int len, int straight )
{
	int	valid = FALSE, bits = 0, value = 0;
	char	*bp = NULL, buffer[MAXOPLINE + 1];
     
	if( getfield( buffer, MAXOPLINE, " \t" ) ) {
		bp = buffer;
		if( *bp == '%' )
			bp++;
		while( ( *bp == '0' || *bp == '1' ) && len-- > 0 ) {
			/* Or with #8 #4 #2 or #1 if '1' */
			if( *bp == '1' )
				if( straight )
					value |= 1 << ( 3 - bits );
				else
					value |= 1 << bits;
			if( ++bits == 4 ) {
				*dest++ = hextab[value];
				value = bits = 0;
			}
			bp++;
			valid = TRUE;
		}
	}
	/* Output cached bits */
	if( bits != 0 )
		*dest++ = hextab[value];
	*dest = '\0';
	if( valid && *bp == '\0' )
		return TRUE;
	else
		return FALSE;
}

unsigned long	getlength(void)
{
	unsigned long	len_expr, len_value;

	len_value = (unsigned long)strlen(this_opcode.value);
	if(entry->op_operand > ARG_FIXED_SIZE) {
		len_expr = (unsigned long)entry->op_size + (unsigned long)offset;
		while(len_expr > len_value)
			this_opcode.value[(short)(len_value++)] = '0';
		this_opcode.value[(short)len_value] = '\0';
	}
	return len_value;
}

struct o_symbol	*getobject(int in_index)
{
	char	string[256];
	struct o_symbol	*sym_rec;

	while(in_index >= O_SYM_CNT) {
		if(fread(in.c, 256, 1, symfile) != 1) {
			sprintf(string, "Error reading from RDSYMB file (%s)", nameptr);
			errmsg(string);
			return NULL;
		}
		in_index -= O_SYM_CNT;
	}
	sym_rec = in.sym.s + in_index;
	in_index += (get_u16((unsigned char *)sym_rec->os_fillcount) & 0x3F) + 1;
	return sym_rec;
}

struct OPCODE *getopcode(char *mnem_in)
{
	char	*saveptr = ptr, *cp = NULL;
	unsigned char	cstate;
	int	savecol = column, i, field;
	long	reloffset, maxoffset;
	struct symbol *symbptr;
	struct expression *exprptr;

	if((entry = all_findoptable(mnem_in)) != NULL) {
		offset = entry->op_offset;
		flags = entry->op_flags;
		class = entry->op_class;
	} else {
		if(assembling && !addingmacro) {
			errmsg("Illegal mnemonic");
			return &null_opcode;
		}
	}
	if(!assembling &&
		(entry == NULL || !(flags & OP_PSEUDO) ||
		(class != CLASS_ELSE && class != CLASS_ENDIF &&
		(class != CLASS_ENDM || imtype != IMTYPE_MACRO)))) {
			label[0] = '\0';
			listthisline = FALSE;
			listflags |= LIST_DISABLELINE;
			return &null_opcode;
	}
	if( addingmacro ) {
		if( class == CLASS_FOR ) {
			errmsg( "Can't nest FOR inside MACRO" );
			listaddress = FALSE;
		} else {
			listaddress = FALSE;
			if( entry == NULL || !(flags & OP_PSEUDO ) ||
				 class != CLASS_ENDM ||
				( label[0] != '\0' &&
				  strncmp( macroentry->op_mnem, label, SYM_NAME_LEN ) ) )
				putmacroline();
			else
				addingmacro = FALSE;
		}
		label[0] = '\0';
		return &null_opcode;
	}
	if( ( flags & CPU_MASK ) > code_level ) {
		errmsg( "Instruction not allowed with this CPU" );
		return &null_opcode;
	}
	if( ( flags & OP_COMBOS ) && !combos ) {
		errmsg( "COMBOS not allowed in SASM compatability mode" );
		return &null_opcode;
	}
	if(flags & OP_MACRO) {
		if(pass2 && !(flags & OP_SHORTA)) {
			errmsg("Illegal mnemonic");
			return &null_opcode;
		}
		impending = IMP_MACRO;
		if(listflags & LIST_MACROS)
			listthisline = TRUE;
		return &null_opcode;
	}
	strcpy(this_opcode.value, entry->op_string);
	this_opcode.length = 0;
	if(flags & OP_PSEUDO) {
		listaddress = 0;
		switch(class) {
			case CLASS_NORMAL:
			case CLASS_LABEL:
				listaddress = TRUE;
				if(!begin_pseudos_ok)
					errmsg("ABS/REL must be at beginning");
				else {
					if((exprptr = doexpr()) != NULL)
						if(absexpr(exprptr))
							if(exprptr->e_value < 0x100000L) {
								if(have_start_addr)
									pc = codestart = exprptr->e_value;
							} else
								zerrmsg("Value too big");
						else
							zerrmsg("Expression must be absolute");
					else
						zerrmsg("Invalid expression");
					absolute = (class == CLASS_NORMAL) ? TRUE : FALSE;
					cp = label;
					symbptr = addsymbol("*START", codestart, (SYM_DEFINED | SYM_RESOLVED));
					symbptr->s_value = codestart;
					symbptr->s_flags = (SYM_DEFINED | SYM_RESOLVED);
					if(!absolute)
						symbptr->s_flags |= SYM_RELATIVE;
					if(*cp != '\0') {
						if(*cp == '=')
							cp++;
						strncpy(prog_symbol, cp, 13);
						i = (SYM_GLOBAL | SYM_RESOLVED);
						if(!absolute)
							i |= SYM_RELATIVE;
						symbptr = addsymbol(cp, pc, i);
						if(pass1)
							addlineref(symbptr);
						if(pass2) {
							if(symbptr->s_flags & SYM_DEFINED)
								errmsg("Op: Duplicate symbol");
							else {
								symbptr->s_flags |= SYM_DEFINED;
								if(symbptr->s_value != pc) {
									errmsg("PC changed (use old value)");
									pc = symbptr->s_value;
								}
							}
						}
					}
				}
				label[0] = '\0';
				begin_pseudos_ok = FALSE;
				break;

			case CLASS_LINK:
				listaddress = TRUE;
				if((exprptr = doexpr()) != NULL)
					if(absexpr(exprptr))
						if(exprptr->e_value >= 0)
							if((exprptr->e_value + pc) > 0x100000L)
								zerrmsg("Number of nibbles too big for file");
							else
								this_opcode.length = exprptr->e_value;
						else
							zerrmsg("Number of nibbles must be non-negative");
					else
						zerrmsg("Expression must be absolute");
				else
					zerrmsg("Invalid expression");
				break;

			case CLASS_END:
				listaddress = TRUE;
				if(imtype == IMTYPE_CODE)
					reading = FALSE;
				else
					impending = IMP_END;
				break;

			case CLASS_IF:
			case CLASS_IFEXPR:
				if(pushcond()) {
					unsigned char	results;
					int	exp;

					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
					if((exprptr = doexpr()) != NULL) {
						if(absexpr(exprptr)) {
							/* this could be SUBCLASS flags too */
							results = ((exp = (int)exprptr->e_value) != 0) ? (exp < 0) ? 1 : 4 : 2;
							if(class == CLASS_IF) {
								if(exp < 100)
									assembling = (flag[exp] & TRUE) ? TRUE : FALSE;
								else
									zerrmsg("Flag value out of range");
							} else
								assembling = (entry->op_subclass & results);
						} else
							zerrmsg("IF expression must be absolute");
					} else
						zerrmsg("Invalid flag expression");
				}
				label[0] = '\0';
				break;

			case CLASS_IFSTR:
				if(pushcond()) {
					int	results;
					char	save1, save2, sep, *s1, *s2;

					sep = *ptr++;
					s1 = ptr;
					s2 = strchr(ptr, sep);
					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
					if(s2 != NULL && (cp = strchr(s2 + 1, sep)) != NULL) {
						save1 = *s2;
						*s2++ = '\0';
						save2 = *cp;
						*cp = '\0';
						results = strcmp(s1, s2);
						*cp = save2;
						*--s2 = save1;
						results = (results == 0) ? SUBCLASS_EQ : (results < 0) ? SUBCLASS_LESS : SUBCLASS_GREAT ;
						assembling = (entry->op_subclass & results) ? TRUE : FALSE;
					} else
						zerrmsg("Missing separator character");
				}
				label[0] = '\0';
				break;

			case CLASS_IFDEF:
			case CLASS_IFNDEF:
				if( pushcond() ) {
					listthisline = ( listflags & LIST_PSEUDO ) ? TRUE : FALSE;
					if( getfield( symname, CHAR_BUF_LEN, " \t" ) ) {
						if( ( symbptr = findsymbol( symname ) ) != NULL &&
							( pass1 || ( pass2 && ( symbptr->s_flags & SYM_DEFINED ) ) ) )
							assembling = (class == CLASS_IFDEF) ? TRUE : FALSE;
						else
							assembling = (class == CLASS_IFNDEF) ? TRUE : FALSE;
						if(pass2 && symbptr != NULL)
							addlineref(symbptr);
					} else
						zerrmsg("Invalid symbol name");
				}
				label[0] = '\0';
				break;

			case CLASS_IFOPC:
			case CLASS_IFNOPC:
				if(pushcond()) {
					int	fl;
					struct optable	*opt;

					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
					if(getfield(symname, CHAR_BUF_LEN, " \t")) {
						if((opt = all_findoptable(symname)) != NULL) {
							fl = opt->op_flags;
							if(pass2 && (fl & OP_MACRO) && !(fl & OP_SHORTA))
								opt = NULL;
						}
						if(opt != NULL)
							assembling = (class == CLASS_IFOPC) ? TRUE : FALSE;
						else
							assembling = (class == CLASS_IFNOPC) ? TRUE : FALSE;
					} else
						zerrmsg("Invalid symbol name");
				}
				label[0] = '\0';
				break;

			case CLASS_IFPASS1:
			case CLASS_IFPASS2:
				if(pushcond()) {
					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
					if(pass1 && class == CLASS_IFPASS1)
						assembling = TRUE;
					else {
						if(!pass2)
							assembling = FALSE;
						else
							assembling = (class == CLASS_IFPASS2) ? TRUE : FALSE;
					}
				}
				label[0] = '\0';
				break;

			case CLASS_IFCRY1:
			case CLASS_IFCRY0:
			case CLASS_IFCRYANY:
			case CLASS_IFREACHED:
				if(pushcond()) {
					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
					switch(class) {
						case CLASS_IFCRY1:
							assembling = (carry == 2) ? TRUE : FALSE;
							break;
						case CLASS_IFCRY0:
							assembling = (carry == 1) ? TRUE : FALSE;
							break;
						case CLASS_IFCRYANY:
							assembling = (carry == 3) ? TRUE : FALSE;
							break;
						case CLASS_IFREACHED:
							assembling = (carry != 0) ? TRUE : FALSE; /* cmp ax,1 ; sbb ax,ax ; inc ax */
							break;
					}
				}
				label[0] = '\0';
				break;

			case CLASS_ELSE:
				if(checkcond()) {
					/* this could be named too */
					if(f_elsedone[f_ptr] == 0x00) {
						assembling = !assembling;
						f_elsedone[f_ptr] = 0x01;
						listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
					} else
						errmsg("Duplicate ELSE statement");
				} else
					if(assembling)
						errmsg("ELSE without matching IF");
				label[0] = '\0';
				break;

			case CLASS_ENDIF:
				if(checkcond()) {
					f_ptr--;
					assembling = TRUE;
					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				} else
					if(assembling)
						errmsg("ENDIF without matching IF");
				label[0] = '\0';
				break;

			case CLASS_MSG:
				fputs(ptr, stderr);
				fputc('\n', stderr);
				label[0] = '\0';
				break;

			case CLASS_RDSYMB:
				rdsymb();
				label[0] = '\0';
				break;

			case CLASS_INC:
				if((exprptr = doexpr()) != NULL) {
					if(exprptr->e_extcount == 0)
						addlabel(label, exprptr->e_value, ((exprptr->e_relcount != 0) ? SYM_RELATIVE : 0) | SYM_RESOLVED);
					else {
						zerrmsg("Expression can not be external");
						addlabel(label, 0L, SYM_RESOLVED);
					}
				} else {
					zerrmsg("Invalid expression");
					addlabel(label, 0L, SYM_RESOLVED);
				}
				label[0] = '\0';
				break;

			case CLASS_EQU:
				if((exprptr = doexpr()) != NULL) {
					if(exprptr->e_extcount == 0)
						xaddlabel(label, exprptr->e_value, ((exprptr->e_relcount != 0) ? SYM_RELATIVE : 0) | SYM_RESOLVED, TRUE);
					else {
						zerrmsg("Expression can not be external");
						xaddlabel(label, 0L, SYM_RESOLVED, TRUE);
					}
				} else {
					zerrmsg("Invalid expression");
					xaddlabel(label, 0L, SYM_RESOLVED, TRUE);
				}
				label[0] = '\0';
				break;

			case CLASS_TITLE:
				if(*title == '\0')
					strncpy(title, ptr, 40);
				if(strncmp(title, ptr, 40))
					errmsg("Only one TITLE statement allowed per file");
				label[0] = '\0';
				break;

			case CLASS_ABASE:
				if((exprptr = doexpr()) != NULL) {
					if(exprptr->e_value >= 0x100000L)
						zerrmsg("Too big base value");
					else
						abase = exprptr->e_value;
				} else
					zerrmsg("Invalid expression");
				break;

			case CLASS_ALLOC:
				if((exprptr = doexpr()) != NULL) {
					if(exprptr->e_extcount == 0) {
						addlabel(label, abase, ((exprptr->e_relcount != 0) ? SYM_RELATIVE : 0) | SYM_RESOLVED);
						abase += exprptr->e_value;
						if(abase >= 0x100000L) {
							zerrmsg("Allocation wrapped around zero");
							abase -= 0x100000L;
						}
					} else {
						zerrmsg("Expression can not be external");
						addlabel(label, 0L, SYM_RESOLVED);
					}
				} else {
					zerrmsg("Invalid expression");
					addlabel(label, 0L, SYM_RESOLVED);
				}
				label[0] = '\0';
				break;

			case CLASS_STITLE:
				strncpy(subtitle, ptr, 40);
			case CLASS_EJECT:
				if(listthisline) {
					if(lineofpage < pagelength)
					/* caution needed here! expression must yields 1 if true */
						lineofpage = pagelength - !(listflags & LIST_PSEUDO);
					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				}
				label[0] = '\0';
				break;

			case CLASS_LIST:
			case CLASS_LISTM:
			case CLASS_UNLIST:
				switch(class) {
					case CLASS_LIST:
						listflags |= LIST_CODE;
						break;
					case CLASS_LISTM:
						listflags |= (LIST_MACROS | LIST_INCLUDES);
						break;
					case CLASS_UNLIST:
						listflags &= ~(LIST_CODE | LIST_MACROS | LIST_INCLUDES);
						break;
				}
				listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				symbptr = addsymbol("*LISTFLAGS", 0L, SYM_RESOLVED);
				symbptr->s_value = listflags & LIST_MASK;
				label[0] = '\0';
				break;

			case CLASS_LISTALL:
				listcount = 0L;
				if((exprptr = doexpr()) != NULL) {
					if(absexpr(exprptr))
						listcount = exprptr->e_value;
					else
						zerrmsg("Expression not absolute");
				} else
					zerrmsg("Invalid expression");
				label[0] = '\0';
				break;

			case CLASS_SETLIST:
			case CLASS_CLRLIST:
				listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				i = 0;
				while(getfield(symname, CHAR_BUF_LEN, " \t,")) {
					uppercase(symname);
					switch(symname[0]) {
						case 'A':
							i |= (LIST_CODE | LIST_MACROS | LIST_INCLUDES | LIST_PSEUDO);
							break;
						case 'C':
							i |= LIST_CODE;
							break;
						case 'I':
							i |= LIST_INCLUDES;
							break;
						case 'M':
							i |= LIST_MACROS;
							break;
						case 'N':
							listthisline = FALSE;
							break;
						case 'P':
							i |= LIST_PSEUDO;
							break;
						default:
							errmsg("List type not CODE,MACRO,INCLUDE,PSEUDO,ALL, or NOLIST");
							break;
					}
					if(*ptr != '\0')
						ptr++;
					skipwhite();
				}
				if(class == CLASS_SETLIST)
					listflags |= i;
				else
					listflags &= ~i;
				symbptr = addsymbol("*LISTFLAGS", 0L, SYM_RESOLVED);
				symbptr->s_value = listflags & LIST_MASK;
				label[0] = '\0';
				break;

			case CLASS_MACRO:
				cp = label;
				if( *cp == '\0' ) {
					errmsg( "Unlabeled MACRO statement (*MACRO used)" );
					cp = "*MACRO";
				}
				if( ( macroentry = findoptable( cp, optable_root ) ) != NULL )
					errmsg( "Can't redefine an existing opcode" );
				if( pass1 && macroentry == NULL ) {
					if( findoptable( cp, macro_root ) != NULL )
						errmsg( "Can't redefine an existing macro" );
					macroentry = newmacro( cp );
				}
				if(pass2 && macroentry == NULL) {
					if((macroentry = findoptable(cp, macro_root)) == NULL)
						error("macro '%s' was not defined in pass 1", cp);
					if(macroentry->op_flags & OP_SHORTA)
						errmsg("Can't redefine an existing macro");
					macroentry->op_flags |= OP_SHORTA;
				}
				addingmacro = TRUE;
				listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				label[0] = '\0';
				return &null_opcode;

			case CLASS_ENDM:
			case CLASS_EXITM:
				if(class == CLASS_EXITM)
					impending = IMP_EXITM;
				else
					errmsg("ENDM and EXITM not permitted outside of a macro");
				label[0] = '\0';
				return &null_opcode;

			case CLASS_FOR:
				if( stdinread ) {
					errmsg( "Can't LOOP while reading stdin" );
					return &null_opcode;
				}
				cp = label;
				if( *cp == '\0' ) {
					errmsg( "Unlabeled FOR statement (*FOR used)" );
					cp = "*FOR";
				}
				for( i = ( LOOP_STACK_MAX - 1 ); i > 0 ; i-- ) {
					loops[i][0] = loops[i - 1][0];
					loops[i][1] = loops[i - 1][1];
					loops[i][2] = loops[i - 1][2];
				}
				/* add push stack */

/* only one FOR-NEXT loop allowed for now,
 * put spanning in following 'if'
 */
				if( pass1 || pass2 ) {
					fnentry = newloop();
					fnentry->lp_filename = inname;
					fnentry->lp_fpos = ftell( infile );
					strncpy( fnentry->lp_label, cp, SYM_NAME_LEN );
					fnentry->lp_value = loops[0][0] = 0L;
					fnentry->lp_end = loops[0][1] = -1L;
					fnentry->lp_step = loops[0][2] = 1L;
					if( isfield() && ( exprptr = doexpr() ) != NULL ) {
						fnentry->lp_value = loops[0][0] = exprptr->e_value;
						skipwhite();
						if( isfield() && ( exprptr = doexpr() ) != NULL ) {
							fnentry->lp_end = loops[0][1] = exprptr->e_value;
							skipwhite();
							if( isfield() && ( exprptr = doexpr() ) != NULL ) {
								fnentry->lp_step = loops[0][2] = exprptr->e_value;
							}
						}
					}
				}
				insideloop = TRUE;
				listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				label[0] = '\0';
				return &null_opcode;

			case CLASS_NEXT:
				if( fnentry == NULL ) {
					errmsg( "NEXT without FOR" );
					return &null_opcode;
				}
				fnentry->lp_value += fnentry->lp_step;
				loops[0][0] = fnentry->lp_value;
				if( fnentry->lp_value <= fnentry->lp_end ) {
				/* check that lp_filename belongs to inname !! */
					fseek( infile, fnentry->lp_fpos, SEEK_SET );
					listthisline = FALSE;
				} else {
					for( i = 0; i < ( LOOP_STACK_MAX - 1 ); i++ ) {
						loops[i][0] = loops[i + 1][0];
						loops[i][1] = loops[i + 1][1];
						loops[i][2] = loops[i + 1][2];
					}
				/* add pop stack */
					FREE( fnentry );
					fnentry = NULL;
					insideloop = FALSE;
					listthisline = (listflags & LIST_PSEUDO) ? TRUE : FALSE;
				}
				label[0] = '\0';
				return &null_opcode;

			case CLASS_EXITF:
				if( fnentry == NULL ) {
					errmsg( "EXITF without FOR" );
					return &null_opcode;
				}
				fnentry->lp_value = 0L;
				fnentry->lp_end = -1L;
				fnentry->lp_step = 1L;
				label[0] = '\0';
				return &null_opcode;

#if	0
			case CLASS_UNDEF:
				if(getfield(symname, CHAR_BUF_LEN, " \t")) {
					struct optable	*opt;

					if((opt = all_findoptable(symname)) != NULL) {
						/* remove optable entry from list */
					} else
						zerrmsg("Non existing opcode");
				} else
					zerrmsg("Invalid opcode");
				break;
#endif

			case CLASS_INCLUDE:
				cp = getterm();
				if(getfield(symname, CHAR_BUF_LEN, cp))
					impending = IMP_INCLUDE;
				else
					errmsg("Invalid file specifier");
				label[0] = '\0';
				break;

			case CLASS_CHARMAP:
				char_map();
				label[0] = '\0';
				break;

			case CLASS_SETFLAG:
			case CLASS_CLRFLAG:
				if((exprptr = doexpr()) != NULL) {
					if(absexpr(exprptr)) {
						i = (int)exprptr->e_value;
						if(i >= 0 && i < FLAGS_NUM) {
							if(class == CLASS_SETFLAG)
								flag[i] |= TRUE; /* 0x01 */
							else
								flag[i] &= ~TRUE;
						} else
							zerrmsg("Flag value out of range");
					} else
						zerrmsg("Flag expression must be absolute");
				} else
					zerrmsg("Invalid flag expression");
				label[0] = '\0';
				break;

			case CLASS_SETCARRY:
			case CLASS_CLRCARRY:
				carry |= (class == CLASS_SETCARRY) ? 2 : 1;
				label[0] = '\0';
				break;

			case CLASS_NOTREACH:
				carry = 0;
				label[0] = '\0';
				break;

			case CLASS_JUMP:
				if(label[0] != '\0') {
					DEBUG_PRN("Setting carry bits (labelled opcode)\n");
					carry = 3; /* hm? maybe FLAG0x?? */
				}
				switch(carry) {
					case 0:
						DEBUG_PRN("No code needed for PS_JUMP\n");
						return &null_opcode;
						/* break; */
					case 1:
						cp = entry->alt1_op_string;
						break;
					case 2:
						cp = entry->alt2_op_string;
						break;
					case 3:
						cp = entry->op_string;
						break;
				}
				DEBUG_PRN("Selected %s as new opcode for PS_JUMP\n", cp);
				if((entry = all_findoptable(cp)) == NULL)
					error("corrupt opcode table entry (%s not in table)", cp);
				strcpy(this_opcode.value, entry->op_string);
				offset = entry->op_offset;
				flags = entry->op_flags;
				class = entry->op_class;
				listaddress = 1;
				break;

			default:
				errmsg("Unknown PSEUDO-OP");
				break;
		}
	}
	switch(entry->op_operand) {
		case ARG_NONE:
			break;
		case ARG_FS:
		case ARG_FS1:
			field = getfs();
			if(field != -1 && *ptr != ',') {
				if(entry->op_operand == 16 &&
					(field == 0 || field == 2 || field == 4 || field == 1))
					errmsg("Single nibble field not allowed here");
				setfield(field);
			} else
				errmsg("Illegal field select");
			break;
		case ARG_FSN:
			field = getfs();
			if(field != -1 && *ptr != ',')
				setfield(field);
			else {
				ptr = saveptr;
				column = savecol;
				if((exprptr = getdigitexpr()) != NULL) {
					setnibble(this_opcode.value + offset - 1, 8);
					if(expr_resolved)
						setnibble(this_opcode.value + offset, (int)exprptr->e_value & 0x0F);
				} else
					errmsg("Invalid field select/digit");
			}
			break;
		case ARG_FSEXP:
		case ARG_FSEXP1:
			field = getfs();
			if(field != -1) {
				if(entry->op_operand == 15 &&
					(field == 0 || field == 2 || field == 4 || field == 1))
					errmsg("Single nibble field not allowed here");
				setfield(field);
				if(*ptr == ',')
					ptr++;
				skipwhite();
				if( ( exprptr = getdigitexpr() ) != NULL ) {
					if( expr_resolved )
						setnibble( this_opcode.value + offset, (int)exprptr->e_value & 0x0F );
				} else
					errmsg( "Invalid expression" );
			} else
				errmsg("Illegal field select");
			break;
		case ARG_NIB:
			if( ( exprptr = getdigitexpr() ) != NULL ) {
				if( expr_resolved )
					setnibble( this_opcode.value + offset, (int)exprptr->e_value & 0x0F );
			} else
				errmsg( "Invalid expression" );
			break;
		case ARG_LABEL:
			exprptr = NULL;
			if(rel_expressions_ok)
				exprptr = doexpr();
			else {
				expr_resolved = TRUE;
				o_ptr = -1;
				e_ptr = -1;
				if(getsymbol()) {
					if(*ptr != ')')
						exprptr = popexpr();
					else
						errmsg("Expression not allowed as target of branch");
				} else
					expr_resolved = FALSE;
			}
			if(exprptr != NULL) {
				if(exprptr->e_extcount == 0 && !absolute) {
					abs_expr.e_value = exprptr->e_value;
					exprptr = &abs_expr;
				}
				if(expr_resolved) {
					reloffset = exprptr->e_value - offset - pc;
					if(entry->op_fs_nib == 1)
						reloffset -= entry->op_size;
					maxoffset = (1L << ((entry->op_size << 2) - 1)) - 1;
					if((reloffset + 1) < (-maxoffset) || reloffset > maxoffset)
						errmsg("Branch out of range");
					setexpr(reloffset);
				}
				addfillref(exprptr);
			}
			break;
		case ARG_EXPR:
		case ARG_EXPR2:
			if((exprptr = doexpr_fill()) != NULL && expr_resolved)
				setexpr(exprptr->e_value);
			else if (codeonly && !expr_resolved) /* brb ! */
				zerrmsg("Unresolved expression");
			break;
		case ARG_BIN:
			if(!getbin(this_opcode.value + offset, MAXOPLINE - offset, (entry->op_bias == 0) ? TRUE : FALSE))
				zerrmsg("Invalid BIN constant");
			break;
		case ARG_HEX:
			if(!gethex(this_opcode.value + offset, MAXOPLINE - offset))
				zerrmsg("Invalid HEX constant");
			break;
		case ARG_LHEX:
			if(gethex(this_opcode.value + offset, MAXOPLINE - offset)) {
				reverse(this_opcode.value + offset);
				i = strlen(this_opcode.value + offset);
				if(i <= 16)
					setnibble(this_opcode.value + offset - 1, i - 1);
				else
					zerrmsg("Invalid HEX constant (too large)");
			} else
				zerrmsg("Invalid HEX constant");
			break;
		case ARG_DHEX:
			if(gethex(this_opcode.value + offset, MAXOPLINE - offset)) {
				i = strlen(this_opcode.value + offset);
				if(i == 2 || i == 4 || i == 5) {
					this_opcode.length = (long)(offset + i);
					reverse(this_opcode.value + offset);
					if(i != 2) {
						cp = this_opcode.value + offset - 1;
						*cp = hextab[fromhex(*cp) + i - 3];
					}
				} else
					zerrmsg("Invalid HEX constant (not 2, 4, or 5 digits)");
			} else
				zerrmsg("Invalid HEX constant");
			break;
		case ARG_LENHEX:
			if(gethex(this_opcode.value + offset, MAXOPLINE - offset)) {
				i = strlen(this_opcode.value + offset) + entry->op_bias;
				if((long)i >= (1L << 4 * offset))
					zerrmsg("Invalid HEX constant (too large)");
				else
					setnibbles(this_opcode.value, i, offset);
			} else
				zerrmsg("Invalid HEX constant");
			break;
		case ARG_LENASC:
			if(getasc(this_opcode.value + offset, MAXOPLINE - offset) && strchr(" \t", *ptr)) {
				i = strlen(this_opcode.value + offset) / 2 + entry->op_bias;
				if((long)i >= (1L << 4 * offset))
					zerrmsg("Invalid ASC constant (too large)");
			else
				setnibbles(this_opcode.value, i, offset);
			} else
				zerrmsg("Invalid ASC constant");
			break;
		case ARG_LSTR:
		case ARG_LASC:
			if(getasc(this_opcode.value + offset, MAXOPLINE - offset) &&
				strchr(" \t", *ptr)) {
				if(entry->op_operand == ARG_LASC)
					breverse(this_opcode.value + offset);
				i = strlen(this_opcode.value + offset);
				if(i <= 16)
					setnibble(this_opcode.value + offset - 1, i - 1);
				else
					zerrmsg("Invalid ASC constant (too large)");
			} else
				zerrmsg("Invalid ASC constant");
			break;
		case ARG_ASC:
			if(getasc(this_opcode.value + offset, MAXOPLINE - offset) &&
				strchr(" \t", *ptr)) {
				i = strlen(this_opcode.value);
				if(entry->op_bias == -1) {
					if(i >= MAXOPLINE - 3)  /* Correct comparison?? */
						zerrmsg("Invalid ASC constant (too long");
					else
						strcpy(this_opcode.value + i, "00");
				} else
					if(entry->op_bias == 1)
						setnibble(this_opcode.value + strlen(this_opcode.value) - 1, 8);
			} else
				zerrmsg("Invalid ASC constant");
			break;
		case ARG_LINK:
			exprptr = NULL;
			o_ptr = -1;
			e_ptr = -1;
			if(getsymbol()) {
				if(*ptr != ')')
					exprptr = popexpr();
				else
					errmsg("Expression not allowed for INC(x), LINK, and SLINK");
			}
			if(exprptr != NULL) {
				if(exprptr->e_extcount == 0 && exprptr->e_extcount == 0 &&
					!absolute) {
					abs_expr.e_value = exprptr->e_value;
					exprptr = &abs_expr;
				}
				addfillref(exprptr);
			}
			break;
		case ARG_INVALID:
		default:
			error("operand type (%ld) is not valid", entry->op_operand);
			break;
	}
	if( this_opcode.length == 0 )
		this_opcode.length = getlength();
	if( this_opcode.length != 0 ) {
		i = ( flags & OP_BRANCH ) | was_test;
		was_test = flags & OP_TEST;
		if( i == OP_TEST )
			errmsg( "Test instruction without GOYES/RTNYES" );
		if( i == OP_BRANCH )
			errmsg( "GOYES/RTNYES without test instruction" );
		if( label[0] != '\0' && ( flags & OP_BRANCH ) ) {
			errmsg( "Label not allowed on GOYES/RTNYES statement" );
			label[0] = '\0';
		}
	}
	if( label[0] != '\0' ) {
		DEBUG_PRN( "Setting carry bits (labelled opcode)\n" );
		carry = 3; /* ? hm, FLAG0x?? maybe */
	}
	cstate = entry->op_carrystate;
	DEBUG_PRN( "Setting carry state (AND with %1X, OR with %1X)\n", ( cstate >> 4 ) & 0x0F, cstate & 0x0F );
	carry = ( ( carry | cstate ) & ( cstate >> 4 ) ) & 0x0F;
	return &this_opcode;
}

void	setnibbles(char *cptr, int value, int nibbles)
{
	while(nibbles-- > 0) {
		*cptr++ = hextab[value & 0x0F];
		value = value >> 4;
	}
}

void	setnibble(char *cptr, int value)
{
	*cptr = hextab[(fromhex(*cptr) | value) & 0x0F];
}

void	setexpr(long value)
{
	int	i;
	char	*cptr;

	i = entry->op_size;
	cptr = this_opcode.value + offset;
	while(i-- > 0) {
		*cptr++ = hextab[(int)value & 0x0F];
		value >>= 4;
	}
	*cptr = '\0';
}

void	setfield(int field)
{
	if((entry->op_flags & OP_SHORTA) && entry->op_Afield == field) {
		strcpy(this_opcode.value, entry->alt1_op_string);
		return;
	}
	if((entry->op_flags & OP_SHORTB) && entry->op_Bfield == field) {
		strcpy(this_opcode.value, entry->alt2_op_string);
		return;
	}
	setnibble(this_opcode.value + entry->op_fs_nib, field);
}

unsigned long	opcodelen( char *mnemonic )
{
	unsigned char	carrystate;
	unsigned long	mylength = 0;
	struct OPCODE	*opc;

	if( ( entry = all_findoptable( mnemonic ) ) != NULL ) {
		if( !assembling || ( mylength = entry->op_length ) == -1 ||
			addingmacro || ( entry->op_flags & OP_PSEUDO ) ) {
			opc = getopcode( mnemonic );
			mylength = opc->length;
		} else {
			carrystate = entry->op_carrystate;
			DEBUG_PRN( "Setting carry state (AND with %1X, OR with %1X)\n", carrystate & 0x0F, ( carrystate >> 4 ) & 0x0F );
			carry = ( carry | carrystate ) & ( carrystate >> 4 ) & 0x0F;
		}
	} else
		if( addingmacro )
			putmacroline();
	return mylength;
}

void	rdsymb(void)
{
	unsigned int	symcount;
	int	in_index = O_SYM_CNT, mod_abs;
	long	symres;
	char	*term, string[CHAR_BUF_LEN];
	struct SATURN	buffer;
	struct o_symbol	*sym;

	term = getterm();
	if(getfield(symname, CHAR_BUF_LEN, term)) {
		if((symfile = openfile(symname, "SASM_RDSYMB", ";\\lib\\sasm", "rb", NULL)) == NULL) {
			sprintf(string, "Unable to open RDSYMB file (%s)", symname);
			errmsg(string);
			return;
		}
		if(fread(&buffer, 256, 1, symfile) == 1)
			if(!strncmp(buffer.id, "Saturn", 6))
				if(fseek(symfile, (((get_u32((unsigned char *)buffer.nibbles) + 511L) / 512L) * 256L) + 256L, SEEK_SET) == 0) {
					symcount = get_u16((unsigned char *)buffer.symbols);
					mod_abs = (int)buffer.absolute;
					while(symcount-- != 0 && (sym = getobject(in_index)) != NULL) {
						symres = (long)get_u16((unsigned char *)sym->os_fillcount) & 0xC000L;
						if( symres == 0x8000L || ( mod_abs != 0 && ( symres & 0x8000L ) != 0 ) )
							addlabel(sym->os_name,
								get_32((unsigned char *)sym->os_value),
								(SYM_FILE | SYM_RESOLVED));
					}
				} else {
					sprintf(string, "Error reading RDSYMB file (%s)", nameptr);
					errmsg(string);
				}
		fclose(symfile);
		FREE(nameptr);
		DEBUG_PRN("Address %ld: free'd memory\n", nameptr);
	} else
		errmsg("Invalid file specifier");
}

void char_map(void)
{
	int	byte1, byte2;
	char	*cp, string[CHAR_BUF_LEN];

	cp = getterm();
	if(getfield(symname, CHAR_BUF_LEN, cp)) {
		if((symfile = openfile(symname, "SASM_CHARMAP", ";\\lib\\sasm", "rt", NULL)) == NULL) {
			sprintf(string, "Unable to open CHARMAP file (%s)", symname);
			errmsg(string);
			return;
		}
		while((cp = fgets(string, CHAR_BUF_LEN, symfile)) != NULL) {
			byte1 = decode(&cp);
			byte2 = decode(&cp);
			if(byte1 == -1 || byte2 == -1)
				continue;
			charset[byte1] = byte2;
		}
		if(ferror(symfile)) {
			sprintf(string, "Error reading CHARMAP file (%s)", nameptr);
			errmsg(string);
		}
		fclose(symfile);
		FREE(nameptr);
		DEBUG_PRN("Address %ld: free'd memory\n", nameptr);
	} else
		errmsg("Invalid file specifier");
}

void	reverse(char *sp)
{
	char	save, *tmp;

	tmp = sp + strlen(sp) - 1;
	while(sp < tmp) {
		save = *sp;
		*sp++ = *tmp;
		*tmp-- = save;
	}
}

void	breverse(char *sp)
{
	char	save1, save2, *tmp;

	tmp = sp + strlen(sp) - 1;
	while(sp < tmp) {
		save1 = *sp;
		save2 = sp[1];
		*sp++ = *(tmp - 1);
		*sp++ = *tmp;
		*tmp-- = save2;
		*tmp-- = save1;
	}
}

void	putmacroline( void )
{
	if( pass1 && ( macroentry->op_flags & OP_MACRO ) ) {
		if( ( *lastmacro = (struct macro *)calloc( 1, sizeof( struct macro ) ) ) == NULL )
			error( "error getting memory for macro line '%s'", linestart );
		DEBUG_PRN( "Address %ld: allocated %d bytes\n", *lastmacro, sizeof( struct macro ) );
		if( ( (*lastmacro)->m_line = (char *)calloc( 1, strlen( linestart ) + 1 ) ) == NULL )
			error( "error getting memory for macro line '%s'", linestart );
		DEBUG_PRN( "Address %ld: allocated %d bytes\n", (*lastmacro)->m_line, strlen( linestart ) );
		strcpy( (*lastmacro)->m_line, linestart );
		(*lastmacro)->m_next = NULL;
		lastmacro = &(*lastmacro)->m_next;
	}

/* Old code, just leave it here ! */
/*		if((cp = malloc(strlen(linestart) + sizeof(struct macro) + 1)) == NULL)
			error("error getting memory for macro line '%s'", linestart);
		DEBUG_PRN("Address %ld: allocated %d bytes\n", cp, strlen(linestart) + sizeof(struct macro) + 1);
		*lastmacro = cp;
		(*lastmacro)->m_line = cp + 8;
		(*lastmacro)->m_next = NULL;
		strcpy((*lastmacro)->m_line, linestart);
		lastmacro++;
 */
}

int	decode( char **cptr )
{
	int	i, charvalue = 0;

	if( **cptr == '\0' )
		return -1;
	if( **cptr == '\\' ) {
		(*cptr)++;
		if( **cptr == '\0' )
			return -1;
		switch( **cptr ) {
			case 'a':
				charvalue = 7;
				break;
			case 'b':
				charvalue = 8;
				break;
			case 'f':
				charvalue = 12;
				break;
			case 'n':
				charvalue = 10;
				break;
			case 'r':
				charvalue = 13;
				break;
			case 't':
				charvalue = 9;
				break;
			case 'v':
				charvalue = 11;
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				i = 0;
				while( **cptr != '\0' && strchr( dectab, **cptr ) && i < 3 ) {
					charvalue = ( charvalue << 3 ) + *(*cptr)++ - '0';
					i++;
				}
				(*cptr)--;
				break;
			case 'x':
				(*cptr)++;
				i = 0;
				while( **cptr != '\0' && i < 2 && strchr( hextab, **cptr ) ) {
					charvalue <<= 4;
					if( strchr( dectab, **cptr ) )
						charvalue += **cptr - '0';
					else {
						if( strchr( "abcdef", **cptr ) )
							charvalue += **cptr + 0x20;
						charvalue += **cptr - '7';
					}
					(*cptr)++;
					i++;
				}
				(*cptr)--;
				if( i == 0 )
					charvalue = **cptr;
				break;
			default:
				charvalue = **cptr;
				break;
		}
	} else
		charvalue = **cptr;
	(*cptr)++;
	return charvalue;
}

void	dofree( struct optable *optr )
{
	struct macro *mp, *mpold;

	mpold = optr->op_macroptr;
	while( mpold != NULL ) {
		mp = mpold->m_next;
		FREE( mpold );
		DEBUG_PRN( "Address %ld: free'd memory\n", mpold );
		mpold = mp;
	}
	FREE( optr );
	DEBUG_PRN( "Address %ld: free'd memory\n", optr );
}

struct expression	*doexpr_fill( void )
{
	struct expression	*dummy;

	if( ( dummy = doexpr() ) != NULL )
		addfillref( dummy );
	return dummy;
}

void	cleanmacros( void )
{
	alloptable( dofree, macro_root );
	macro_root = NULL;
}
