/*
 * sasm.c - SATURN assembler, main loop
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_UNISTD_H
#  include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_STRING_H */
#ifndef	macintosh
#  include <malloc.h>
#endif
#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#else
#  include <time.h>
#endif /* HAVE_SYS_TIME_H */
#endif /* TIME_WITH_SYS_TIME */
#ifdef HAVE_DOS_H
#  include <dos.h>
#endif

#include "opc.h"
#include "sasm.h"

/* If you have troubles reading this file try
 * setting tabulators to 3 spaces */

int
	linelength, linewidth = 4,
	pagelength = 60, pagewidth = 0,
	codewidth = 5;
int	column, ccolumn = 30, source_col;
int	imptr, listflags, listcount, page, lineofpage;
int	offset_label_index = 0, out_position;
int	label_index_m2, label_index_m1, label_index_p1, label_index_p2;
int	code_level = CPU_SWITCH2;
int
	listdefault	=	TRUE,
	listing	=	TRUE,
	objdefault	=	TRUE,
	combos	=	TRUE,
	pass1_listing	=	FALSE,
	msg_append	=	FALSE,
	debug	=	FALSE,
	blanks_in_expr	=	FALSE,
	errors_to_stderr	=	FALSE,
	extended_errors	=	FALSE,
	codeonly	=	FALSE,
	hexcode	=	FALSE,
	rel_expressions_ok	=	FALSE,
	quote_ok	=	TRUE,
	force_global	=	FALSE,
	verbose	=	FALSE,
	have_start_addr	=	FALSE;
int
	singlefile = FALSE,
	stdinread = FALSE,
	listaddress = FALSE,
	pass1 = FALSE,
	pass2 = FALSE,
	absolute = FALSE,
	begin_pseudos_ok = FALSE,
	imtype = FALSE,
	impending = FALSE,
	reading = FALSE,
	assembling = FALSE,
	addingmacro = FALSE,
	insideloop = FALSE,
	listthisline = FALSE;
unsigned int ext_fillrefs, ext_symbols;
long	codestart = 0, nibble_offset, out_offset, pc, pc_offset;
long	abase, errorcount, first_error, last_error, mainline, currentline = 0;
long	loops[LOOP_STACK_MAX][3];
/* char	*opname; obsolete because of opc.c */
char	*progname, *inname, *objname, *listname, *dbgname;
char	*msgname, *curr_name;
char	*listprefix = NULL, *listsuffix = ".l";
char	*objprefix = NULL, *objsuffix = ".o";
char	*inbase, *objbase;
char	*linestart, *ptr;
char	dectab[] = "0123456789";
char	hextab[] = "0123456789ABCDEFabcdef";
char	basefmt[258], addrfmt[20], inclfmt[10], codefmt[25];
char	lastcodefmt[15], charset[256], out[256];
char	line[CHAR_BUF_LEN + 1], mnemonic[CHAR_BUF_LEN + 1];
char	label[CHAR_BUF_LEN + 1], tmpline[CHAR_BUF_LEN + 1];
char	date[26], prog_symbol[13], title[41], subtitle[41], flag[FLAGS_NUM];
char	*pptr[PPTR_STACK_MAX][9];
FILE	*infile, *listfile, *objfile, *dbgfile, *temp, *include;
struct macro	*macroptr;
struct imacro imstack[MACRO_STACK_MAX];
union {
	char	buffer[257];
	struct SATURN	saturn;
} headerout;

extern int	optind, carry, f_ptr;
extern char	symname[], *nameptr, *optarg, *version, *creation;
extern struct optable	*entry;
extern struct loop	*fnentry;
extern struct symbol	*s_root, *sys_root, s_terminal;

static int cum_errors = 0;

void	usage( int flag )
{
/*	fprintf(stderr, "SASM %s, Copyright (C) 1995 Mario Mikocevic Mozgy\n", version);
	fprintf(stderr, "SASM comes with ABSOLUTELY NO WARRANTY\n");
 */
	fprintf(stderr, "sasm %s\n", version);
	if( flag )
		return;
	fprintf(stderr, "Usage: %s [options] ... [sourcefile] ...\n", progname);
	fprintf(stderr, "options: (use -V? for full list)\n");
	fprintf(stderr, "-A\t\twrite an annotated listing to stdout [default for -s]\n");
	fprintf(stderr, "-a listfile\twrite an annotated listing to listfile\n");
	if(verbose)
		fprintf(stderr, "-b\t\tblanks OK within expressions [default = end of expression]\n");
	fprintf(stderr, "-c column\tfields starting after column 'column' are ignored\n");
	fprintf(stderr, "-D symbol[=value]\tdefine a symbol [default value = 1]\n");
	if(verbose)
		fprintf(stderr, "-d debugfile\tdebug mode, specified output file\n");
	fprintf(stderr, "-E\t\twrite C-like error messages to stderr\n");
	fprintf(stderr, "-e\t\twrite error messages to stderr\n");
	fprintf(stderr, "-f flaglist\tset assembly flags selected by flaglist (comma-separated)\n");
	if(verbose)
		fprintf(stderr, "-G\t\tforce all entry points to be global (overrides SASM_GLOBAL=0)\n");
	if(verbose)
		fprintf(stderr, "-g\t\tcancels forcing globals (overrides SASM_GLOBAL=1)\n");
	if(verbose)
		fprintf(stderr, "-H\t\tsuppress object file header and symbol (write raw code)\n");
	if(verbose)
		fprintf(stderr, "-h\t\tsuppress object file header and symbol (write hex code)\n");
	if(verbose)
		fprintf(stderr, "-L\t\tcreate a pass 1 listing if listing is enabled\n");
	if(verbose)
		fprintf(stderr, "-M msgfile\tredirect messages to msgfile (append to msgfile)\n");
	if(verbose)
		fprintf(stderr, "-m msgfile\tredirect messages to msgfile (overwrite msgfile)\n");
	fprintf(stderr, "-N\t\tdo not produce an annotated listing\n");
	if(verbose)
		fprintf(stderr, "-n nibls\tput 'nibls' nibbles of code per list line [default=%d]\n", codewidth);
	fprintf(stderr, "-o objfile\twrite the object file to objfile [default=<name>.o]\n");
	fprintf(stderr, "-P codetype\tset processor (0=1LF2, 1=1LK7, 2=1LR2 [default=%d])\n", code_level >> 6);
	fprintf(stderr, "-p pagelen\tdo a page break each 'pagelen' lines [default=%d]\n", pagelength);
	if(verbose)
		fprintf(stderr, "-q\t\tallow quote ['] as ASCII text delimiter [default unless -s]\n");
	fprintf(stderr, "-s\t\tforce SASM compatibility mode\n");
	if(verbose)
		fprintf(stderr, "-S nnnnn\toverride start address to hex nnnnn\n");
/* obsolete with opc.c */
/*	if(verbose)
		fprintf(stderr, "-t optable\tselect another opcode file [default=%s]\n", opname); */
	fprintf(stderr, "-V\t\tverbose mode (print a message when starting a new file)\n");
	fprintf(stderr, "-v\t\tprint version\n");
	fprintf(stderr, "-w width\tset output page width to 'width' columns [default=80]\n");
	if(verbose)
		fprintf(stderr, "-x\t\tallow expressions as operand of REL(n) and GOTO\n");
}

void errmsg( char *message )
{
	char	msgbuffer[CHAR_BUF_LEN + 1];

	if( listing && ( pass2 || pass1_listing ) ) {
		sprintf( msgbuffer, "*** ERROR: %s ***", message );
		putlist( msgbuffer );
		listthisline = TRUE;
	}
	if( pass2 ) {
		if( errors_to_stderr ) {
			if( !extended_errors ) {
				if( fprintf( stderr, "\"%s\", line %ld: %s\n",
					curr_name, ( currentline != 0 ) ? currentline : mainline, message ) < 0 )
					error( "%s", "error writing to stderr" );
			} else {
				if( fprintf( stderr, "*** ERROR in \"%s\" on line %ld: %s ***\n%s\n",
					curr_name, ( ( currentline != 0 ) ? currentline : mainline ), message , linestart ) < 0 )
					error( "%s", "error writing to stderr" );
			}
		}
		if( last_error != currentline )
			if( errorcount++ == 0 )
				first_error = mainline;
		last_error = mainline;
	}
}

void putnib(int value)
{
	long	offset;
	int	shift;
	const	char	hex_data[] = "0123456789ABCDEF";

	value &= 0x0F;
	if(hexcode) {
		if(fputc(hex_data[value], objfile) == EOF)
			error("error writing to %s", objname);
	} else {
		while(out_offset + 512 <= nibble_offset) {
			if(fwrite(out, 256, 1, objfile) != 1)
				error("error writing to %s", objname);
			out_offset += 512;
		}
		if(nibble_offset < out_offset || nibble_offset > 0x100000L)
			if(nibble_offset >= 0x100000L)
				error("too many nibbles in output file %s", objname);
			else
				error("can't back up %s", objname);
		else {
			offset = nibble_offset++ - out_offset;
			if(codeonly)
				shift = (offset & 0x01) ? 4 : 0;
			else
				shift = (offset & 0x01) ? 0 : 4;
			offset >>= 1;
			/* hm? why long offset for char var of length 256 ?? */
			out[(int)offset] &= 0x0F << (4 - shift);
			out[(int)offset] += value << shift;
		}
	}
}

void flushnibs(void)
{
	if(!codeonly) {
		while((nibble_offset % 512L) != 0L)
			putnib(0);
	} else {
		if(hexcode)
			return;
		if((nibble_offset % 2L) != 0L)
			putnib(0);
	}
	if(nibble_offset > out_offset)
		if(fwrite(out, (int)((nibble_offset - out_offset) / 2L), 1, objfile) != 1)
			error("error writing to %s", objname);
}

void writeheader(void)
{
	char	buffer[CHAR_BUF_LEN + 1];

	if(!codeonly) {
		strncpy(headerout.saturn.id, "Saturn", 6);
		set_16((unsigned char *)headerout.saturn.length,
			(unsigned int)(((get_u32((unsigned char *)headerout.saturn.nibbles) + 511) / 512) +
			((ext_symbols + ext_fillrefs + (O_SYM_CNT - 1)) / O_SYM_CNT) + 1));
		set_32((unsigned char *)headerout.saturn.nibbles, (unsigned long)(pc + pc_offset - codestart));
		set_16((unsigned char *)headerout.saturn.symbols, (unsigned int)ext_symbols);
		set_16((unsigned char *)headerout.saturn.references, (unsigned int)ext_fillrefs);
		set_32((unsigned char *)headerout.saturn.start, (unsigned long)codestart);
		set_8((unsigned char *)headerout.saturn.absolute, (unsigned char)absolute);
		set_8((unsigned char *)headerout.saturn.filler, '\0');
		sprintf(buffer, "%-26s", date);
		strncpy(headerout.saturn.date, buffer, 26);
		sprintf(buffer, "%-40s", title);
		strncpy(headerout.saturn.title, buffer, 40);
		strncpy(headerout.saturn.softkeys, "                    ", 20);
/* check creation of next string */
		sprintf(buffer, "SASM %-21s", version);
		strncpy(headerout.saturn.version, buffer, 26);
		set_32((unsigned char *)headerout.saturn.romid, 0L);
		rewind(objfile);
		if(fwrite(headerout.buffer, 256, 1, objfile) != 1)
			error("error writing to %s", objname);
	}
}

void putobject(union SFC *sym_rec)
{
	int	i;

	if(out_position == 0) {
		strncpy(out, "Symbol table", O_SYM_MOD);
		out_position = O_SYM_MOD;
	}
	i = 0;
	do {
		out[out_position + i] = sym_rec->c[i];
	} while(++i < sizeof(union SFC)); /* sizeof(union SFC) == O_SYM_LEN */
	out_position += sizeof(union SFC);
	if((out_position + sizeof(union SFC)) > 256) {
		if(fwrite(out, 256, 1, objfile) != 1)
			error("error writing to %s", objname);
		out_position = 0;
	}
}

void flushobject(void)
{
	if(out_position > O_SYM_MOD)
		if(fwrite(out, 256, 1, objfile) != 1)
			error("error writing to %s", objname);
}

void dosymbol(struct symbol *sym_ptr)
{
	unsigned long	fillcount;
	unsigned char	flags;
	int	i;
	long	value;
	struct o_fillref	fill_ref;
	struct o_symbol	symref;
	struct fillref	*fillptr;

	fillptr = sym_ptr->s_fillref;
	flags = sym_ptr->s_flags;
	if( ( ( flags & SYM_RESOLVED ) &&
	    ( flags & SYM_GLOBAL ) ) ||
	    ( fillptr != NULL ) ) {
		i = 0;
		while(sym_ptr->s_name[i] != '\0' && i < SYM_NAME_LEN) {
			symref.os_name[i] = sym_ptr->s_name[i];
			i++;
		}
		while(i < SYM_NAME_LEN)
			symref.os_name[i++] = ' ';
		fillcount = (long)sym_ptr->s_fillcount & 0x3FFFL;
		value = sym_ptr->s_value;
		if((flags & SYM_RESOLVED) && !(flags & SYM_FILE)) {
			fillcount |= 0x8000L;
			if((flags & SYM_RELATIVE)) {
				fillcount |= 0x4000L;
				value -= codestart;
			}
		}
		set_32((unsigned char *)symref.os_value, (unsigned long)value);
		set_16((unsigned char *)symref.os_fillcount, (unsigned int)fillcount);
		putobject((union SFC *)(&symref));
		ext_symbols++;
		while(fillptr != NULL) {
			set_8((unsigned char *)fill_ref.of_class, (unsigned char)fillptr->f_class);
			set_8((unsigned char *)fill_ref.of_subclass, (unsigned char)fillptr->f_subclass);
			set_32((unsigned char *)fill_ref.of_address, (unsigned long)(fillptr->f_address - codestart));
			set_32((unsigned char *)fill_ref.of_bias, (unsigned long)fillptr->f_bias);
			set_16((unsigned char *)fill_ref.of_length, (unsigned int)fillptr->f_length);
			putobject((union SFC *)(&fill_ref));
			ext_fillrefs++;
			fillptr = fillptr->f_next;
		}
	}
}

void writesymbols(void)
{
	if(!codeonly) {
		allsymbols(dosymbol, s_root->s_right);
		flushobject();
	}
}

void writeopcode(unsigned long opcode_len, char *opcode_nibs)
{
	int	n;

	while(opcode_len-- != 0) {
		if((n = *opcode_nibs++) != 0)
			n = ( strchr( dectab, n ) ) ? n - '0': n - '7';
		putnib( n );
	}
	begin_pseudos_ok = FALSE;
}

int iscomment(void)
{
	if(*ptr == '*' || *ptr == '\0')
		return TRUE;
	return FALSE;
}

int isfield(void)
{
	if(column < ccolumn && *ptr != '\0')
		return TRUE;
	return FALSE;
}

int getfield(char *fieldp, int maxlength, char *separators)
{
	char	*tmp;
	int	length = 0;

	if(isfield() && !strchr(separators, *ptr)) {
		tmp = ptr;
		if((ptr = strpbrk(tmp, separators)) == NULL)
			ptr = linestart + linelength;
		length = ptr - tmp;
		column += length;
		if(length > maxlength) {
			errmsg("Field too long");
			length = maxlength;
		}
		strncpy(fieldp, tmp, length);
	}
	fieldp[length] = '\0';
	return (length != 0) ? TRUE : FALSE; /* cmp ax,1 ; sbb ax,ax ; inc ax */
}

void skipwhite(void)
{
	char	c;

	while((c = *ptr) == ' ' || c == '\t') {
		ptr++;
		if(c == ' ')
			column++;
		else
			column += 8 - ((column - 1) % 8);
	}
}

char *getline(char *lineptr, FILE *file)
{
	char	*pt;

	if(fgets(lineptr, CHAR_BUF_LEN, file) == NULL)
		return NULL;
	if(strlen(lineptr) == (CHAR_BUF_LEN - 1) && lineptr[CHAR_BUF_LEN - 1] != '\n') {
		char	tempbuf[CHAR_BUF_LEN];
		errmsg("Input line too long (extra characters ignored)");
		while((pt = fgets(tempbuf, CHAR_BUF_LEN, file)) != NULL &&
			strlen(tempbuf) == (CHAR_BUF_LEN - 1) &&
			tempbuf[CHAR_BUF_LEN - 1] != '\n');
	}
	if((pt = strchr(lineptr, '\n')) != NULL)
		*pt = '\0';
	if((pt = strrchr(lineptr, '\r')) != NULL && pt[1] == '\0')
		*pt = '\0';
	if(lineptr[0] == 0x1A)
		lineptr[0] = '\0';
	return lineptr;
}

void copyline(char *outline, char *in_line, int length)
{
	static	char	blank[9] = "        ";
	char	c;
	int	col = 0;

	while((c = *in_line++) != '\0' && col < length) {
		if(c != '\t') {
			*outline++ = c;
			col++;
		} else {
			strcpy(outline, (col % 8) + blank);
			outline += 8 - (col % 8);
			col += 8 - (col % 8);
		}
	}
	*outline = '\0';
}

void doheader( void )
{
	char	buffer[CHAR_BUF_LEN + 1];

	if( listing ) {
		lineofpage = 1;
		sprintf( buffer, "%cSaturn Assembler    %-34.34s %s", '\f', title, date );
		putlist( buffer );
		sprintf( buffer, "%-14.14s %-15.15s %-24.24s %-14.14s Page %4d", version, creation, subtitle, inbase, page++ );
		putlist( buffer );
		putlist( "" );
	}
}

void putlist(char	*lineout)
{
	register int	length, oldlen;

	oldlen = length = strlen(lineout) - 1;
	if(++lineofpage > pagelength)
		doheader();
	if(pagewidth != 0 && pagewidth <= length)
		length = pagewidth - 1;
	while(length >= 0 && lineout[length] == ' ')
		length--;
	if(oldlen != length && length >= 0)
		lineout[++length] = '\0';
	if(lineout[0] != '\0') /* stupid MeSsyDOS can't fputs "" !?!? */
		if(fputs(lineout, listfile) < 0)
			error("error writing to %s", listname);
	if(putc('\n', listfile) == EOF)
		error("error writing to %s", listname);
/*	if(fputs(lineout, listfile) < 0 || putc('\n', listfile) == EOF)
		error("error writing to %s", listname); */
}

void putsymbolref(struct symbol *sym_ptr)
{
	int	pwidth, col;
	char	linebuf[CHAR_BUF_LEN + 1];
	struct lineref	*lineptr;

	lineptr = sym_ptr->s_lineref;
	pwidth = (pagewidth != 0) ? pagewidth : 80;
	if(debug || (lineptr != NULL &&
		(!(sym_ptr->s_flags & SYM_RESOLVED) ||
		!(sym_ptr->s_flags & SYM_FILE) || lineptr->l_next != NULL) &&
		(lineptr->l_line != 0 || lineptr->l_next != NULL))) {
	/* should change format of linebuf according to SYM_NAME_LEN */
		sprintf(linebuf, "%c%-12.12s  ", (sym_ptr->s_flags & SYM_GLOBAL) ? '=' : ' ', sym_ptr->s_name);
		if(sym_ptr->s_flags & SYM_RESOLVED)
			sprintf(linebuf + 15, "%3.3s%8ld #%08.8lX -", (sym_ptr->s_flags & SYM_RELATIVE) ? "Rel" : "Abs", sym_ptr->s_value, sym_ptr->s_value);
		else
			sprintf(linebuf + 15, "%s", (sym_ptr->s_flags & SYM_EXTERNAL) ? "Ext                   -" : "    ******* ********* -");
		col = strlen(linebuf);
		while(lineptr != NULL) {
			if((col + 6) > pwidth) {
				putlist(linebuf);
				sprintf(linebuf, "%38s", "");
				col = 38;
			}
			sprintf( linebuf + col, " %5d", (int)lineptr->l_line );
			col += 6;
			lineptr = lineptr->l_next;
		}
		putlist(linebuf);
	}
}

void listsymbols(void)
{
	lineofpage = pagelength;
	strcpy(subtitle, "Symbol Table");
	allsymbols(putsymbolref, s_root->s_right);
	if(debug) {
		lineofpage = pagelength;
		strcpy(subtitle, "Symbols Defined on Command Line");
		allsymbols(putsymbolref, sys_root->s_right);
	}
}

void liststatistics( void )
{
	int	pwidth, len;
	register int	i, j;

	pwidth = ( pagewidth == 0 ) ? 80 : pagewidth;
	strcpy( subtitle, "Statistics" );
	doheader();
	putlist( "Input Parameters\n" );
	sprintf( line, "  Source file name is %s\n", inname );
	putlist( line );
	sprintf( line, "  Listing file name is %s\n", listname );
	putlist( line );
	sprintf( line, "  Object file name is %s\n", objname );
	putlist( line );
	sprintf( line, "  Flags set on command line" );
	len = -1;
	i = 0;
	while( i < FLAGS_NUM ) {
		if( !( flag[i] & SYM_RESOLVED ) ) {
			i++;
			continue;
		}
		j = len + 3 + ( ( i > 9 ) ? 1 : 0 );
		if( j >= pwidth ) {
			line[len++] = ',';
			line[len] = '\0';
			len = -1;
		}
		if( len < 0 ) {
			putlist( line );
			sprintf( line, "    %d", i );
		} else
			sprintf( line + len, ", %d", i );
		len = strlen( line );
		i++;
	}
	putlist( line );
	if( len < 0 )
		putlist( "    None" );
	if( debug ) {
		sprintf(line, "\n  Flags set when assembly completed" );
		len = -1;
		i = 0;
		while( i < FLAGS_NUM ) {
			if( !( flag[i] & SYM_EXTERNAL ) ) {
				i++;
				continue;
			}
			j = len + 3 + ( ( i > 9 ) ? 1 : 0 );
			if( j >= pwidth ) {
				line[len++] = ',';
				line[len] = '\0';
				len = -1;
			}
			if( len < 0 ) {
				putlist( line );
				sprintf( line, "    %d", i );
			} else
				sprintf( line + len, ", %d", i );
			len = strlen( line );
			i++;
		}
		putlist( line );
		if( len < 0 )
			putlist( "    None" );
	}
	putlist( "\nErrors\n" );
	if( errorcount <= 1 )
		if( errorcount == 0 )
			strcpy( line, "  None" );
		else
			sprintf( line, "  1 error (line %ld)", first_error );
	else
		sprintf( line, "  %ld errors (first %ld/last %ld)", errorcount, first_error, last_error );
	putlist( line );
}

void definesymbol(char *symbol)
{
	char	*cp;
	int	c = (SYM_RESOLVED | SYM_DEFINED), len;
	long	value = 1;
	char	symbname[SYM_NAME_LEN + 1];

	if(*symbol == ':')
		symbol++;
	else
		if(*symbol == '=') {
			symbol++;
			c |= SYM_RESOLVED;
		}
	if((cp = strrchr(symbol, '=')) != NULL)
		len = cp - symbol;
	else
		len = strlen(symbol);
	if(len > SYM_NAME_LEN)
		len = SYM_NAME_LEN;
	strncpy(symbname, symbol, len);
	symbname[len] = '\0';
	if(cp != NULL && cp[1] != '\0' && sscanf(cp + 1, "%ld", &value) != 1) {
		fprintf(stderr, "Non-numeric expression for -D (%s)\n", cp + 1);
		errorcount++;
	} else
		sys_addsymbol(symbname, value, c);
}

void get_options( int argc, char **argv )
{
	char	i, *cp, ch;
	register int	j;

/*	while((i = getopt(argc, argv, "Aa:bc:D:d:Eef:GgHhLM:m:Nn:o:P:p:qS:st:vw:x")) != (char)(-1)) */
	while((i = getopt(argc, argv, "Aa:bc:D:d:Eef:GgHhLM:m:Nn:o:P:p:qS:sVvw:x")) != (char)(-1))
		switch( i ) {
		case 'A':
			listing = TRUE;
			setname( &listname, "-" );
			listdefault = FALSE;
			break;
		case 'a':
			listing = TRUE;
			setname(&listname, optarg);
			if((cp = strrchr(listname, '\\')) != NULL)
				cp++;
			else
				cp = listname;
				if(listprefix != NULL) {
					FREE(listprefix);
					DEBUG_PRN("Address %ld: free'd memory\n", listprefix);
				}
				if(*cp != '.' || strchr(cp, '\\'))
					listdefault = FALSE;
				else {
					listdefault = TRUE;
					listsuffix = cp;
					if(listname != cp) {
						if((listprefix = calloc(1, cp - listname + 1)) == NULL)
							error("error getting memory for %s", listname);
						DEBUG_PRN("Address %ld: allocated %d bytes\n", listprefix, cp - listname + 1);
						strncpy(listprefix, listname, cp - listname);
					}
				}
			break;
		case 'b':
			blanks_in_expr = TRUE;
			break;
		case 'c':
			if( sscanf( optarg, "%d", &ccolumn ) != 1 ) {
				fprintf( stderr, "Non-numeric comment column (%s)\n", optarg );
				errorcount++;
			}
			if( ccolumn < 0 ) {
				fprintf( stderr, "Comment column must be non-negative (%d)\n", ccolumn ); /* was '(%ld)' */
				errorcount++;
			}
			if(ccolumn == 0)
				ccolumn = CHAR_BUF_LEN + 1;
			break;
		case 'D':
			definesymbol(optarg);
			break;
		case 'd':
#if	defined(__DEBUG__)
			setname( &dbgname, optarg );
			if( dbgfile != NULL && dbgfile != stderr )
				fclose( dbgfile );
			if( !strcmp( dbgname, "-" ) ) {
				setname( &dbgname, "[stderr]" );
				dbgfile = stderr;
			} else
				if( ( dbgfile = fopen( dbgname, "wt" ) ) == NULL )
					error( "error creating %s for writing", dbgname );
			setbuf( dbgfile, NULL );
			debug = TRUE;
#else
			fprintf( stderr, "Not compiled with DEBUG on !\n" );
			errorcount++;
#endif
			break;
		case 'E':
			errors_to_stderr = TRUE;
			extended_errors = TRUE;
			break;
		case 'e':
			errors_to_stderr = TRUE;
			extended_errors = FALSE;
			break;
		case 'f':
			j = -1;
			while(j > -2)
				switch(*optarg) {
					case ' ':
					case '\0':
					case ',':
					case ':':
					case ';':
						if(j >= 0 && j < FLAGS_NUM) {
							DEBUG_PRN("Setting flag %d\n", j);
							flag[j] = 0x02;
						} else
							if(j >= 0) {
								fprintf(stderr, "Flag number out of range (%d)\n", j);
								errorcount++;
							}
						if(*optarg == '\0')
							j = -2;
						else {
							optarg++;
							j = -1;
						}
						break;
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
					case '8':
					case '9':
						j = (j < 0) ? 0 : j * 10;
						j += *optarg++ - '0';
						break;
					default:
						fprintf(stderr, "Invalid flag digit (%c)\n", *optarg++);
						errorcount++;
						break;
				}
			break;
		case 'G':
			force_global = TRUE;
			break;
		case 'g':
			force_global = FALSE;
			break;
		case 'H':
			codeonly = TRUE;
			hexcode = FALSE;
			break;
		case 'h':
			codeonly = TRUE;
			hexcode = TRUE;
			break;
		case 'L':
			pass1_listing = TRUE;
			break;
		case 'M':
			setname(&msgname, optarg);
			msg_append = TRUE;
			break;
		case 'm':
			setname(&msgname, optarg);
			msg_append = FALSE;
			break;
		case 'N':
			listing = FALSE;
			break;
		case 'n':
			if( sscanf( optarg, "%d", &codewidth ) != 1 ) {
				fprintf( stderr, "Non-numeric code field width (%s)\n", optarg );
				errorcount++;
			}
			if(codewidth <= 0) {
				fprintf(stderr, "Code field width must be greater than 0 (%d invalid)\n", codewidth);
				codewidth = 5;
				errorcount++;
			}
			if(codewidth > 243) {
				fprintf(stderr, "Code field width must be less than %d (%d invalid)\n", 243, codewidth);
				codewidth = 5;
				errorcount++;
			}
			break;
		case 'o':
			setname(&objname, optarg);
			if((cp = strrchr(objname, '\\')) != NULL)
				cp++;
			else
				cp = objname;
			if(objprefix != NULL) {
				FREE(objprefix);
				DEBUG_PRN("Address %ld: free'd memory\n", objprefix);
			}
			if(*cp == '.' && strchr(cp, '\\') == NULL) {
				objdefault = TRUE;
				objsuffix = cp;
				if(objname != cp) {
					if((objprefix = (char *)calloc(1, cp - objname + 1)) == NULL)
						error("error getting memory for %s", objname);
					DEBUG_PRN("Address %ld: allocated %d bytes\n", objprefix, cp - objname + 1);
					strncpy(objprefix, objname, cp - objname);
				}
			} else
				objdefault = FALSE;
			break;
		case 'P':
			ch = *optarg;
			if(ch < '0' || ch > '3' || optarg[1] != '\0')
				fprintf(stderr, "Invalid code generation level (%s)\n", optarg);
			else
				code_level = (ch << 6) - 0x0C00; /* ?? */
			break;
		case 'p':
			if(sscanf(optarg, "%d", &pagelength) != 1) {
				fprintf(stderr, "Non-numeric page length (%s)\n", optarg);
				errorcount++;
			} else if(pagelength < 4) {
				fprintf(stderr, "Page length must be at least 4 (%d)\n", pagelength);
				pagelength = 60;
				errorcount++;
			}
			break;
		case 'q':
			quote_ok = TRUE;
			break;
		case 'S':
			if( strspn( optarg, hextab ) != strlen( optarg ) ||
				sscanf( optarg, "%lx", &codestart ) != 1 ||
				codestart >= 0x100000L || codestart < 0 ) {
				codestart = 0;
				have_start_addr = FALSE;
				errorcount++;
				fprintf( stderr, "Invalid start address (%s)\n", optarg );
			}
			break;
		case 's':
			blanks_in_expr = FALSE;
			ccolumn = 30;
			codeonly = FALSE;
			extended_errors = FALSE;
			quote_ok = FALSE;
			rel_expressions_ok = FALSE;
			linewidth = 5;
			codewidth = 4;
			listing = TRUE;
			pass1_listing = FALSE;
			setname(&listname, "-");
			listdefault = FALSE;
			setname(&msgname, "-");
			setname(&objname, "code");
			objdefault = FALSE;
			verbose = FALSE;
			combos = FALSE;
			break;
/* obsolete with opc.c */
/*		case 't':
			setname(&opname, optarg);
			break; */
		case 'V':
			verbose = TRUE;
			break;
		case 'v':
			usage( TRUE );
			exit( 1 );
			break;
		case 'w':
			if(sscanf(optarg, "%d", &pagewidth) != 1) {
				fprintf(stderr, "Non-numeric page width (%s)\n", optarg);
				errorcount++;
			} else
				if(pagewidth < 44) {
					fprintf(stderr, "Page width must be at least 44 (%d)\n", pagewidth);
					errorcount++;
				}
			break;
		case 'x':
			rel_expressions_ok = TRUE;
			break;
		default:
			errorcount++;
			break;
		}
}

void copy_sys( struct symbol *symb )
{
	struct symbol	*newsymb;
	char	name[SYM_NAME_LEN + 1];

	strncpy( name, symb->s_name, SYM_NAME_LEN );
	name[SYM_NAME_LEN] = '\0';
	newsymb = addsymbol( name, symb->s_value, symb->s_flags );
	addlineref( newsymb );
}

void dopptr( void )
{
	char	*cp;
	int	i;

	if( pptr[imptr][0] == NULL ) {
		cp = imstack[imptr].im_buffer;
		i = 0;
		while( i < 9 && cp != NULL ) {
			if( *cp == '<' && strchr( cp, '>' ) ) {
				pptr[imptr][i++] = cp + 1;
				cp = strchr( cp, '>' );
				*cp++ = '\0';
				if( ( cp = strchr( cp, ',' ) ) != NULL )
					cp++;
				continue;
			}
			pptr[imptr][i] = cp;
			if( ( cp = strchr( cp, ',' ) ) != NULL ) {
				*cp++ = '\0';
				i++;
				continue;
			}
			if( ( cp = strpbrk( pptr[imptr][i], " \t" ) ) != NULL )
				*cp = '\0';
			cp = NULL;
			i++;
		}
		while( i < 9 )
			pptr[imptr][i++] = NULL;
	}
}

void get_loopline( void )
{
	char	c, *cp;
	int	space;

	ptr = strncpy( tmpline, line, CHAR_BUF_LEN );
	cp = line;
	memset( line, 0, CHAR_BUF_LEN + 1 );
	DEBUG_PRN( "Loop line: %s\n", ptr );
	if( *ptr == '!' )
		ptr++;
	while( ( c = *ptr++ ) != '\0' &&
		( space = ( CHAR_BUF_LEN - 6 ) - ( cp - line ) ) > 0 ) {
		if( c != '$' ) {
			*cp++ = c;
			continue;
		}
		c = *ptr++;
		switch( c ) {
			case '$':
				*cp = '$';
				break;
			case '<':
				if( space >= strlen( curr_name ) )
					strcpy( cp, curr_name );
				else
					strncpy( cp, curr_name, space );
				break;
			case '0':
				sprintf( cp, "%ld", currentline );
				break;
			case '1':
/*
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
 */
				if( fnentry->lp_value < 0 )
					sprintf( cp, "(0)-(%ld)", -fnentry->lp_value );
				else
					sprintf( cp, "%ld", fnentry->lp_value );
				break;

			case 'E':
			case 'e':
				sprintf( cp, "%ld", fnentry->lp_end );
				break;

			case 'S':
			case 's':
				sprintf( cp, "%ld", fnentry->lp_step );
				break;

			case '(':
				if( strchr( dectab, *ptr ) && *ptr != '0' ) {
					char	*save_ptr, fmtchar, fmtstring[10];
					char	char_in[] = "HXDOUChxdouc", char_out[] = "XXdoucxxdouc";
					int	digit, fmtvalue;

					if( ( digit = *ptr++ - '1' ) > LOOP_STACK_MAX ) {
						strcpy( cp, "$(" );
						break;
					}
					fmtvalue = -1;
					fmtchar = 'd';
					while( ( c = *ptr++ ) != '\0' && c != ')' ) {
						if( strchr( dectab, c ) ) {
							if( fmtvalue == -1 )
								fmtvalue = 0;
							fmtvalue = ( c + fmtvalue * 10 ) - '0';
							continue;
						}
						if( ( save_ptr = strchr( char_in, c ) ) != NULL ) {
							fmtchar = char_out[save_ptr - char_in];
							continue;
						}
						if( c != ':' || fmtvalue != -1 )
							errmsg( "Invalid format character" );
					}
					if( c == '\0' ) {
						errmsg( "Invalid parameter indirection" );
						break;
					}
					if( fmtvalue <= 0 )
						fmtvalue = 1;
					if( space < fmtvalue ) {
						fmtvalue = space;
						errmsg( "Parameter too big (truncated)" );
					}
					sprintf( fmtstring, "%%0%dl%c", fmtvalue, fmtchar );
					sprintf( cp, fmtstring, loops[digit][0] );
				} else
					strcpy( cp, "$(" );
				break;
			default:
				*cp++ = '$';
				*cp = c;
				break;
		} /* switch */
		cp += strlen( cp );
	} /* while */
	if( c != '\0' )
		errmsg( "Expanded loop line too long (truncated)" );
	*cp = '\0';
	DEBUG_PRN( "Substitute: %s\n", line );
	ptr = line;
}

void get_macroline( void )
{
	char	c, *mptr, *cp;
	int	space;

	if( macroptr != NULL ) {
		ptr = macroptr->m_line;
		macroptr = macroptr->m_next;
		mptr = line;
		memset( line, 0, CHAR_BUF_LEN + 1 );
		DEBUG_PRN( "Macro line: %s\n", ptr );
		if( *ptr == '!' )
			ptr++;
		while( ( c = *ptr++ ) != '\0' &&
			( space = ( CHAR_BUF_LEN - 6 ) - ( mptr - line ) ) > 0 ) {
			if( c != '$' ) {
				*mptr++ = c;
				continue;
			}
			c = *ptr++;
			switch( c ) {
				case '$':
					*mptr = '$';
					break;
				case '<':
					if( space >= strlen( curr_name ) )
						strcpy( mptr, curr_name );
					else
						strncpy( mptr, curr_name, space );
					break;
				case '0':
					sprintf( mptr, "%ld", currentline );
					break;
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					cp = pptr[imptr][c - '1'];
					if( cp != NULL ) {
						if( space >= strlen( cp ) )
							strcpy( mptr, cp );
						else
							strncpy( mptr, cp, space );
					}
					break;

				case '(':
					if( strchr( dectab, *ptr ) && *ptr != '0' ) {
						char	*save_ptr, fmtchar, fmtstring[10];
						char	char_in[] = "HXDOUhxdou", char_out[] = "XXdouxxdou";
						int	digit, fmtvalue;
						struct expression	*exprp;

						digit = *ptr++ - '1';
						fmtvalue = -1;
						fmtchar = 'd';
						exprp = NULL;
						while( ( c = *ptr++ ) != '\0' && c != ')' ) {
							if( strchr( dectab, c ) ) {
								if( fmtvalue == -1 )
									fmtvalue = 0;
								fmtvalue = ( c + fmtvalue * 10 ) - '0';
								continue;
							}
							if( ( save_ptr = strchr( char_in, c ) ) != NULL ) {
								fmtchar = char_out[save_ptr - char_in];
								continue;
							}
							if( c != ':' || fmtvalue != -1 )
								errmsg( "Invalid format character" );
						}
						if( c == '\0' ) {
							errmsg( "Invalid parameter indirection" );
							break;
						}
						if( fmtvalue <= 0 )
							fmtvalue = 1;
						if( space < fmtvalue ) {
							fmtvalue = space;
							errmsg( "Parameter too big (truncated)" );
						}
						sprintf( fmtstring, "%%0%dl%c", fmtvalue, fmtchar );
						if( ( cp = pptr[imptr][digit] ) != NULL ) {
							save_ptr = ptr;
							ptr = cp;
							skipwhite();
							column = 1;
							exprp = doexpr();
							ptr = save_ptr;
						}
						sprintf( mptr, fmtstring, ( exprp != NULL ) ? exprp->e_value : 0L );
					} else
						strcpy( mptr, "$(" );
					break;
				default:
					*mptr++ = '$';
					*mptr = c;
					break;
			} /* switch */
			mptr += strlen( mptr );
		} /* while */
		if( c != '\0' )
			errmsg( "Expanded macro line too long (truncated)" );
		*mptr = '\0';
		DEBUG_PRN( "Substitute: %s\n", line );
		ptr = line;
	} else
		ptr = NULL;
}

void imget(void)
{
	struct symbol	*symbptr;
	char	string[CHAR_BUF_LEN];

	do {
		if(impending == IMP_INCLUDE || impending == IMP_MACRO) {
			if(imtype == IMTYPE_INCLUDE) {
				imstack[imptr].im_addr = ftell(include);
				fclose(include);
			} else
				if(imtype == IMTYPE_MACRO)
					imstack[imptr].im_filename = (char *)macroptr;
			if(++imptr < MACRO_STACK_MAX) {
				imstack[imptr].im_type = impending;
				pptr[imptr][0] = NULL;
				imstack[imptr].im_flagptr = f_ptr;
				imstack[imptr].im_listflags = listflags;
				if(impending == IMP_INCLUDE) {
					imstack[imptr].im_line = currentline;
					currentline = 0;
					imstack[imptr].im_addr = 0;
					imstack[imptr].im_filename = curr_name;
					strcpy(imstack[imptr].im_buffer, symname);
					curr_name = imstack[imptr].im_buffer;
				} else {
					imstack[imptr].im_filename = (char *)entry->op_macroptr;
					strcpy(imstack[imptr].im_buffer, ptr);
				}
				impending = IMP_INSERT;
			} else {
				errmsg("INCLUDE or MACRO nested too deeply");
				imptr--;
				impending = IMP_NONE;
			}
		}
		if((impending == IMP_END || impending == IMP_EXITM) && imtype == IMTYPE_CODE) {
			errmsg("ENDM and EXITM not permitted outside of a macro");
			impending = IMP_NONE;
		}
		if(impending == IMP_END) {
			if(imtype == IMTYPE_INCLUDE) {
				if(include != NULL)
					fclose(include);
				currentline = imstack[imptr].im_line;
				curr_name = imstack[imptr].im_filename;
			}
			if(imtype == IMTYPE_MACRO)
				while(imptr > -1 && imstack[imptr].im_type == -1)
					imptr--;
			if(imptr == -1)
				reading = FALSE;
		}
		if(impending == IMP_EXITM && imtype == IMTYPE_INCLUDE) {
			errmsg("ENDM and EXITM not permitted outside of a macro");
			impending = IMP_NONE;
		}
		if(impending == IMP_END || impending == IMP_EXITM) {
			f_ptr = imstack[imptr].im_flagptr;
			listflags &= LIST_MASK2;
			listflags |= imstack[imptr].im_listflags & LIST_MASK;
			symbptr = addsymbol("*LISTFLAGS", 0L, SYM_RESOLVED);
			symbptr->s_value = listflags & LIST_MASK;
			if(imptr > -1)
				imptr--;
			impending = IMP_INSERT;
		}
		if(impending == IMP_INSERT) {
			impending = IMP_NONE;
			assembling = TRUE;
			if(imptr > -1)
				imtype = imstack[imptr].im_type;
			else
				imtype = IMTYPE_CODE;
			if(imtype == IMTYPE_INCLUDE) {
				if((include = openfile(imstack[imptr].im_buffer, "SASM_INCLUDE", ";\\lib\\sasm", "rb", NULL)) == NULL) {
					currentline = imstack[imptr].im_line;
					curr_name = imstack[imptr].im_filename;
					sprintf(string, "Unable to open INCLUDE file (%s)", imstack[imptr].im_buffer);
					errmsg(string);
					impending = IMP_END;
				} else {
					if(fseek(include, imstack[imptr].im_addr, SEEK_SET) != 0)
						error("error setting location in include file %s", nameptr);
					FREE(nameptr);
					DEBUG_PRN("Address %ld: free'd memory\n", nameptr);
				}
			}
			if(imtype == IMTYPE_MACRO) {
				macroptr = (struct macro *)imstack[imptr].im_filename;
				dopptr();
			}
		}
		if(impending == IMP_NONE) {
			if(imtype == IMTYPE_INCLUDE) {
				if((ptr = getline(line, include)) != NULL) {
					currentline++;
					DEBUG_PRN("Include line: %s\n", ptr);
				} else {
					impending = IMP_END;
					if(addingmacro) {
						errmsg("End of file while defining macro");
						addingmacro = FALSE;
					}
				}
			}
			if(imtype == IMTYPE_MACRO) {
				get_macroline();
				if(ptr == NULL)
					impending = IMP_EXITM;
			}
		}
	} while(impending != IMTYPE_CODE);
}

char *makename(char *base, char *prefix, char *suffix)
{
	char	*destname, *cp;
	unsigned int	namlen;

	namlen = strlen(base) + strlen(suffix) + 1;
	if(prefix != NULL)
		namlen += strlen(prefix);
	if((destname = calloc(1, namlen)) == NULL)
		error("error getting memory for %s in makename", base);
	DEBUG_PRN("Address %ld: allocated %d bytes\n", destname, namlen);
	if(prefix != NULL && *prefix != '\0') {
		strcpy(destname, prefix);
		if((cp = strrchr(base, '\\')) != NULL)
			base = cp + 1;
	}
	strcat(destname, base);
	if((cp = strrchr(destname, '\\')) != NULL)
		cp++;
	else
		cp = destname;
	if((cp = strrchr(cp, '.')) != NULL)
		*cp = '\0';
	strcat(destname, suffix);
	return destname;
}

void setup_formats(void)
{
	if(listing) {
		source_col = linewidth + codewidth + 8;
		sprintf(basefmt, "%%%dld", linewidth);
		memset(basefmt + 4, ' ', codewidth + 8);
		basefmt[codewidth + 12] = '\0';
		sprintf(addrfmt, "%%05lX %%-%d.%ds ", codewidth, codewidth);
		sprintf(inclfmt, "%*s+", linewidth, "");
		sprintf(codefmt, "%*s%%-%d.%ds", linewidth + 7, "", codewidth, codewidth);
		sprintf(lastcodefmt, "%*s%%s", linewidth + 7, "");
	}
}

void main( int argc, char **argv, char **env )
{
	char *current_date;
	struct symbol *symbptr;
	struct tm *time_struct;
	char *cp;
	char listline[MAXLISTLINE];
	char *opptr = NULL;
	time_t clock;
	unsigned long oplen;
	int i;
	struct OPCODE *opcode;

	progname = argv[0];
	listfile = stdout;
	sys_initialize();
/* obsolete with opc.c */
/*	setname( &opname, "sasm.opc" ); */
	setname( &msgname, "-" );
	if( ( cp = getenv( "SASM_GLOBAL" ) ) != NULL )
		force_global = ( *cp == '1' ? TRUE : FALSE );
	get_options( argc, argv );
	if(strcmp(msgname, "-")) {
		if(!freopen(msgname, ((msg_append == FALSE) ? "wt" : "at"), stderr))
			error("error creating %s for writing", msgname);
		setbuf(stderr, 0);
	}
	if( verbose ) {
		fprintf( stderr, "Saturn Assembler %s\n", version );
	}
	if(errorcount > 0) {
		usage( FALSE );
		exit(-3);
	}
	readoptable();
	setup_formats();
	if(((argc - optind) > 1) && (!objdefault || listdefault)) {
		fprintf(stderr, "Options -A, -a, and -o not allowed with multiple files\n");
		usage( FALSE );
		exit(-4);
	}
	singlefile = ((argc - optind) > 1) ? FALSE : TRUE;
	while(singlefile || optind < argc) {
		s_initialize();
		cleanmacros();
		stdinread = FALSE;
		if(singlefile) {
			if(argc == optind || !strcmp("-", argv[optind])) {
				setname(&inname, "[stdin]");
				infile = stdin;
				stdinread = TRUE;
				if((temp = tmpfile()) == NULL)
					error("%s", "error creating a temporary file");
				optind++;
			}
			singlefile = FALSE;
		}
		if(!stdinread) {
			setname(&inname, argv[optind]);
			DEBUG_PRN("input file name is %s\n", inname);
				if((infile = freopen(inname, "rt", stdin)) == NULL) {
					FREE(inname);
					DEBUG_PRN("Address %ld: free'd memory\n", inname);
					inname = makename(argv[optind], "", ".a");
					DEBUG_PRN("Try input file name with .a appended (%s)\n", inname);
					if((infile = freopen(inname, "rt", stdin)) == NULL)
						error("error opening %s for reading", argv[optind]);
				}
				optind++;
				DEBUG_PRN("input file %s opened successfully\n", inname);
		}
		if(listing) {
			if(listdefault)
				if(objdefault && !stdinread) {
					cp = listname;
					listname = makename(inname, listprefix, listsuffix);
					if(cp != NULL) {
						FREE(cp);
						DEBUG_PRN("Address %ld: free'd memory\n", cp);
					}
				} else
					setname(&listname, "-");
			if(!strcmp(listname, "-")) {
				setname(&listname, "[stdout]");
				listfile = stdout;
			} else {
				if((listfile = fopen(listname, "wt")) == NULL)
					error("error creating %s for writing", listname);
				setbuf(listfile, 0);
			}
		}
		if(objdefault)
			if(!stdinread) {
				cp = objname;
				objname = makename(inname, objprefix, objsuffix);
				if(cp != NULL) {
					FREE(cp);
					DEBUG_PRN("Address %ld: free'd memory\n", cp);
				}
			} else
				setname(&objname, "code");
		currentline = 0;
		allsymbols(copy_sys, sys_root->s_right);
		if(code_level >= (CPU_SWITCH2 | CPU_SWITCH1))
			addsymbol("*LEVEL3", 1L, SYM_DEFINED | SYM_RESOLVED);
		if(code_level >= CPU_SWITCH2) {
			addsymbol("*LEVEL2", 1L, SYM_DEFINED | SYM_RESOLVED);
			addsymbol("*1LR2", 1L, SYM_DEFINED | SYM_RESOLVED);
		}
		if(code_level >= CPU_SWITCH1) {
			addsymbol("*LEVEL1", 1L, SYM_DEFINED | SYM_RESOLVED);
			addsymbol("*1LK7", 1L, SYM_DEFINED | SYM_RESOLVED);
		}
		if(code_level >= 0) {
			addsymbol("*LEVEL0", 1L, SYM_DEFINED | SYM_RESOLVED);
			addsymbol("*1LF2", 1L, SYM_DEFINED | SYM_RESOLVED);
		}
		addsymbol("*LEVEL", code_level >> 6, SYM_DEFINED | SYM_RESOLVED);
		if( time( &clock ) == -1 )
			error( "%s", "unable to read current time" );
		current_date = ctime( &clock );
		strncpy(date, current_date, 24);
		date[24] = '\0';
		time_struct = localtime(&clock);
		i = SYM_DEFINED | SYM_RESOLVED;
		addsymbol("*TM_SECOND", (long)(time_struct->tm_sec), SYM_DEFINED | SYM_RESOLVED);
		addsymbol("*TM_MINUTE", (long)(time_struct->tm_min), i);
		addsymbol("*TM_HOUR", (long)(time_struct->tm_hour), i);
		addsymbol("*TM_DAYMONTH", (long)(time_struct->tm_mday), i);
		addsymbol("*TM_MONTH", (long)(time_struct->tm_mon), i);
		addsymbol("*TM_YEAR", (long)(time_struct->tm_year), i);
		addsymbol("*TM_DAYWEEK", (long)(time_struct->tm_wday), i);
		addsymbol("*TM_DAYYEAR", (long)(time_struct->tm_yday), i);
		if((inbase = strrchr(inname, '\\')) != NULL)
			inbase++;
		else
			inbase = inname;
		if((objbase = strrchr(objname, '\\')) != NULL)
			objbase++;
		else
			objbase = objname;
		prog_symbol[0] = '*';
		strncpy(prog_symbol + 1, objbase, 11);
		pass1 = TRUE;
		pass2 = FALSE;
		absolute = FALSE;
		ext_fillrefs = 0;
		ext_symbols = 0;
		title[0] = '\0';
		subtitle[0] = '\0';
		addsymbol("*START", codestart, SYM_DEFINED | SYM_RELATIVE | SYM_RESOLVED);
		if(verbose) {
			if(fprintf(stderr, "Assembling %s\n    object file = %s", inname, objname) < 0)
				error("%s", "error writing to stderr");
			if(listing) {
				if(fprintf(stderr, ", list file = %s\n", listname) < 0)
					error("%s", "error writing to stderr");
			} else {
				if(fprintf(stderr, " (No listing)\n") < 0)
					error("%s", "error writing to stderr");
			}
		}
		while(pass1 || pass2) {
			DEBUG_PRN("Starting pass %d\n", pass1 + 2 * pass2);
			i = 0;
			while(i < FLAGS_NUM) {
				flag[i] = (flag[i] & SYM_RESOLVED) ?
					flag[i] | SYM_EXTERNAL : flag[i] & ~SYM_EXTERNAL;
				i++; /* hm! maybe I could rewrite this */
			}
			f_ptr = -1;
			carry = 3;
			curr_name = inname;
			begin_pseudos_ok = TRUE;
			imptr = -1;
			imtype = IMTYPE_CODE;
			impending = IMP_NONE;
			listflags = (LIST_PSEUDO | LIST_CODE);
			listcount = 0;
			symbptr = addsymbol("*LISTFLAGS", 0L, SYM_RESOLVED);
			symbptr->s_value = (long)(listflags & LIST_MASK);
			reading = TRUE;
			assembling = TRUE;
			addingmacro = FALSE;
			subtitle[0] = '\0';
			lineofpage = pagelength;
			page = 1;
			label_index_m2 = label_index_m1 = 0;
			label_index_p2 = label_index_p1 = 0;
			abase = 0L;
			if(pass2) {
				if(!strcmp(objname, "-")) {
					if(!codeonly)
						error("%s", "object file cannot be [stdout]");
					if(listing && listfile == stdout) {
						fprintf(stderr, "List and code files both [stdout] (list disabled)\n");
						listing = FALSE;
					}
					setname(&objname, "[stdout]");
					objfile = stdout;
					DEBUG_PRN("Object file is [stdout]\n");
				} else {
					if((objfile = fopen(objname, ((hexcode) ? "wt" : "wb"))) == NULL)
						error("error opening %s for writing", objname);
					if(codeonly) {
						DEBUG_PRN("Opened object file (code only): %s\n", objname);
					} else {
						DEBUG_PRN("Opened object file %s; writing header now\n", objname);
						writeheader();
						DEBUG_PRN("Done writing object file header\n");
					}
				}
				nibble_offset = 0;
				out_offset = 0;
				out_position = 0;
				if(codeonly)
					if(hexcode) {
						DEBUG_PRN("Object code type: %s\n", "Text");
					}
					else
						DEBUG_PRN("Object code type: %s\n", "Binary");
			}
			errorcount = 0;
			last_error = 0;
			first_error = 0;
			pc = codestart;
			pc_offset = 0;
			currentline = 0;
			mainline = 0;
			i = 0;
			while( i < 256 ) {
				charset[i] = i;
				i++;
			}
			while( reading ) {
				if( impending != IMP_NONE || imtype != IMTYPE_CODE )
					imget();
				if( !reading )
					break;
				if( imtype == IMTYPE_CODE ) {
					if( ( ptr = getline( line, infile ) ) == NULL ) {
						if( addingmacro ) {
							errmsg( "End of file while defining macro" );
							addingmacro = FALSE;
						}
						break;
					}
					mainline = ++currentline;
					if( stdinread )
						if( fputs( line, temp ) == -1 || fputc( '\n', temp ) == -1 )
							error( "error writing to %s", "[intermediate file]" );
				}
				linestart = ptr;
				column = 1;
				oplen = 0;
				if( insideloop )
					get_loopline();
				if( pass2 || pass1_listing ) {
					opptr = "";
					mnemonic[0] = '\0';
					label[0] = '\0';
					listflags &= ~LIST_DISABLELINE;
					if(imtype == IMTYPE_CODE && (listflags & LIST_CODE))
						listflags |= LIST_ON;
					else
						if(imtype == IMTYPE_MACRO && (listflags & LIST_MACROS))
							listflags |= LIST_ON;
						else
							if(imtype == IMTYPE_INCLUDE && (listflags & LIST_INCLUDES))
								listflags |= LIST_ON;
							else
								listflags &= ~LIST_ON;
					listthisline = (assembling) ? ((listflags & LIST_ON) ? TRUE : FALSE) : FALSE;
					listaddress = FALSE;
				}
				if(!iscomment()) {
					linelength = strlen(linestart);
					if(*ptr == ' ') {
						ptr++;
						column++;
					}
					if(getfield(label, CHAR_BUF_LEN, " \t"))
						listaddress = TRUE;
					skipwhite();
					if(getfield(mnemonic, CHAR_BUF_LEN, " \t") || addingmacro) {
						DEBUG_PRN("Got mnemonic: %s\n", mnemonic);
						mnemonic[SYM_NAME_LEN] = '\0';
						skipwhite();
						if(pass1 && !pass1_listing)
							oplen = opcodelen(mnemonic);
						else {
							listaddress = TRUE;
							opcode = getopcode(mnemonic);
							oplen = opcode->length;
							opptr = opcode->value;
							if(pass2 && oplen != 0)
								writeopcode(oplen, opptr);
						}
					}
					if(label[0] != '\0' && assembling && !addingmacro) {
						DEBUG_PRN("Adding a new label: %s\n", label);
						addlabel(label, pc, (force_global ? SYM_GLOBAL : 0) | (absolute ? 0 : SYM_RELATIVE) | SYM_RESOLVED);
						if(mnemonic[0] == '\0') {
							DEBUG_PRN("Setting carry bits (stand-alone label)\n");
							carry = 3; /* could be FLAG0x?? */
						}
					}
				}
				if(pass2 || pass1_listing) {
					if(listflags & LIST_DISABLELINE)
						listthisline = FALSE;
					else
						if(listflags & LIST_COUNT_ON)
							if(--listcount < 0)
								listthisline = TRUE;
							else {
								listcount = 0;
								listflags &= ~LIST_COUNT_ON;
							}
					if( listing && listthisline ) {
						if( ( currentline > 9999 && linewidth < 5 ) ||
							( currentline > 99999L && linewidth < 6 ) ) {
							linewidth++;
							setup_formats();
						}
						sprintf(listline, basefmt, currentline);
						if(listaddress)
							sprintf(listline + linewidth + 1, addrfmt, pc, opptr);
						listline[source_col - 1] = ' ';
						if(imtype != IMTYPE_CODE)
							if(imtype == IMTYPE_INCLUDE)
								listline[linewidth] = '-';
							else
								strncpy(listline, inclfmt, linewidth + 1);
						copyline(listline + source_col, linestart, (MAXLISTLINE - 1) - source_col);
						putlist(listline);
						if(strlen(opptr) > codewidth) {
							opptr += codewidth;
							if(strlen(opptr) > codewidth) {
								sprintf(listline, codefmt, opptr);
								putlist(listline);
								opptr += codewidth;
							}
							sprintf(listline, lastcodefmt, opptr);
							putlist(listline);
							if(strlen(opptr) > (codewidth + 1))
								putlist("");
						}
					}
					if(listcount != 0)
						listflags |= LIST_COUNT_ON;
				}
				pc += oplen;
				if(pc >= 0x100000L) {
					errmsg("PC wrapped around to 00000");
					pc -= 0x100000L;
					pc_offset += 0x100000L;
				}
			}
			if(ferror(infile))
				error("error reading from %s", inname);
			if(stdinread) {
				stdinread = FALSE;
				fclose(infile);
				infile = temp;
			}
			if(pass1) {
				DEBUG_PRN("Rewinding input file: %s\n", inname);
				rewind(infile);
				pass1 = FALSE;
				pass2 = TRUE;
				addsymbol("*END", pc, (absolute ? 0 : SYM_RELATIVE) | SYM_RESOLVED | SYM_DEFINED);
			} else
				pass2 = FALSE;
		}
		DEBUG_PRN( "Flushing nibbles remaining in object file buffer\n" );
		flushnibs();
		DEBUG_PRN( "Writing symbol table\n" );
		writesymbols();
		DEBUG_PRN( "Rewriting object file header\n" );
		writeheader();
		DEBUG_PRN( "Done rewriting object file header\n" );
		if( listing ) {
			listsymbols();
			liststatistics();
		}
		cum_errors = ( errorcount ) ? -1 : 0;
		if( infile != stdin )
			fclose( infile );
		if( objfile != stdout )
			fclose( objfile );
		if( listing && listfile != stdout )
			fclose( listfile );
	}
	exit( cum_errors );
}
