/*
 * symbols.c - SATURN assembler, symbol parser
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#ifndef	macintosh
# include <malloc.h>
#endif

#include "opc.h"
#include "symbols.h"

struct symbol	s_terminal, *f, *g, *gg, *sys_root, *s_root;

static struct expression	temp;

extern int debug, pass1, pass2;
extern int label_index_m2, label_index_m1, label_index_p1, label_index_p2;
extern long pc, currentline, codestart;
extern char *dbgname;
extern FILE *dbgfile;
extern struct optable *entry;
extern char prog_symbol[];

struct symbol	*newsymbol(void)
{
	struct symbol *dummy;

	if((dummy = (struct symbol *)calloc(1, sizeof(struct symbol))) == NULL)
		error("error getting memory for %s", "symbol");
	DEBUG_PRN("Address %ld: allocated %d bytes\n", dummy, sizeof(struct symbol));
	dummy->s_left = &s_terminal;
	dummy->s_right = &s_terminal;
	return dummy;
}

struct lineref	*newlineref(void)
{
	struct lineref *dummy;

	if((dummy = (struct lineref *)calloc(1, sizeof(struct lineref))) == NULL)
		error("error getting memory for %s", "line reference");
	DEBUG_PRN("Address %ld: allocated %d bytes\n", dummy, sizeof(struct lineref));
	return dummy;
}

struct fillref	*newfillref(void)
{
	struct fillref	*dummy;

	if((dummy = (struct fillref *)calloc(1, sizeof(struct fillref))) == NULL)
		error("error getting memory for %s", "fill reference");
	DEBUG_PRN("Address %ld: allocated %d bytes\n", dummy, sizeof(struct fillref));
	return dummy;
}

struct symbol	*s_rotate(char *v, struct symbol *y)
{
	int	flag;
	struct symbol	*s, *gs;

	flag = (strncmp(v, y->s_name, SYM_NAME_LEN) < 0) ? TRUE : FALSE;
	s = (flag) ? y->s_left : y->s_right;
	if(strncmp(v, s->s_name, SYM_NAME_LEN) < 0) {
		gs = s->s_left;
		s->s_left = gs->s_right;
		gs->s_right = s;
	} else {
		gs = s->s_right;
		s->s_right = gs->s_left;
		gs->s_left = s;
	}
	if(flag)
		y->s_left = gs;
	else
		y->s_right = gs;
	return gs;
}

struct symbol	*s_split(struct symbol *root, char *sym, struct symbol *dummy)
{
	dummy->s_red = (char)1;
	dummy->s_left->s_red = '\0';
	dummy->s_right->s_red = '\0';
	if(f->s_red != '\0') {
		g->s_red = (char)1;
		if(((strncmp(sym, f->s_name, SYM_NAME_LEN) < 0) ? TRUE : FALSE) ==
			((strncmp(sym, g->s_name, SYM_NAME_LEN) < 0) ? TRUE : FALSE))
			f = s_rotate(sym, g);
		dummy = s_rotate(sym, gg);
		dummy->s_red = '\0';
	}
	root->s_right->s_red = '\0';
	return dummy;
}

struct symbol	*tree_addsymbol(struct symbol *root, char *sym,
	long value, int flags)
{
	int c;
	char *cp;
	struct symbol *x;

	if(*sym == ':')
		sym++;
	else
		if(*sym == '=') {
			sym++;
			flags |= SYM_EXTERNAL;
		}
	if((cp = strchr(sym, ' ')) != NULL)
		if((cp - sym) < SYM_NAME_LEN)
			*cp = '\0';
	strncpy(s_terminal.s_name, sym, SYM_NAME_LEN);
	gg = g = f = x = root;
	while((c = strncmp(sym, x->s_name, SYM_NAME_LEN)) != 0) {
		gg = g;
		g = f;
		f = x;
		x = (c < 0) ? x->s_left : x->s_right;
		if((x->s_left)->s_red != '\0' && (x->s_right)->s_red != '\0')
			x = s_split(root, sym, x);
	}
	if(x == &s_terminal) {
		x = newsymbol();
		strncpy(x->s_name, sym, SYM_NAME_LEN);
		x->s_value = value;
		x->s_flags = flags;
		if(strncmp(sym, f->s_name, SYM_NAME_LEN) < 0)
			f->s_left = x;
		else
			f->s_right = x;
		s_split(root, sym, x);
	}
	return x;
}

struct symbol	*findsymbol(char *name)
{
	int	c;
	char	*cp;
	struct symbol	*dummy;

	if(*name == '=' || *name == ':')
		name++;
	if((cp = strchr(name, ' ')) != NULL && (cp - name) < SYM_NAME_LEN)
		*cp = '\0';
	strncpy(s_terminal.s_name, name, SYM_NAME_LEN);
	dummy = s_root;
	while((c = strncmp(name, dummy->s_name, SYM_NAME_LEN)) != 0)
		dummy = (c < 0) ? dummy->s_left : dummy->s_right;
	if(dummy == &s_terminal) {
		dummy = sys_root;
		while((c = strncmp(name, dummy->s_name, SYM_NAME_LEN)) != 0)
			dummy = (c < 0) ? dummy->s_left : dummy->s_right;
		if(dummy == &s_terminal)
			return NULL;
	}
	return dummy;
}

struct symbol	*addsymbol(char *sym, long value, int flags)
{
	return tree_addsymbol(s_root, sym, value, flags);
}

struct symbol	*sys_addsymbol(char *sym, long value, int flags)
{
	return tree_addsymbol(sys_root, sym, value, flags);
}

void	addlineref(struct symbol *sym)
{
	struct lineref **lp;

	lp = &sym->s_lineref;
	while(*lp != NULL)
		lp = &(*lp)->l_next;
	*lp = newlineref();
	(*lp)->l_line = currentline;
}

void	addfillref(struct expression *expr)
{
	int	type = SYM_EXTERNAL;
	long	value = 0;
	struct symbol	*sym;
	struct fillref	**fp;

	if(pass2 && (expr->e_extcount != 0 || expr->e_relcount != 0)) {
		if(expr->e_relcount != 0 && entry->op_class != CLASS_LABEL) {
			temp = *expr;
			strncpy(temp.e_name, prog_symbol, SYM_NAME_LEN);
			expr = &temp;
			type = (SYM_RELATIVE | SYM_GLOBAL | SYM_RESOLVED);
			value = codestart;
		}
		if((sym = findsymbol(expr->e_name)) == NULL)
			sym = addsymbol(expr->e_name, value, type);
		fp = &sym->s_fillref;
		sym->s_fillcount++;
		while(*fp != NULL)
			fp = &(*fp)->f_next;
		*fp = newfillref();
		(*fp)->f_address = pc + entry->op_offset;
		(*fp)->f_bias = (sym->s_flags | SYM_RESOLVED) ? expr->e_value - sym->s_value : expr->e_value;
		(*fp)->f_class = entry->op_class;
		(*fp)->f_subclass = entry->op_subclass;
		(*fp)->f_length = (int)entry->op_size;
	}
}

/* Added conversion of offset labels to normal labels:
 *      "--" --> "lBm2%X"
 *      "-"  --> "lBm1%X"
 *      "+"  --> "lBp1%X"
 *      "++" --> "lBp2%X"
 */

void	xaddlabel(char *label, long value, int flags, int redefine)
{
	char	*lab = label;
	struct symbol *dummy = NULL;

	if(!strcmp(label,"--"))
		sprintf(label, "lBm2%X", label_index_m2++);
	else if(!strcmp(label,"-"))
		sprintf(label, "lBm1%X", label_index_m1++);
	else if(!strcmp(label,"+"))
		sprintf(label, "lBp1%X", label_index_p1++);
	else if(!strcmp(label,"++"))
		sprintf(label, "lBp2%X", label_index_p2++);
	if(*label == ':')
		lab++;
	else
		if(*label == '=') {
			lab++;
			flags |= SYM_GLOBAL;
		}
	if(*lab == '\0') {
		errmsg("Invalid symbol");
		return;
	}
	if(pass1) {
		dummy = addsymbol(label, value, flags);
		addlineref(dummy);
	}
	if(pass2) {
		if((dummy = findsymbol(label)) != NULL) {
			if((dummy->s_flags & SYM_DEFINED) && !redefine)
				errmsg("Duplicate symbol");
			else {
				dummy->s_flags |= SYM_DEFINED;
				if(dummy->s_value != value && !redefine)
					errmsg("Symbol changed (use old value)");
			}
		} else
			errmsg("Symbol was not defined in pass 1");
	}
	if(redefine && dummy != NULL)
		dummy->s_value = value;
}

void	addlabel(char *label, long value, int flags)
{
	xaddlabel(label, value, flags, FALSE);
}

void	allsymbols(void doit(struct symbol *), struct symbol *x)
{
	struct symbol *dummy;

	if(x != &s_terminal) {
		if(x->s_left != &s_terminal)
			allsymbols(doit, x->s_left);
		dummy = x->s_right;
		doit(x);
		if(dummy != &s_terminal)
			allsymbols(doit, dummy);
	}
}

void	free_symb(struct symbol *sym)
{
	struct lineref	*lp;
	struct fillref	*fp;

	lp = sym->s_lineref;
	while(lp != NULL) {
		FREE(lp);
		DEBUG_PRN("Address %ld: free'd memory\n", lp);
		lp = lp->l_next;
	}
	fp = sym->s_fillref;
	while(fp != NULL) {
		FREE(fp);
		DEBUG_PRN("Address %ld: free'd memory\n", fp);
		fp = fp->f_next;
	}
	FREE(sym);
	DEBUG_PRN("Address %ld: free'd memory\n", sym);
}

void	s_initialize(void)
{
	s_terminal.s_left = &s_terminal;
	s_terminal.s_right = &s_terminal;
	if(s_root != NULL)
		allsymbols(free_symb, s_root);
	s_root = newsymbol();
}

void	sys_initialize(void)
{
	s_terminal.s_left = &s_terminal;
	s_terminal.s_right = &s_terminal;
	sys_root = newsymbol();
}
