/*
 * code.c - SATURN loader, creating code
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>

#include "code.h"

extern	FILE *objfile, *modulefile;
extern	int codeonly, hexcode;
extern	char *objname, *modulename, hextable[];
#if	defined(__FLAT_MEM__)
extern	unsigned char *memory;
#else
extern	unsigned char *memory[];
#endif

unsigned char *tmpmemblock;

/* s_8AF2 */


unsigned char *getblock( void ) /* 000A */
{
	unsigned char *dummy;

#if	defined(__FLAT_MEM__)
	if( ( dummy = (unsigned char *)calloc( 64, 0x2000 ) ) == NULL )
		error( "G_BLK: %s", "unable to get memory block" );
#else
	if( ( dummy = (unsigned char *)calloc( 1, 0x2000 ) ) == NULL )
		error( "G_BLK: %s", "unable to get more memory" );
#endif
#if	defined(__FLAT_MEM__)
	DEBUG_PRN( "Address %ld: allocated %ld bytes (getblock)\n", dummy, 0x80000L );
#else
	DEBUG_PRN( "Address %ld: allocated %d bytes (getblock)\n", dummy, 0x2000 );
#endif
	return dummy;
}

unsigned char *gettmpblock( void ) /* 00EC */
{
	if( tmpmemblock == NULL )
		tmpmemblock = getblock();
	return tmpmemblock;
}

int getnibble( long adr ) /* 0110 */
{
	unsigned char *np;
#if	!defined(__FLAT_MEM__)
	int page;
#endif

	DEBUG_PRN( "Getting nibble from address %05lX\n", adr );
#if defined(__FLAT_MEM__)
	np = ( ( adr & 0xFFFFFL ) >> 1 ) + memory;
#else
	page = (int)( adr >> 14 );
	if( memory[page] == NULL )
		return 0;
	np = ( ( adr & 0x3FFF ) >> 1 ) + memory[page];
#endif
	if( ( adr & 1L ) != 0 )
		return *np & 0x0F;
	return ( *np >> 4 ) & 0x0F;
}

void fillnibble( long adr, long value ) /* 021E */
{
	unsigned char *np;
#if	!defined(__FLAT_MEM__)
	int page;
#endif

	DEBUG_PRN( "Putting nibble to address %05lX\n", adr );
#if	defined(__FLAT_MEM__)
	np = ( ( adr & 0xFFFFFL ) >> 1 ) + memory;
#else
	page = adr >> 14;
	if( memory[page] == NULL )
		memory[page] = getblock();
	np = ( ( adr & 0x3FFF ) >> 1 ) + memory[page];
#endif
	if( ( adr & 1 ) != 0 )
		*np = ( *np & 0xF0 ) | ( value & 0x0F );
	else
		*np = ( *np & 0x0F ) | ( ( value & 0x0F ) << 4 );
}

void fillbyte( long adr, long value ) /* 033C */
{
	unsigned char *bp;
#if	!defined(__FLAT_MEM__)
	int page;
#endif

	DEBUG_PRN( "Putting byte to address %05lX\n", adr );
#if	defined(__FLAT_MEM__)
	bp = ( ( adr & 0xFFFFFL ) >> 1 ) + memory;
#else
	page = adr >> 14;
	if( memory[page] == NULL )
		memory[page] = getblock();
	bp = ( ( adr & 0x3FFF ) >> 1 ) + memory[page];
#endif
	*bp = ( ( ( value & 0x0F ) << 8 ) | ( value & 0xF0 ) ) >> 4;
}


void read_code( long adr, long len ) /* 044A */
{
	register char si, di;
	unsigned char *mem_adr;
#if	!defined(__FLAT_MEM__)
	unsigned char **bp_20;
	int page;
	long page_adr;
#endif
	long mem_offset, bp_24, bp_28;

#if	defined(__FLAT_MEM__)
	mem_offset = ( adr & 0xFFFFFL ) >> 1;
	mem_adr = memory + mem_offset;
	if( ( adr & 1L ) != 0 ) {
		di = getc( modulefile );
		if( ferror( modulefile ) )
			error( "error reading from %s", modulename );
		*mem_adr = ( *mem_adr & 0xF0 ) | ( ( di >> 4 ) & 0x0F );
		++mem_adr;
		--len;
		while( len >= 2L ) {
			si = ( di & 0x0F ) << 4;
			di = getc( modulefile );
			if( ferror( modulefile ) )
				error( "error reading from %s", modulename );
			*mem_adr++ = ( ( di >> 4 ) & 0x0F ) | si;
			len -= 2L;
		}
	} else {
		bp_28 = bp_24 = len >> 1;
		if( bp_28 > 0x2000 )
			bp_28 = 0x2000;
/* reading in 0x2000 chunks 'cos on most OSs (int) is only 2 byte wide */
		while( bp_24 > 0L ) {
			if( fread( mem_adr, 1, (int)bp_28, modulefile ) != bp_28 )
				error( "error reading from %s", modulename );
			mem_adr += bp_28;
			bp_24 -= bp_28;
			bp_28 = bp_24;
			if( bp_28 > 0x2000 )
				bp_28 = 0x2000;
		}
		if( ( len & 1 ) != 0 ) {
			di = getc( modulefile );
			if( ferror( modulefile ) )
				error( "error reading from %s", modulename );
			di >>= 4;
		}
	}
	if( ( len & 1 ) != 0 ) {
		*mem_adr = ( ( *mem_adr >> 4 ) & 0x0F ) | ( ( di & 0x0F ) << 4 );
		++mem_adr;
	}
#else
	page = adr >> 14;
	bp_20 = memory + page;
	mem_adr = *bp_20; /* memory[page] */
	if( mem_adr == NULL )
		mem_adr = *bp_20 = getblock();
	mem_offset = ( adr & 0x3FFFL ) >> 1;
	page_adr = 0x2000L - mem_offset;
	mem_adr += mem_offset;
	if( ( adr & 1L ) != 0 ) {
		di = getc( modulefile );
		if( ferror( modulefile ) )
			error( "error reading from %s", modulename );
		*mem_adr = ( *mem_adr & 0xF0 ) | ( ( di >> 4 ) & 0x0F );
		++mem_adr;
		if( --page_adr == 0L ) {
			mem_adr = memory[++page];
			if( mem_adr == NULL )
				mem_adr = memory[page] = getblock();
			page_adr = 0x2000L;
		}
		--len;
		while( len >= 2L ) {
			si = ( di & 0x0F ) << 4;
			di = getc( modulefile );
			if( ferror( modulefile ) )
				error( "error reading from %s", modulename );
			*mem_adr++ = ( ( di >> 4 ) & 0x0F ) | si;
			if( --page_adr == 0L ) {
				mem_adr = memory[++page];
				if( mem_adr == NULL )
					mem_adr = memory[page] = getblock();
				page_adr = 0x2000L;
			}
			len -= 2L;
		}
	} else {
		bp_24 = len >> 1;
		bp_28 = page_adr;
		if( bp_28 > bp_24 )
			bp_28 = bp_24;
		while( bp_24 > 0L ) {
			/* umh, dangerous int<->long casts */
			if( fread( mem_adr, 1, (int)bp_28, modulefile ) != bp_28 )
				error( "error reading from %s", modulename );
			mem_adr += bp_28;
			bp_24 -= bp_28;
			page_adr -= bp_28;
			if( page_adr == 0 ) {
				mem_adr = memory[++page];
				if( mem_adr == NULL )
					mem_adr = memory[page] = getblock();
				page_adr = 0x2000;
			}
			bp_28 = bp_24;
			if( bp_28 > page_adr )
				bp_28 = page_adr;
		}
		if( ( len & 1 ) != 0 ) {
			di = getc( modulefile );
			if( ferror( modulefile ) )
				error( "error reading from %s", modulename );
			di >>= 4;
		}
	}
	if( ( len & 1 ) != 0 ) {
		*mem_adr = ( ( *mem_adr >> 4 ) & 0x0F ) | ( ( di & 0x0F ) << 4 );
		++mem_adr;
		/* hmm, following if(){} is redundant, we don't need
		 * to allocate space that we don't use in this routine
		 * any more and anyway same test is performed at the start
		 */
		if( --page_adr == 0 ) {
			mem_adr = memory[++page];
			if( mem_adr == NULL )
				mem_adr = memory[page] = getblock();
			page_adr = 0x2000;
		}
	}
#endif
}

void write_code( long adr, long len ) /* 08C4 */
{
	register char di = '\0', si = '\0';
	unsigned char *mem_adr;
#if	!defined(__FLAT_MEM__)
	int page;
#endif
	int nibs_rest;
	long page_adr, mem_offset, bp_26, bp_30;

	if( hexcode ) {
		page_adr = adr;
		while( page_adr-- != 0 )
			if( fputc( '0', objfile ) == EOF )
				error( "WR_CO: error writing to %s", objname );
		while( len-- != 0 )
			if( fputc( hextable[getnibble( adr++ )], objfile ) == EOF )
				error( "WR_CO: error writing to %s", objname );
		return;
	}
	if( codeonly ) {
		len += adr;
		adr = 0L;
	}
#if	defined(__FLAT_MEM__)
	mem_offset = ( adr & 0xFFFFFL ) >> 1;
	mem_adr = memory + mem_offset;
#else
	page = adr >> 14;
	mem_offset = ( adr & 0x3FFF ) >> 1;
	page_adr = 0x2000 - mem_offset;
	if( ( mem_adr = memory[page] ) == NULL )
		mem_adr = gettmpblock();
	mem_adr += mem_offset;
#endif
	nibs_rest = 256 - ( ( ( len + 1L ) / 2L ) & 0xFFL );
	if( ( adr & 1 ) != 0 ) {
		di = *mem_adr++;
#if	!defined(__FLAT_MEM__)
		if( --page_adr == 0 ) {
			if( ( mem_adr = memory[++page] ) == NULL )
				mem_adr = gettmpblock();
			page_adr = 0x2000;
		}
#endif
		while( len > 2L ) {
			if( codeonly ) {
				si = di & 0x0F;
				di = *mem_adr++;
#if	!defined(__FLAT_MEM__)
				if( --page_adr == 0 ) {
					if( ( mem_adr = memory[++page] ) == NULL )
						mem_adr = gettmpblock();
					page_adr = 0x2000;
				}
#endif
				si |= di & 0xF0;
			} else {
				si = ( di & 0x0F ) << 4;
				di = *mem_adr++;
#if	!defined(__FLAT_MEM__)
				if( --page_adr == 0 ) {
					if( ( mem_adr = memory[++page] ) == NULL )
						mem_adr = gettmpblock();
					page_adr = 0x2000;
				}
#endif
				si |= ( di >> 4 ) & 0x0F;
			}
			putc( si, objfile );
			if( ferror( objfile ) )
				error( "WR_CO: error writing to %s", objname );
			len -= 2;
		}
	} else {
		bp_26 = len >> 1;
		if( codeonly ) {
			while( --bp_26 >= 0L ) {
				si = *mem_adr++;
#if	!defined(__FLAT_MEM__)
				if( --page_adr == 0 ) {
					if( ( mem_adr = memory[++page] ) == NULL )
						mem_adr = gettmpblock();
					page_adr = 0x2000;
				}
#endif
				si = ( ( si & 0x0F ) << 4 ) | ( ( si >> 4 ) & 0x0F );
				putc( si, objfile );
				if( ferror( objfile ) )
					error( "WR_CO: error writing to %s", objname );
			}
		} else {
#if	defined(__FLAT_MEM__)
			bp_30 = bp_26;
			if( bp_30 > 0x2000 )
				bp_30 = 0x2000;
			while( bp_26 > 0L ) {
				if( fwrite( mem_adr, 1, (int)bp_30, objfile ) != bp_30 )
					error( "WR_Fw: error writing to %s", objname );
				mem_adr += bp_30;
				bp_26 -= bp_30;
				bp_30 = bp_26;
				if( bp_30 > 0x2000 )
					bp_30 = 0x2000;
			}
#else
			bp_30 = page_adr;
			if( bp_30 > bp_26 )
				bp_30 = bp_26;
			while( bp_26 != 0L ) {
				if( fwrite( mem_adr, 1, (int)bp_30, objfile ) != bp_30 )
					error( "WR_Cw: error writing to %s", objname );
				mem_adr += bp_30;
				bp_26 -= bp_30;
				page_adr -= bp_30;
				if( page_adr == 0L ) {
					if( ( mem_adr = memory[++page] ) == NULL )
						mem_adr = gettmpblock();
					page_adr = 0x2000;
				}
				bp_30 = page_adr;
				if( bp_30 > bp_26 )
					bp_30 = bp_26;
			}
#endif
		}
		if( ( len & 1 ) != 0 ) {
			si = *mem_adr++;
#if	!defined(__FLAT_MEM__)
			if( --page_adr == 0 ) {
				if( ( mem_adr = memory[++page] ) == NULL )
					mem_adr = gettmpblock();
				page_adr = 0x2000;
			}
#endif
			di = si >> 4;
		}
	}
	if( ( len & 1 ) != 0 ) {
		si = ( !codeonly ) ? ( di & 0x0F ) << 4 : di & 0x0F;
		putc( si, objfile );
		if( ferror( objfile ) )
			error( "WR_CO: error writing to %s", objname );
	}
	if( !codeonly ) {
		while( nibs_rest-- != 0 ) {
			putc( si, objfile );
			if( ferror( objfile ) )
				error( "WR_CO: error writing to %s", objname );
		}
	}
}

long compute_crc( long crc_start, long crc_offset, long crc_reads, long crc_sectors ) /* 0FAE */
{
	int crc_table[16] = {
			 0, 0x168, 0x2D0, 0x3B8, 0x5C8, 0x4A0, 0x718, 0x670,
		0xBF8, 0xA90, 0x928, 0x840, 0xE30, 0xF58, 0xCE0, 0xD88
	};
	long crc = 0L, bp_8, bp_12, bp_16, bp_20;
	int bp_22;
	unsigned char *mem_adr, *bp_42;
#if	!defined(__FLAT_MEM__)
	int bp_34, bp_38, bp_46, bp_50;
#endif
	long bp_54;

	DEBUG_PRN( "COMPUTECRC: Start = %05lX, Offset = %05lX\n", crc_start, crc_offset );
	DEBUG_PRN( "COMPUTECRC: Reads = %05lX, Sectors = %05lX\n", crc_reads, crc_sectors );
	while( crc_sectors-- != 0L ) {
		bp_8 = crc_start;
#if	defined(__FLAT_MEM__)
		bp_22 = ( bp_8 & 0xFFFFFL ) >> 1;
		mem_adr = memory + bp_22;
#else
		bp_34 = crc_start >> 14;
		bp_22 = ( bp_8 & 0x3FFF ) >> 1 ;
		bp_38 = 0x2000 - bp_22;
		if( ( mem_adr = memory[bp_34] ) == NULL )
			mem_adr = gettmpblock();
		mem_adr += bp_22;
#endif
		bp_16 = crc_start + crc_offset;
#if	defined(__FLAT_MEM__)
		bp_22 = ( bp_16 & 0xFFFFFL ) >> 1;
		bp_42 = memory + bp_22;
#else
		bp_46 = bp_16 >> 14;
		bp_22 = ( bp_16 & 0x3FFF ) >> 1 ;
		bp_50 = 0x2000 - bp_22;
		if( ( bp_42 = memory[bp_46] ) == NULL )
			bp_42 = gettmpblock();
		bp_42 += bp_22;
#endif
		bp_12 = crc_reads;
		while( bp_12-- != 0L ) {
			if( ( bp_8 & 1L ) != 0 ) {
				bp_54 = (long)*mem_adr++;
#if	!defined(__FLAT_MEM__)
				if( --bp_38 == 0L ) {
					if( ( mem_adr = memory[++bp_34] ) == NULL )
						mem_adr = gettmpblock();
					bp_38 = 0x2000L;
				}
#endif
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( crc << 4 ) ^ ( bp_54 & 0x0F );
			}
			bp_20 = ( bp_8 + 1L ) & 0xFFFFFFFEL; /* do this better */
		      while( ( bp_8 + 15L ) > bp_20 ) {
				bp_54 = (long)*mem_adr++;
#if	!defined(__FLAT_MEM__)
				if( --bp_38 == 0L ) {
					if( ( mem_adr = memory[++bp_34] ) == NULL )
						mem_adr = gettmpblock();
					bp_38 = 0x2000L;
				}
#endif
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( ( bp_54 >> 4 ) & 0x0F ) ^ ( crc << 4 );
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( crc << 4 ) ^ ( bp_54 & 0x0F );
				bp_20 += 2L;
			}
			if( ( bp_8 & 1L ) != 0 ) {
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( ( (int)*mem_adr >> 4 ) & 0x0F ) ^ ( crc << 4 );
			}
			bp_8 += 16L;
			if( ( bp_16 & 1L ) != 0 ) {
				bp_54 = (long)*bp_42++;
#if	!defined(__FLAT_MEM__)
				if( --bp_50 == 0L ) {
					if( ( bp_42 = memory[++bp_46] ) == NULL )
						bp_42 = gettmpblock();
					bp_50 = 0x2000L;
				}
#endif
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( crc << 4 ) ^ ( bp_54 & 0x0F );
			}
			bp_20 = ( bp_16 + 1L ) & ~(-2L); /* 0xFFFFFFFEL */
	      while( ( bp_16 + 15L ) > bp_20 ) {
				bp_54 = (long)*bp_42++;
#if	!defined(__FLAT_MEM__)
				if( --bp_50 == 0L ) {
					if( ( bp_42 = memory[++bp_46] ) == NULL )
						bp_42 = gettmpblock();
					bp_50 = 0x2000L;
				}
#endif
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( ( bp_54 >> 4 ) & 0x0F ) ^ ( crc << 4 );
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( crc << 4 ) ^ ( bp_54 & 0x0F );
				bp_20 += 2L;
			}
			if( ( bp_16 & 1L ) != 0 ) {
				crc = crc_table[ ( crc >> 12 ) & 0x0F ] ^ ( ( (int)*bp_42 >> 4 ) & 0x0F ) ^ ( crc << 4 );
			}
			bp_16 += 16L;
		}
		crc_start += crc_offset << 1;
	}
	return crc;
}

int compute_cksum( long cksum_start, long cksum_end ) /* 16AC */
{
	long cksum = 0L;
	long bp_16;
	long bp_20;
	int bp_22;
	unsigned char *bp_30;
	long bp_34;
#if	!defined(__FLAT_MEM__)
	int bp_4, page;
#endif
	DEBUG_PRN( "COMPUTECKSUM: Start = %05lX, End = %05lX\n", cksum_start, cksum_end );
	bp_16 = cksum_start;
#if	defined(__FLAT_MEM__)
	bp_22 = ( bp_16 & 0xFFFFFL ) >> 1;
	bp_30 = memory + bp_22;
#else
	page = bp_16 >> 14;
	bp_22 = ( bp_16 & 0x3FFF ) >> 1;
	bp_4 = 0x2000 - bp_22;
	if( ( bp_30 = memory[page] ) == NULL )
		bp_30 = gettmpblock();
	bp_30 += bp_22;
#endif
	if( ( bp_16 & 1L ) != 0 ) {
		bp_34 = (long)*bp_30++;
#if	!defined(__FLAT_MEM__)
		if( --bp_4 == 0 ) {
			if( ( bp_30 = memory[++page] ) == NULL )
				bp_30 = gettmpblock();
			bp_4 = 0x2000;
		}
#endif
		cksum = ( ( ( bp_34 ^ cksum ) & 0x0F ) * 0x1081 ) ^ ( cksum >> 4 );
	}
	bp_20 = ( bp_16 + 1L ) & 0xFFFFFFFEL; /* do this better */
		while( ( cksum_end - 1L ) > bp_20 ) {
		bp_34 = (long)*bp_30++;
#if	!defined(__FLAT_MEM__)
		if( --bp_4 == 0 ) {
			if( ( bp_30 = memory[++page] ) == NULL )
				bp_30 = gettmpblock();
			bp_4 = 0x2000;
		}
#endif
		cksum = ( ( ( ( bp_34 >> 4 ) ^ cksum ) & 0x0F ) * 0x1081 ) ^ ( cksum >> 4 );
		cksum = ( ( ( bp_34 ^ cksum ) & 0x0F ) * 0x1081 ) ^ ( cksum >> 4 );
		bp_20 += 2L;
	}
	if( ( cksum_end & 1L ) != 0 )
		cksum = ( ( ( ( (int)*bp_30 >> 4 ) ^ cksum ) & 0x0F ) * 0x1081 ) ^ ( cksum >> 4 );
	cksum &= 0xFFFF;
	cksum |= cksum << 16;
	return (int)( ( cksum & 0x000F0F00L ) >> 4 ) | ( ( cksum & 0x00F0F000L ) >> 12 );
}

long compute_kcrc( long kcrc_start, long kcrc_offset, long kcrc_reads, long kcrc_sectors ) /* 1A06 */
{
	long kcrc = 0L;
	long bp_8, bp_12, bp_16, bp_20;
	int bp_22;
	unsigned char *bp_30;
	unsigned char *bp_42;
	long bp_54;
#if	!defined(__FLAT_MEM__)
	int bp_34, bp_38, bp_46, bp_50;
#endif

	DEBUG_PRN( "COMPUTEKCRC: Start = %05lX, Offset = %05lX\n", kcrc_start, kcrc_offset );
	DEBUG_PRN( "COMPUTEKCRC: Reads = %05lX, Sectors = %05lX\n", kcrc_reads, kcrc_sectors );
	while( kcrc_sectors-- != 0L ) {
		bp_8 = kcrc_start;
#if	defined(__FLAT_MEM__)
		bp_22 = ( bp_8 & 0xFFFFFL ) >> 1;
		bp_30 = memory + bp_22;
#else
		bp_34 = bp_8 >> 14;
		bp_22 = ( bp_8 & 0x3FFF ) >> 1 ;
		bp_38 = 0x2000 - bp_22;
		if( ( bp_30 = memory[bp_34] ) == NULL )
			bp_30 = gettmpblock();
		bp_30 += bp_22;
#endif
		bp_16 = kcrc_start + kcrc_offset;
#if	defined(__FLAT_MEM__)
		bp_22 = ( bp_16 & 0xFFFFFL ) >> 1;
		bp_42 = memory + bp_22;
#else
		bp_46 = bp_16 >> 14;
		bp_22 = ( bp_16 & 0x3FFF ) >> 1 ;
		bp_50 = 0x2000 - bp_22;
		if( ( bp_42 = memory[bp_46] ) == NULL )
			bp_42 = gettmpblock();
		bp_42 += bp_22;
#endif
		bp_12 = kcrc_reads;
		while( bp_12-- != 0L ) {
			if( ( bp_8 & 1L ) != 0 ) {
				bp_54 = (long)*bp_30++;
#if	!defined(__FLAT_MEM__)
				if( --bp_38 == 0L ) {
					if( ( bp_30 = memory[++bp_34] ) == NULL )
						bp_30 = gettmpblock();
					bp_38 = 0x2000L;
				}
#endif
				kcrc = ( ( ( bp_54 ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
			}
			bp_20 = ( bp_8 + 1L ) & ~(-2L);
	      while( ( bp_8 + 15L ) > bp_20 ) {
				bp_54 = (long)*bp_30++;
#if	!defined(__FLAT_MEM__)
				if( --bp_38 == 0L ) {
					if( ( bp_30 = memory[++bp_34] ) == NULL )
						bp_30 = gettmpblock();
					bp_38 = 0x2000L;
				}
#endif
				kcrc = ( ( ( ( bp_54 >> 4 ) ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
				kcrc = ( ( ( bp_54 ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
				bp_20 += 2L;
			}
			if( ( bp_8 & 1L ) != 0 ) {
				kcrc = ( ( ( ( (int)*bp_30 >> 4 ) ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
			}
			bp_8 += 16L;
			if( ( bp_16 & 1L ) != 0 ) {
				bp_54 = (long)*bp_42++;
#if	!defined(__FLAT_MEM__)
				if( --bp_50 == 0L ) {
					if( ( bp_42 = memory[++bp_46] ) == NULL )
						bp_42 = gettmpblock();
					bp_50 = 0x2000L;
				}
#endif
				kcrc = ( ( ( bp_54 ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
			}
			bp_20 = ( bp_16 + 1L ) & ~(-2L);
	      while( ( bp_16 + 15L ) > bp_20 ) {
				bp_54 = (long)*bp_42++;
#if	!defined(__FLAT_MEM__)
				if( --bp_50 == 0L ) {
					if( ( bp_42 = memory[++bp_46] ) == NULL )
						bp_42 = gettmpblock();
					bp_50 = 0x2000L;
				}
#endif
				kcrc = ( ( ( ( bp_54 >> 4 ) ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
				kcrc = ( ( ( bp_54 ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
				bp_20 += 2L;
			}
			if( ( bp_16 & 1L ) != 0 ) {
				kcrc = ( ( ( ( (int)*bp_42 >> 4 ) ^ kcrc ) & 0x0FL ) * 0x1081L ) ^ ( kcrc >> 4 );
			}
			bp_16 += 16L;
		}
		kcrc_start += kcrc_offset << 1;
	}
/*		return (long)s_8AF2_20AE( kcrc, -1 );
 */
	return kcrc;
}

int s_8AF2_20AE( /* ?? long ?? */ int bp6, int bp8 ) /* 20AE */
{
	int bp_2 = 0; /* this line doesn't exists in source */
	int bp_4 = bp6; /* ?? long ?? */
	int bp_6 = bp8;

	bp_4 = ( ( ( bp_4 >> 12 ) * 0x811 ) ^ ( bp_4 << 4 ) ) ^ ( bp_6 & 0x0F );
	bp_6 >>= 4;
	bp_4 = ( ( ( bp_4 >> 12 ) * 0x811 ) ^ ( bp_4 << 4 ) ) ^ ( bp_6 & 0x0F );
	bp_6 >>= 4;
	bp_4 = ( ( ( bp_4 >> 12 ) * 0x811 ) ^ ( bp_4 << 4 ) ) ^ ( bp_6 & 0x0F );
	bp_6 >>= 4;
	bp_4 = ( ( ( bp_4 >> 12 ) * 0x811 ) ^ ( bp_4 << 4 ) ) ^ ( bp_6 & 0x0F );
	bp_6 = -1999;
	bp_2 = ( ( bp_4 ^ -1999 ) & 0x0F ) | ( bp_2 << 4 ); /* uh oh bp_2 ?!?! */
	bp_4 = ( bp_4 >> 4 ) ^ 0x1081;
	bp_6 >>= 4;
	bp_2 = ( ( bp_4 ^ bp_6 ) & 0x0F ) | ( bp_2 << 4 );
	bp_4 = ( ( bp_6 & 0x0F ) * 0x1081 ) ^ ( bp_4 >> 4 );
	bp_6 >>= 4;
	bp_2 = ( ( bp_4 ^ bp_6 ) & 0x0F ) | ( bp_2 << 4 );
	bp_4 = ( ( bp_6 & 0x0F ) * 0x1081 ) ^ ( bp_4 >> 4 );
	bp_6 >>= 4;
	bp_2 = ( ( bp_4 ^ bp_6 ) & 0x0F ) | ( bp_2 << 4 );
	return bp_2;
}

long compute_ecrc( long ecrc_start, long ecrc_end ) /* 21D8 */
{
	long ecrc = 0L;
	long bp_16;
	long mem_offset;
	unsigned char *mem_adr;
	long bp_38;
#if	!defined(__FLAT_MEM__)
	int bp_4, page;
#endif
	DEBUG_PRN( "COMPUTEECRC: Start = %05lX, End = %05lX\n", ecrc_start, ecrc_end );
	bp_16 = ecrc_start;
#if	defined(__FLAT_MEM__)
	mem_offset = ( bp_16 & 0xFFFFFL ) >> 1;
	mem_adr = memory + mem_offset;
#else
	page = bp_16 >> 14;
	mem_offset = ( bp_16 & 0x3FFF ) >> 1;
	bp_4 = 0x2000 - mem_offset;
	if( ( mem_adr = memory[page] ) == NULL )
		mem_adr = gettmpblock();
	mem_adr += mem_offset;
#endif
	if( bp_16 < ecrc_end )
		if( ( bp_16++ & 1L ) != 0 ) {
			bp_38 = (long)*mem_adr++;
#if	!defined(__FLAT_MEM__)
			if( --bp_4 == 0 ) {
				if( ( mem_adr = memory[++page] ) == NULL )
					mem_adr = gettmpblock();
				bp_4 = 0x2000;
			}
#endif
			ecrc = ( ( ecrc >> 4 ) & 0x0FL ) + ( ecrc << 4 ) + ( bp_38 & 0x0FL ) + 1L;
			ecrc += ( ( ecrc & 0xF000 ) >> 12 ) + 1L;
		}
	bp_16 &= ~(1L);
	while( ( ecrc_end & ~(1L) ) > bp_16 ) {
		bp_38 = (long)*mem_adr++;
#if	!defined(__FLAT_MEM__)
		if( --bp_4 == 0 ) {
			if( ( mem_adr = memory[++page] ) == NULL )
				mem_adr = gettmpblock();
			bp_4 = 0x2000;
		}
#endif
		ecrc = ( ( bp_38 >> 4 ) & 0x0FL ) + ( ( ecrc >> 4 ) & 0x0FL ) + ( ecrc << 4 ) + 1L;
		ecrc += ( ( ecrc & 0xF000 ) >> 12 ) + 1L;
		ecrc = ( ( ecrc >> 4 ) & 0x0FL ) + ( ecrc << 4 ) + ( bp_38 & 0x0FL ) + 1L;
		ecrc += ( ( ecrc & 0xF000 ) >> 12 ) + 1L;
		bp_16 += 2L;
	}
	if( bp_16 < ecrc_end ) {
		ecrc = (long)( ( (int)*mem_adr >> 4 ) & 0x0F ) + ( ( ecrc >> 4 ) & 0x0F ) + ( ecrc << 4 ) + 1L;
		ecrc += ( ( ecrc & 0xF000 ) >> 12 ) + 1L;
	}
	ecrc &= 0xFFFF;
	ecrc |= ecrc << 16;
	return ( ( ecrc & 0x000F0F00L ) >> 4 ) | ( ( ecrc & 0x00F0F000L ) >> 12 );
}
