/*
 * newob.c - SATURN loader, creating new objects
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif

#include "struct.h"
#include "newob.h"

int	d_23e0;
int	d_23e2;
int	d_23e4;
struct symbol	*d_5c42;
char	d_5c46[256];

extern	int	symbolcount, refcount, errorcount;
extern	char	char_table[];
extern	struct	module	*module_last, d_0372;


void	s_8A85_000A( void )
{
	symbolcount = 0;
	refcount = 0;
	d_5c42 = NULL;
}

void	s_8A85_0038( void )
{
	sprintf( d_5c46, "f(%d %4.2f) i(%d)\n", d_23e0, ( d_23e0 == 0 ) ? 0 : ( (double)d_23e2 / (double)d_23e0 ), d_23e4 );
}

void	do_all_symbols( void doit( struct symbol * ) )
{
	s_8D4E_03EE( d_5c42, doit );
}

struct	symbol	*newsymbol( char	*sym )
{
	struct	symbol	*dummy;

	DEBUG_PRN( "Adding symbol '%s' to symbol table\n", sym );
	if( ( dummy = ( struct symbol *)malloc( sizeof( struct symbol ) ) ) == NULL )
		error( "error malloc'ing %ld bytes for new symbol", sizeof( struct symbol ) );
	DEBUG_PRN( "Address %ld: allocated %d bytes (newsymbol)\n", dummy, sizeof( struct symbol ) );
	strncpy( dummy->s_name, sym, SYM_NAME_LEN );
	dummy->s_name[SYM_NAME_LEN] = '\0';
	dummy->s_value = 0L;
	dummy->s_flags = SYM_RELOCATABLE;
	dummy->s_module = module_last;
	dummy->s_fref_first = NULL;
	dummy->s_fref_next = &dummy->s_fref_first;
	++symbolcount;
	return dummy;
}

struct fillref *newfillref( void )
{
	struct fillref *dummy;

	DEBUG_PRN( "Adding fill reference to symbol table\n" );
	if( ( dummy = malloc( sizeof( struct fillref ) ) ) == NULL )
		error( "error malloc'ing %ld bytes for new fill reference", sizeof( struct fillref ) );
	DEBUG_PRN( "Address %ld: allocated %d bytes (newfillref)\n", dummy, sizeof( struct fillref ) );
	dummy->f_next = NULL;
	++refcount;
	return dummy;
}

struct	symbol	*s_8A85_0414( char *sym )
{
	char	*cp;

	sym[SYM_NAME_LEN] = '\0';
	if( ( cp = strchr( sym, ' ' ) ) != NULL )
		*cp = '\0';
	return s_8D4E_000C( sym, &d_5c42, FALSE );
}

struct	symbol	*s_8A85_0462( char *sym )
{
/*	struct	symbol	*symbp;
 */
	char	*cp;

	if( (unsigned char)*sym >= 0x80 || ( char_table[(int)*sym] & CHR_OK1ST ) == 0 )
		errmsg( "Illegal first character in symbol" );
	sym[SYM_NAME_LEN] = '\0';
	if( ( cp = strchr( sym, ' ' ) ) != NULL )
		*cp = '\0';
	return s_8D4E_000C( sym, &d_5c42, TRUE );

/*	sumbp = s_8D4E_000C( sym, &d_5c42, TRUE );
	return symbp;
 */
}

void	s_8A85_04DE( char *bp6 )
{
	struct	symbol	*bp_4;
	char	*bp_8;
	char	bp_22[13]; /* SYM_NAME_LEN + 1 */
	long	bp_26;
	int	bp_28;

	bp_26 = 1L;

	if( *bp6 == '=' )
		bp6++;
	bp_28 = ( ( bp_8 = strrchr( bp6, '=' ) ) != NULL ) ? bp_8 - bp6 : strlen( bp6 );
	if( bp_28 > 12 )
		bp_28 = 12;
	strncpy( bp_22, bp6, bp_28 );
	bp_22[bp_28] = '\0';
	if( bp_8 != NULL && bp_8[1] != '\0' ) {
		if( sscanf( bp_8 + 1, "%ld", &bp_26 ) != 1 )
			fprintf( stderr, "Non-numeric expression for -D (%s)\n", bp_8 + 1 );
		errorcount++;
	} else {
		bp_4 = s_8A85_0462( bp_22 );
		bp_4->s_value = bp_26;
		bp_4->s_flags = SYM_UNDEFINED;
		bp_4->s_module = &d_0372;
	}
}


void	addfillref( struct symbol *symbp, union SFC *unp )
{
	struct	fillref	*dummy;

	dummy = newfillref();
	*symbp->s_fref_next = dummy;
	symbp->s_fref_next = &dummy->f_next;
	++symbp->s_flags;
	dummy->f_address = module_last->address + get_32( (unsigned char *)unp->f.of_address );
	dummy->f_bias = get_32( (unsigned char *)unp->f.of_bias );
	dummy->f_class = *unp->f.of_class;
	dummy->f_subclass = *unp->f.of_subclass;
	dummy->f_length = get_16( (unsigned char *)unp->f.of_length );
	dummy->f_module = module_last;
}
