/*
 * read.c - SATURN loader, input routines
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif

#include "read.h"

extern	char	errmsgbuf[], *ptr, *oldptr, char_table[256];
extern	int	column, oldcolumn;
extern	long	codeaddr;


char	get_chr( void )
{
	char	c;

	c = *ptr;
	oldcolumn = column;
	oldptr = ptr;
	if( c == '\t' ) {
		column += 8 - ( ( column - 1 ) % 8 );
		ptr++;
	} else if( c != '\0' ) {
		ptr++;
		column++;
	}
	return c;
}

void	backchar( void )
{
	ptr = oldptr;
	column = oldcolumn;
}

int	getfield( char *fieldp, int maxlength, char *separators )
{
	char	*tmp, *tmpchr;
	int	length = 0;

	if( strchr( separators, *ptr ) == NULL ) {
		if( ( tmp = strpbrk( ptr, separators ) ) == NULL ) {
			length = strlen( ptr );
			tmp = ptr + length;
		} else
			length = tmp - ptr; /* sub ax,ptr */
		if( length > maxlength ) {
			errmsg( "Field too long" );
			length = maxlength;
		}
		strncpy( fieldp, ptr, length );
		if( ( tmpchr = strchr( ptr, '\t' ) ) != NULL && tmpchr < tmp ) {
			while( ptr < tmp )
				get_chr();
		} else {
			column += length;
			ptr = tmp;
		}
	}
	fieldp[length] = '\0';
	return ( length != 0 ) ? TRUE : FALSE; /* cmp ax,1 ; sbb ax,ax ; inc ax */
}

void	skipwhite( void )
{
	char	c;

	while( ( c = *ptr ) == ' ' || c == '\t' ) {
		ptr++;
		if( c == ' ' )
			column++;
		else
			column += 8 - ( ( column - 1 ) % 8 );
	}
}

void	next_token( void )
{
	if( *ptr == ',' )
		get_chr();
	skipwhite();
}

char	*getline( char *lineptr, FILE *file )
{
	char	*pt;

	column = 0L;
	if( fgets( lineptr, 256, file ) == NULL )
		return NULL;
	if( ( strlen( lineptr ) == 255 ) && lineptr[255] != '\n' ) {
		char	tempbuf[256];
		errmsg( "Input line too long (extra characters ignored)" );
		/* might do next line as in sasm.c */
		while( strlen( pt = fgets( tempbuf, 256, file ) ) == 255 && tempbuf[255] != '\n' );
	}
	if( ( pt = strchr( lineptr, '\n' ) ) != NULL )
		*pt = '\0';
	if( ( pt = strrchr( lineptr, '\r' ) ) != NULL && pt[1] == '\0' )
		*pt = '\0';
	if( *lineptr == 26 /* 0x1A */ )
		*lineptr = '\0';
	return lineptr;
}

long	getaddress( void )
{
	char	c;
	long	addr;

	addr = 0L;
	next_token();
	c = get_chr();
	if( c == '*' ) {
		addr = codeaddr;
		c = *ptr;
		if( c != '\0' && strchr( " \t,", c ) != NULL ) {
			errmsg( "Invalid address (* must be used alone)" );
			addr = -1L;
		}
		return addr;
	}
	while( strchr( " \t,", c ) == NULL ) {
		if( ( char_table[(int)c] & CHR_HEX ) != 0 ) {
			addr = chartohex( c ) + ( addr << 4 );
			c = get_chr();
			continue;
		}
		sprintf( errmsgbuf, "Unrecognized hexadecimal character '%c'", c );
		errmsg( errmsgbuf );
		addr = -1L;
		break;
	}
	if( addr != -1L )
		backchar();
	if( addr >= 0x100000L ) {
		errmsg( "Address too large" );
		addr = -1L;
	}
	return addr;
}

int	getcommand( void )
{
	char	c;
	int	cmd2 = 0; /* 2-char command */

	next_token();
	c = get_chr();

/* commented original
 *
 *	if( ( tmpchr = char_table[c] & 2 ) != '\0' ) ** byte
 *		c -= ( tmpchr == '\0' ) ? 0 : 0x20 ; ** a bit of redudant code
 *	cmd2 = 256 * c;
 *	c = get_chr();
 *	if( ( tmpchr = char_table[c] & 2 ) != '\0' ) ** byte
 *		c -= ( tmpchr == '\0' ) ? 0 : 0x20 ; ** a bit of redudant code
 */

	if( char_table[(int)c] & CHR_LOWCASE )
		c -= 'a' - 'A';
	cmd2 = c << 8;
	c = get_chr();
	if( char_table[(int)c] & CHR_LOWCASE )
		c -= 'a' - 'A';
	if( !strchr( " \t,", c ) )
		cmd2 |= (int)c;
	else
		backchar();
	while( !strchr( " \t,", c ) )
		c = get_chr();
	backchar();
	return cmd2;
}

char	*gettoken( void )
{
	static char	tokenbuf[256];

	next_token();
	if( getfield( tokenbuf, 256, " \t" ) == FALSE )
		return NULL;
	else
		return tokenbuf;
}

char	*getsymbol( char *symbuf )
{
	next_token();
	if( getfield( symbuf, 12, " \t" ) == FALSE )
		return NULL;
	else
		return symbuf;

}

char	*getstring( void )
{
	char	*str, ch;

	next_token();
	if( *ptr == '\0' )
		return NULL;
	ch = *ptr;
	if( ch == '"' || ch == '\'' )
		if( ( str = strrchr( ++ptr, ch ) ) != NULL )
			*str = '\0';
	return ptr;
}

void	remove_tabs( char *lineptr, char *strptr, int line_len )
{
	char	ch, tab_space[] = "        ";
	int	tab_pos = 0;

	while( ( ch = *strptr++ ) != '\0' && tab_pos < line_len ) {
		if( ch != '\t' ) {
			*lineptr++ = ch;
			tab_pos++;
		} else {
			strcpy( lineptr, tab_space + ( tab_pos % 8 ) );
			lineptr += tab_pos % 8;
			tab_pos += tab_pos % 8;
		}
	}
	*lineptr = '\0';
}
