/*
 * sload.c - SATURN loader, main loop
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif
#if	defined(__MSDOS__)
#include <io.h>
#endif
#if	defined(__FREE__)
# include <malloc.h>
#endif
#include <time.h>
/* #include <sys/time.h> */

#include "struct.h"
#include "sload.h"

FILE	*dbgfile, *msgfile, *listfile, *promptfile, *infile, *objfile, *modulefile;
char	*dbgname, *msgname, *listname, *promptname, *inname, *objname, *modulename;
char	*progname;
char	*ptr, *oldptr, *prompt_string;
char	line[256], date[26], title[40], errmsgbuf[256], softkeys[20];
#if	defined(__FLAT_MEM__)
char	*memory;
#else
char	*memory[64]; /* original is 128, but this is byte count not nibble */
#endif
char	hextable[] = "0123456789ABCDEF";
char	listversion[] = "@(#) Saturn Loader, Ver 2.1.8 (GNU)";
char	*listver = listversion + 5;
int	column, oldcolumn;
int	option_flags, user_flags[100];
int	errorcount, symbolcount, refcount;
int	aborterr = 0, list_mode = LIST_DEFAULT;
long	pagelength = 60L, pagewidth = 80L, lineofpage;
long	codestart = 0L, codeaddr = 0L, romid = 0L;
struct	module	*module_root = NULL, *module_last = NULL;
struct	crc	*crc_root = NULL;
struct	checksum	*cksum_root = NULL;
int
	debug				=	FALSE,
	prompt_mode		=	FALSE,
	aborted			=	FALSE,
	no_resolve		=	FALSE,
	msg_append		=	FALSE,
	update			=	FALSE,
	codeonly			=	FALSE,
	hexcode			=	FALSE,
	full_filenames	=	FALSE,
	listing			=	TRUE,
	absolute			=	TRUE,
	reading			=	TRUE;

extern	int	opterr, optind, sp;
extern	char	*optarg, *nameptr, *version;

struct module	*d_001c;
int	d_0024;
char	d_00ee[14];
char	d_021e[14];
long	d_032e;
char	d_0332[14];
char	*d_0348;
struct module	d_0372;
struct module	*d_03c8;
int	d_05ea = FALSE;
int	d_05f4 = FALSE;
long	d_05f6 = 0L;
int	d_05fa = FALSE;
int	d_0602 = FALSE;
long	d_061a = 0L;
long	d_061e = 0L;
int	d_0622 = FALSE;
int	d_0688;

unsigned	char	*d_25f6;
int	d_2be7;
char	ds_393a[256], ds_3a3a[256];
int	d_3b3a, d_3b3c;
char	d_3b3e[0x2000];
int	d_5b3e;
unsigned	char	*d_5c78[128];

unsigned char	char_table[256] = { /* d_31ff */
	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,
	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,
	CHR_CONTROL,	CHR_PRCNTRL,	CHR_PRCNTRL,	CHR_PRCNTRL,
	CHR_PRCNTRL,	CHR_PRCNTRL,	CHR_CONTROL,	CHR_CONTROL,
	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,
	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,
	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,
	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,	CHR_CONTROL,

	( CHR_SPACE ) | ( CHR_PRNABLE ),	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,
	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,
	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,
	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,

	CHR_HEXDEC,	CHR_HEXDEC,	CHR_HEXDEC,	CHR_HEXDEC,
	CHR_HEXDEC,	CHR_HEXDEC,	CHR_HEXDEC,	CHR_HEXDEC,
	CHR_HEXDEC,	CHR_HEXDEC,	CHR_PUNCT,	CHR_PUNCT,
	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,

	CHR_PUNCT,	CHR_UPHEX,	CHR_UPHEX,	CHR_UPHEX,
	CHR_UPHEX,	CHR_UPHEX,	CHR_UPHEX,	CHR_UPCASE,
	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,
	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,

	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,
	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,
	CHR_UPCASE,	CHR_UPCASE,	CHR_UPCASE,	CHR_PUNCT,
	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,

	CHR_PUNCT,	CHR_LOHEX,	CHR_LOHEX,	CHR_LOHEX,
	CHR_LOHEX,	CHR_LOHEX,	CHR_LOHEX,	CHR_LOWCASE,
	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,
	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,

	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,
	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,
	CHR_LOWCASE,	CHR_LOWCASE,	CHR_LOWCASE,	CHR_PUNCT,
	CHR_PUNCT,	CHR_PUNCT,	CHR_PUNCT,	CHR_CONTROL,

	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD,
	CHR_BAD,	CHR_BAD,	CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD, CHR_BAD

/*	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x28, 0x28, 0x28, 0x28, 0x28, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x48, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
	0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
	0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84, 0x84,
	0x84, 0x84, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
	0x10, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x10, 0x10, 0x10, 0x10, 0x10,
	0x10, 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0x02,
	0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02,
	0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x02,
	0x02, 0x02, 0x02, 0x10, 0x10, 0x10, 0x10, 0x20,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 */
};

void	USAGE( char *fmt_str, ... )
{
	va_list arg_ptr;
	int err;

	va_start( arg_ptr, fmt_str );
	err = vfprintf( msgfile, fmt_str, arg_ptr );
	va_end( arg_ptr );
	if( err < 0 )
		error( "error writing to message file (%s)", msgname );
}

void	usage( int flag ) /* 0000 */
{
	USAGE( "sload %s\n", version );
	if( flag )
		return;
	USAGE( "Usage: %s [options] ... [sourcefile]\n", progname );
	USAGE( "options:\n" );
	USAGE( "-D symb[=val]\tdefine a symbol [default value = 1]\n" );
	USAGE( "-d debugfile\tdebug mode, specified output file\n" );
	USAGE( "-e\t\twrite error messages to message file\n" );
	USAGE( "-f flaglist\tset flags selected by flaglist (comma-separated)\n" );
	USAGE( "-H\t\tsuppress object file header and symbol (write raw code)\n" );
	USAGE( "-h\t\tsuppress object file header and symbol (write hex code)\n" );
	USAGE( "-l listfile\twrite the listing to listfile\n" );
	USAGE( "-M msgfile\tredirect messages to msgfile (append to msgfile)\n" );
	USAGE( "-m msgfile\tredirect messages to msgfile (overwrite msgfile)\n" );
	USAGE( "-N\t\tdo not produce an annotated listing\n" );
	USAGE( "-o objfile\twrite the object file to objfile [default=<name>.o]\n" );
	USAGE( "-p pagelen\tdo a page break each 'pagelen' lines [default=60]\n" );
	USAGE( "-P\t\tprompt for input\n" );
	USAGE( "-u\t\tforce all files to update mode\n" );
	USAGE( "-v\t\tprint version\n" );
	USAGE( "-w pagewidth\tset output page width to 'width' columns [default=80]\n" );

}

void  pagebreak( void ) /* 0854 */
{
	if( listing ) {
		lineofpage = 1L; /* dword */
		if( putc( 12, listfile ) == EOF )
			error( "error writing to list file (%s)", listname );
	}
}

void  do_error( char *msg ) /* 08D4 */
{
	register int len;

	len = strlen( msg ) - 1;
	if( listing ) {
		++lineofpage;
		if( lineofpage > pagelength )
			pagebreak();
		while( len >= 0 && msg[len] == ' ' )
			msg[len--] = '\0';
		if( fputs( msg, listfile ) == EOF || putc( '\n', listfile ) == EOF )
			error( "error writing to list file (%s)", listname );
	}
}

void  last_line( void ) /* 0994 */
{
	lineofpage = pagelength;
}

void  errmsg( char *message ) /* 09B0 */
{
	char	msgbuffer[256];

	++errorcount;
	sprintf( msgbuffer, "%s%s", prompt_string, message );
	if( d_0622 ) {
		if( fprintf( msgfile, msgbuffer ) < 0 || putc( '\n', msgfile ) == EOF )
			error( "error writing to message file (%s)", msgname );
	}
	do_error( msgbuffer );
}

void  write_prompt( char *msg ) /* 0A7E */
{
	if( promptfile != NULL )
		if( fprintf( promptfile, "%s", msg ) == EOF )
			error( "error writing to prompt file (%s)", promptname );
}

void  openprompt( void ) /* 0ACE */
{
	setname( &promptname, "[stdout]" );
	DEBUG_PRN( "Prompt file name set to '%s'\n", promptname );
#if	0
	setname( &promptname, "CON:" ); /* uh oh, HUGE bug here */
	DEBUG_PRN( "Prompt file name set to '%s'\n", promptname );
	if( ( promptfile = openfile( promptname, "", "", "w+t", NULL ) ) == NULL )
		error( "unable to open %s for prompting", promptname );
#endif
	FREE( nameptr );
	DEBUG_PRN( "Address %ld: free'd memory (openprompt)\n", nameptr );
	promptfile = stdout;
}

int  check_boundaries( long first_start, long first_len, long second_start, long second_len ) /* 0C78 */
{
	if( ( first_start + first_len ) > second_start && ( second_start + second_len ) > first_start )
		return TRUE;
	else
		return FALSE;
}

int  my_abs( int value ) /* 0CBC */
{
	return ( value < 0 ) ? -value : value;
}

int  nth_nibble( long value, int n ) /* 0CE4 */
{
	switch( n ) {
		default:
			value >>= 4;
		case 7:
			value >>= 4;
		case 6:
			value >>= 4;
		case 5:
			value >>= 4;
		case 4:
			value >>= 4;
		case 3:
			value >>= 4;
		case 2:
			value >>= 4;
		case 1:
			value >>= 4;
		case 0:
			return ( value & 0x0F );
/*			break; */
	}
}

int	chartohex( char c ) /* 0D78 */
{
	if( char_table[(int)c] & CHR_DECIMAL )
		return c - '0';
	if( char_table[(int)c] & CHR_LOWCASE )
		return c - ( 'a' - 'A' ) - '7';
	else
		return c - '7';
}

char	*s_83EA_0DB4( char *bp6 ) /* 0DB4 */
{
	char	*bp_4;

	bp_4 = bp6;
	while( *bp_4 != '\0' && *bp_4++ == '\\' )
		bp6 = bp_4;
	return bp6;
}

char	*s_83EA_0DF8( long bp6, struct module *bp10 ) /* 0DF8 */
{
	sprintf( ds_393a, "%s(%.5lX)", s_83EA_0DB4( bp10->filename ), bp6 - bp10->address );
	return ds_393a;
}

char	*s_83EA_0E3E( long bp6, struct module *bp10 ) /* 0E3E */
{
	sprintf( ds_3a3a, "%.5lX %s", bp6, s_83EA_0DF8( bp6, bp10 ) );
	return ds_3a3a;
}

void  fillin_ref( long refaddr, long refvalue, int reflen ) /* 0E7C */
{
	long	fillvalue, filladdr;

	fillvalue = refvalue;
	filladdr = refaddr;
	if( ( filladdr & 1L ) != 0L ) {
		fillnibble( filladdr++, fillvalue );
		--reflen;
		fillvalue >>= 4;
	}
	if( reflen != 0 ) {
		switch( reflen ) {
			case 7:
				fillbyte( filladdr, fillvalue );
				filladdr += 2L;
				fillvalue >>= 8;
			case 5:
				fillbyte( filladdr, fillvalue );
				filladdr += 2L;
				fillvalue >>= 8;
			case 3:
				fillbyte( filladdr, fillvalue );
				filladdr += 2L;
				fillvalue >>= 8;
			case 1:
				fillnibble( filladdr, fillvalue );
				break;
			case 8:
				fillbyte( filladdr, fillvalue );
				filladdr += 2L;
				fillvalue >>= 8;
			case 6:
				fillbyte( filladdr, fillvalue );
				filladdr += 2L;
				fillvalue >>= 8;
			case 4:
				fillbyte( filladdr, fillvalue );
				filladdr += 2L;
				fillvalue >>= 8;
			case 2:
				fillbyte( filladdr, fillvalue );
				break;
			default:
				error( "fillin_ref called with len = %d", reflen );
				break;
		}
	}
}

void  s_83EA_102C( struct symbol *bp6 ) /* 102C */
{
	struct fillref *bp_4 = NULL;
	int bp_6;
	long bp_10 = 0L, bp_14 = 0L, bp_18, bp_22;
	struct fillref *bp_26 = bp6->s_fref_first;
	char	bp_282[256];

	while( bp_26 != NULL ) {
		switch( bp_26->f_class ) {
			case '\x00':
				if( ( bp6->s_flags & 0x8000 ) == 0 ) {
					fillin_ref( bp_26->f_address, d_061a, (long)bp_26->f_length );
					break;
				}
				fillin_ref( bp_26->f_address, bp6->s_value + bp_26->f_bias, (long)bp_26->f_length );
				if( ( list_mode & 8 ) == 0 )
					break;
				if( bp_26->f_subclass != '\x01' )
					break;
				if( bp_26->f_length != 5 )
					break;
				bp_22 = bp6->s_value + bp_26->f_bias - bp_26->f_address;
				if( bp_22 < -0x7FFCL || bp_22 > 0x7FFFL )
					break;
				bp_6 = 1;
				if( bp_22 >= -0x7FEL && bp_22 <= 0x7FEL )
					bp_6 = 3;
				sprintf( errmsgbuf,
							"Shorten(%d) ref to %-12s in %s",
							bp_6, bp6->s_name,
							s_83EA_0DF8( bp_26->f_address, bp_26->f_module ) );
				do_error( errmsgbuf );
				d_061e += (long)bp_6;
				break;
			case '\x01':
				if( ( bp6->s_flags & 0x8000 ) == 0 ) {
					fillin_ref( bp_26->f_address, d_061a, (long)bp_26->f_length );
					break;
				}
				bp_22 = bp6->s_value + bp_26->f_bias - bp_26->f_address;
				if( bp_26->f_subclass == '\x01' )
					bp_22 -= (long)bp_26->f_length;
				fillin_ref( bp_26->f_address, bp_22, (long)bp_26->f_length );
				bp_18 = ( 1L << ( ( ( bp_26->f_length & 0xFF ) << 2 ) - 1 ) ) - 1;
				if( ( bp_22 + 1 ) < -bp_18 || bp_22 > bp_18 ) {
					bp_6 = sprintf( bp_282, "Range Error: %-12s %s in ",
										 bp6->s_name,
										 s_83EA_0DF8( bp6->s_value, bp6->s_module ) );
					sprintf( &bp_282[bp_6], "%s D=%05lX",
								s_83EA_0E3E( bp_26->f_address, bp_26->f_module ),
								bp6->s_value - bp_26->f_address );
					if( bp_26->f_length != 5 )
						errmsg( bp_282 );
					break;
				}
				if( ( list_mode & 8 ) == 0 )
					break;
				if( bp_26->f_length != 4 )
					break;
				if( bp_26->f_subclass != '\x00' ||
					 bp_22 < -0x801 || bp_22 > 0x7FE ) {
					if( bp_26->f_subclass != '\x01' )
						break;
					if( bp_22 < -0x802 || bp_22 > 0x7FD )
						break;
				}
				sprintf( errmsgbuf,
							"Shorten(2) ref to %-12s in %s", bp6->s_name,
							s_83EA_0DF8( bp_26->f_address, bp_26->f_module ) );
				do_error( errmsgbuf );
				d_061e += 2;
				break;

			case '\x02':
				if( bp_26->f_subclass == '\x00' ) {
					if( bp_4 == NULL )
						bp_4 = bp_26;
					else {
						sprintf( errmsgbuf, "Multiple SLINK references to symbol %-12s", bp6->s_name );
						errmsg( errmsgbuf );
					}
				} else {
					fillin_ref( bp_26->f_address, bp_26->f_bias + bp_14, (long) bp_26->f_length );
					bp_14 = bp_26->f_address + (long)bp_26->f_length;
				}
				break;
			case '\x03':
				fillin_ref( bp_26->f_address, bp_26->f_bias + bp_10, (long)bp_26->f_length );
				++bp_10;
				break;
			default:
				sprintf( errmsgbuf,
							"Reference of Unknown Type. Symbol(%-12s) at %.5lX Reference type is %d.%d",
							bp6->s_name,
							bp_26->f_address,
							bp_26->f_class,
							bp_26->f_subclass );
				errmsg( errmsgbuf );
				break;

		}
		bp_26 = bp_26->f_next;
	}
	if( bp_4 != NULL ) {
		fillin_ref( bp_4->f_address, bp_4->f_bias + bp_14, (long)bp_4->f_length );
		return;
	}
	if( bp_14 != 0L ) {
		sprintf( errmsgbuf, "Missing SLINK reference to symbol %-12s", bp6->s_name );
		errmsg( errmsgbuf );
	}
}

void  resolve_symbols( void ) /* 1574 */
{
	struct	symbol	*bp_4;
	char	bp_6;
	char	*bp_10;
	long	bp_14;

	if( d_05fa != 0 ) {
		if( ( bp_4 = s_8A85_0414( d_021e ) ) != NULL && ( bp_4->s_flags & SYM_UNDEFINED ) != 0 )
			d_061a = bp_4->s_value;
		else {
			sprintf( errmsgbuf, "Unresolved symbol '%-12s' not defined", d_021e );
			errmsg( errmsgbuf );
		}
	}
	do_all_symbols( s_83EA_102C );
	if( ( list_mode & LIST_SHORTNREF ) != 0 ) {
		sprintf( errmsgbuf, "Shortening references would save %5ld nibbles", d_061e );
		do_error( errmsgbuf );
	}
	if( ( bp_4 = s_8A85_0414( d_0332 ) ) != NULL && bp_4->s_value >= codestart ) {
		bp_10 = d_0348;
		bp_14 = bp_4->s_value;
		while( ( bp_6 = *bp_10++ ) != '\0' && bp_14 < 0xFFFFFL ) {
			fillnibble( bp_14, nth_nibble( (long)bp_6, 0 ) );
			fillnibble( bp_14 + 1, nth_nibble( (long)bp_6, 1 ) );
			bp_14 += 2L;
		}
		if( d_0348 == date && bp_14 < 0xFFFFDL ) {
			fillnibble( bp_14++, 0 );
			fillnibble( bp_14++, 2 );
			fillnibble( bp_14++, 0 );
			fillnibble( bp_14++, 2 );
		}
	}
}

void  resolve_cksums( struct checksum *bp6 ) /* 178A */
{
	long	bp_4;
	int	bp_6;
	long	bp_10;
	struct	symbol	*bp_14;

	while( bp6 != NULL ) {
		if( ( bp_14 = s_8A85_0414( bp6->ch_name ) ) != NULL ) {
			bp_6 = 0;
			bp_4 = bp_14->s_value;
			if( ( bp_4 + 1 ) > 0x100000L ) {
				sprintf( errmsgbuf, "CHecksum address too large (%s)", bp_14->s_name );
				errmsg( errmsgbuf );
				continue;
			}
			fillnibble( bp_4, 0 );
			fillnibble( bp_4 + 1, 0 );
			bp_10 = bp6->d_13;
			while( bp6->d_17 >= bp_10 ) {
				bp_6 += ( ( bp_10 & 1 ) != 0 ) ? getnibble( bp_10 ) << 4 : getnibble( bp_10 );
				if( bp_6 > 255 )
					bp_6 = ( bp_6 & 0x00FF ) + ( bp_6 >> 8 );
				++bp_10;
			}
			bp_6 = ( bp_6 <= 1 ) ? 1 - bp_6 : ( ~bp_6 & 0x00FF ) + 1;
			/* check this casts once again */
			fillnibble( bp_4, nth_nibble( (long)bp_6, 0 ) );
			fillnibble( bp_4 + 1, nth_nibble( (long)bp_6, 1 ) );
		} else {
			sprintf( errmsgbuf, "Checksum symbol %s not found", bp6->ch_name );
			errmsg( errmsgbuf );
		}
		bp6 = bp6->ch_next;
	}
}

void  resolve_crcs( struct crc *bp6 ) /* 1954 */
{
	struct module	*bp_4;
	long	bp_8, bp_12 = 0L;

	while( bp6 != NULL ) {
		if( bp6->cr_type == 2 || bp6->cr_type == 3 )
			bp_8 = bp6->cr_offset;
		else
			bp_8 = ( ( ( bp6->cr_sectors << 1 ) - 1L ) * bp6->cr_offset ) +
					 ( bp6->cr_reads << 4 ) + bp6->cr_start - 4L;
		if( ( bp_8 + 3L ) > 0x100000L ) {
			errmsg( "CRc address error" );
			continue;
		}
		if( ( bp_8 + 4L ) > codeaddr )
			codeaddr = bp_8 + 4L;
		fillnibble( bp_8, 0x0F );
		fillnibble( bp_8 + 1L, 0x0F );
		fillnibble( bp_8 + 2L, 0x0F );
		fillnibble( bp_8 + 3L, 0x0F );
		switch( bp6->cr_type ) {
			case 0:
				bp_12 = compute_crc( bp6->cr_start, bp6->cr_offset, bp6->cr_reads, bp6->cr_sectors );
				break;
			case 1:
				bp_12 = compute_kcrc( bp6->cr_start, bp6->cr_offset, bp6->cr_reads, bp6->cr_sectors );
				break;
			case 2:
				bp_12 = compute_ecrc( bp6->cr_start, bp_8 );
				break;
			case 3:
				bp_12 = (long)compute_cksum( bp6->cr_start, bp6->cr_offset );
				break;
		}
		bp_4 = module_root;
		while( bp6->cr_type != 3 && bp_4 != NULL ) {
			if( check_boundaries( bp_4->address, bp_4->length, bp_8, 4L ) != 0 ) {
				sprintf( errmsgbuf,
							"Warning:  Module %s overlaps following CRC fill location",
							s_83EA_0DB4( bp_4->filename ) );
				errmsg( errmsgbuf );
			}
			bp_4 = bp_4->next;
		}
		sprintf( errmsgbuf, "Filling 4 nibble CRC check value %04lX at address %.5lX", bp_12, bp_8 );
		do_error( errmsgbuf );
		fillnibble( bp_8, nth_nibble( bp_12, 3 ) );
		fillnibble( bp_8 + 1L, nth_nibble( bp_12, 2 ) );
		fillnibble( bp_8 + 2L, nth_nibble( bp_12, 1 ) );
		fillnibble( bp_8 + 3L, nth_nibble( bp_12, 0 ) );
		bp6 = bp6->cr_next;
	}
}

void  list_summary( void ) /* 1CAE */
{
	int symnum = 0, refnum = 0;
	long codelen = 0L;
	struct module *mp;

	do_error( listver );
	do_error( "" );
	if( objname != NULL ) {
		if( ( option_flags & OPT_SYMBOLS ) != 0 )
			symnum = symbolcount;
		if( ( option_flags & OPT_REFS ) != 0 )
			refnum = refcount;
		if( ( option_flags & OPT_CODE ) != 0 )
			codelen = ( codeaddr > codestart ) ? codeaddr - codestart : 0L;
		sprintf( errmsgbuf, "Output Module:\nModule=%s", objname );
		do_error( errmsgbuf );
		if( codelen != 0L )
			sprintf( errmsgbuf,
						"  Start=%.5lX End=%.5lX Length=%.5lX   Symbols=%4d References=%4d", /* was %4ld */
						codestart, codeaddr - 1L, codelen, symnum, refnum );
		else
			sprintf( errmsgbuf,
						"  Start=%.5lX Module Contains No Code  Symbols=%4d References=%4d", /* was %4ld */
						codestart, symnum, refnum );
		do_error( errmsgbuf );
		sprintf( errmsgbuf, "  Date=%-26s Title=%s", date, title );
		do_error( errmsgbuf );
		if( d_05fa ) {
			sprintf( errmsgbuf, "  Unresolved references filled with value of %s (%ld)", d_021e, d_05f6 );
			do_error( errmsgbuf );
		}
		do_error( "" );
	}
	do_error( "" );
	if( module_root != NULL ) {
		do_error( "Source modules:" );
		mp = module_root;
		while( mp != NULL ) {
			sprintf( errmsgbuf, "Module=%s", mp->filename );
			do_error( errmsgbuf );
			if( mp->error )
				do_error( "  ***** Error while reading this module *****" );
			if( mp->length != 0L )
				sprintf( errmsgbuf,
							"  Start=%.5lX End=%.5lX Length=%.5lX",
							mp->address,
							mp->address + mp->length - 1L,
							mp->length);
			else
				sprintf( errmsgbuf, "  Start=%.5lX Module Contains No Code", mp->address );
			do_error( errmsgbuf );
			sprintf( errmsgbuf, "  Date=%-26s Title=%s", mp->date, mp->title );
			do_error( errmsgbuf );
			do_error( "" );
			mp = mp->next;
		}
	}
}

void  s_83EA_1FC4( struct symbol *bp6 ) /* 1FC4 */
{
	int	i;
	char	msgbuf[256];
	struct fillref	*bp_4;

	if( (int)*bp6->s_name != d_0024 ) {
		do_error( "" );
		d_0024 = (int)*bp6->s_name;
	}
	i = sprintf( msgbuf, "%-12s %c ", bp6->s_name, ( bp6->s_flags & 0x4000 ) ? '*' : '=' );
	if( bp6->s_flags & 0x8000 )
		i += sprintf( msgbuf + i, "%.5lX -", bp6->s_value );
	else
		i += sprintf( msgbuf + i, "      -" );
	bp_4 = bp6->s_fref_first;
	while( bp_4 != NULL ) {
		if( (long)( i + 6 ) > pagewidth ) {
			do_error( msgbuf );
			i = sprintf( msgbuf, "                     +" );
		}
		i += sprintf( msgbuf + i, " %.5lX", bp_4->f_address );
		bp_4 = bp_4->f_next;
	}
	do_error( msgbuf );
}

void  s_83EA_210E( struct symbol *bp6 ) /* 210E */
{
	int 	bp_2, i;
	char	msgbuf[256];

	bp_2 = ( bp6->s_flags & 0x4000 ) ? TRUE : FALSE;
	i = sprintf( msgbuf, "%-12s %c ", bp6->s_name, ( bp_2 ) ? '*' : '=' );
	if( bp6->s_flags & 0x8000 ) {
		i += sprintf( msgbuf + i, "%.5lX ", bp6->s_value );
		if( bp_2 )
			strcpy( msgbuf + i, s_83EA_0DF8( bp6->s_value, bp6->s_module ) );
		else
			strcpy( msgbuf + i, s_83EA_0DB4( bp6->s_module->filename ) );
	}
	do_error( msgbuf );
}

void  s_83EA_21E6( struct symbol *bp6, struct fillref *bp10 ) /* 21E6 */
{
	int	i;
	char	msgbuf[256];

	i = sprintf( msgbuf, "%-15s%-28s Ty/Nib=%d.%d/%d", "",
					 s_83EA_0E3E( bp10->f_address, bp10->f_module ),
					 bp10->f_class, bp10->f_subclass, bp10->f_length );
	if( bp10->f_class == '\x01' )
		i += sprintf( msgbuf + i, " Dist=%05lX",
						  (long)my_abs( bp10->f_address - bp6->s_value ) );
/* check my_abs() for making (long) return !!! */
	if( bp10->f_bias != 0L ) {
		i += sprintf( msgbuf + i, " Offset= %3ld", bp10->f_bias );
	}
	do_error( msgbuf );
}

void  s_83EA_22BE( struct symbol *bp6 ) /* 22BE */
{
	struct fillref	*bp_4;

	bp_4 = bp6->s_fref_first;
	if( *bp6->s_name != d_0024 ) {
		do_error( "" );
		d_0024 = *bp6->s_name;
	}
	s_83EA_210E( bp6 );
	while( bp_4 != NULL ) {
		s_83EA_21E6( bp6, bp_4 );
		bp_4 = bp_4->f_next;
	}
}

void  s_83EA_2344( struct symbol *bp6 ) /* 2344 */
{
	struct fillref	*bp_4;
	int	bp_6;

	bp_4 = bp6->s_fref_first;
	bp_6 = TRUE;
	if( bp6->s_module == d_001c )
		while( bp_4 != NULL ) {
			if( bp_4->f_module == d_03c8 ) {
				if( bp_6 ) {
					d_0024 = *bp6->s_name;
					s_83EA_210E( bp6 );
					bp_6 = FALSE;
				}
				s_83EA_21E6( bp6, bp_4 );
			}
			bp_4 = bp_4->f_next;
		}
}

void  s_83EA_23F6( struct symbol *bp6 ) /* 23F6 */
{
	if( ( bp6->s_flags & SYM_UNDEFINED ) == 0 ) {
		if( d_0024 == -1 ) {
			d_0024 = 0;
			last_line();
			do_error( "Unresolved References" );
			do_error( "" );
		}
		if( (long)( d_0024 + 17 ) > pagewidth ) {
			do_error( errmsgbuf );
			d_0024 = 0;
		}
		d_0024 += sprintf( errmsgbuf + d_0024, "%-12s(%3d)   ", bp6->s_name, bp6->s_flags & SYM_REFSMASK );
	}
}

void  list_xref( void ) /* 24B0 */
{
	last_line();
	do_error( "Saturn Cross Reference Listing" );
	d_0024 = -1;
	do_all_symbols( s_83EA_1FC4 );
	do_error( "" );
}

void  list_lxref( void ) /* 24F0 */
{
	last_line();
	do_error( "Saturn Long Cross Reference Listing" );
	d_0024 = -1;
	do_all_symbols( s_83EA_22BE );
	do_error( "" );
}

void  list_references( void ) /* 2530 */
{
	d_03c8 = module_root;
	while( d_03c8 != NULL ) {
		last_line();
		d_001c = module_root;
		sprintf( errmsgbuf, "Saturn Reference Listing For Module %s", d_03c8->filename );
		do_error( errmsgbuf );
		do_error( "" );
		while( d_001c != NULL ) {
			d_0024 = -1;
			do_all_symbols( s_83EA_2344 );
			if( d_0024 != -1 )
				do_error( "" );
			d_001c = d_001c->next;
		}
		d_03c8 = d_03c8->next;
	}
	do_error( "" );
}

void  list_symbols( void ) /* 2636 */
{
	d_001c = module_root;
	while( d_001c != NULL ) {
		last_line();
		d_03c8 = module_root;
		sprintf( errmsgbuf, "Saturn Symbol Listing For Module %s", d_001c->filename );
		do_error( errmsgbuf );
		do_error( "" );
		while( d_03c8 != NULL ) {
			d_0024 = -1;
			do_all_symbols( s_83EA_2344 );
			if( d_0024 != -1 )
				do_error( "" );
			d_03c8 = d_03c8->next;
		}
		d_001c = d_001c->next;
	}
	do_error( "" );
}

void  list_unresolved( void ) /* 273C */
{
	d_0024 = -1;
	do_all_symbols( s_83EA_23F6 );
	if( d_0024 != -1 ) {
		do_error( errmsgbuf );
		do_error( "" );
	}
}

void  list_code( void ) /* 2786 */
{
	long bp_4;
	int bp_6 = 0;

	last_line();
	do_error( "Saturn Hex Code Listing" );
	bp_4 = codestart & 0xFFFFFFC0L; /* dword */
	while( bp_4 < codeaddr ) {
		if( ( bp_4 & 0x3FL ) == 0 ) { /* byte and */
			errmsgbuf[bp_6] = '\0';
			do_error( errmsgbuf );
			if( ( bp_4 & 0xFFL ) == 0 ) /* byte test */
				do_error( "" );
			bp_6 = sprintf( errmsgbuf, "%.5lX -", bp_4 );
		}
		if( ( bp_4 & 0x07L ) == 0 ) /* byte test */
			errmsgbuf[bp_6++] = ' ';
		errmsgbuf[bp_6++] = ( bp_4 < codestart ) ? 0x78 : hextable[getnibble( bp_4 )];
		++bp_4;
	}
	errmsgbuf[bp_6] = '\0';
	do_error( errmsgbuf );
	do_error( "" );
}

void  write_listing( void ) /* 28AC */
{
	if( listing ) {
		if( ( list_mode & LIST_SUMMARY ) != 0 )
			list_summary();
		if( ( list_mode & LIST_XREF ) != 0 )
			list_xref();
		if( ( list_mode & LIST_LXREF ) != 0 )
			list_lxref();
		if( ( list_mode & LIST_REFERENCES ) != 0 )
			list_references();
		if( ( list_mode & LIST_SYMBOLS ) != 0 )
			list_symbols();
		if( ( list_mode & LIST_UNRESOLVED ) != 0 )
			list_unresolved();
		if( ( list_mode & LIST_CODE ) != 0 )
			list_code();
	}
}

int  getflag( int bp6 ) /* 2908 */
{
	int i = 0;

	while( user_flags[i] != bp6 )
		if( ++i >= 100 )
			return FALSE;
	return TRUE;
}

int  setflag( int bp6 ) /* 2944 */
{
	int	i = 0;

	if( getflag( bp6 ) )
		return TRUE;
	while( user_flags[i] != bp6 && i >= 100 ) {
		if( user_flags[i] == 0 ) {
			user_flags[i] = bp6;
			return TRUE;
		}
		++i;
	}
	return FALSE;
}

int  clearflag( int bp6 ) /* 299C */
{
	int	i = 0;
	do {
		if( user_flags[i] == bp6 ) {
			user_flags[i] = 0;
			return TRUE;
		}
	} while( i < 100 );
	return FALSE;
}

int  s_83EA_29DA( void ) /* 29DA */
{
	int	bp_2;

	bp_2 = d_05ea;
	d_05ea = FALSE;
	return bp_2;
}

/*
 *  ABort
 */
void  do_abort( void ) /* 29F6 */
{
	reading = FALSE;
	aborted = TRUE;
	aborterr = FALSE;
}

/*
 *  ANnotate <Comment>
 */
void  do_annotate( void ) /* 2A10 */
{
	remove_tabs( errmsgbuf, ptr, 256 );
	do_error( errmsgbuf );
}

/*
 *  BOundary <Hex addr> - Start next module on multiple of <addr>
 */
void  do_boundary( void ) /* 2A4A */
{
	long	bp_4 = 0L;

	d_0602 = TRUE;
	if( ( bp_4 = getaddress() ) > 0L )
		d_032e = ( ( bp_4 + codeaddr - 1L ) / bp_4 ) * bp_4;
	else
		d_032e = codeaddr;
}

/*
 *  CHeck    <Symbol name> <Hex addr> <Hex addr>
 */
void  do_check( void ) /* 2AC4 */
{
	struct checksum	**bp_4, *bp_12;
	long	bp_8, bp_16;

	bp_4 = &cksum_root;
	if( getsymbol( d_00ee ) != NULL &&
		 ( bp_16 = getaddress() ) != -1L &&
		 ( bp_8 = getaddress() ) != -1L ) {
		while( *bp_4 != NULL )
			bp_4 = &(*bp_4)->ch_next;
		if( ( *bp_4 = calloc( 1, sizeof( struct checksum ) ) ) == NULL )
			error( "error calloc'ing %ld bytes for new checksum", (long)sizeof( struct checksum ) );
		bp_12 = *bp_4;
		DEBUG_PRN( "Address %ld: allocated %d bytes (do_check)\n", bp_12, sizeof( struct checksum ) );
		strncpy( bp_12->ch_name, d_00ee, 12 );
		bp_12->ch_name[12] = '\0';
		bp_12->d_13 = bp_16;
		bp_12->d_17 = bp_8;
	}
}

/*
 *  ** <Comment>
 */
void  do_comment( void ) /* 2C5E */
{
/* it's empty routine :) */
}

/*
 * parameter 3
 *
 *  CKsum    <Start symbol> <End symbol>
 *     (Kermit type CRC--CRC value stored in 4 nibs before End)
 */
void  do_cksum( int bp6 ) /* 2C66 */
{
	char	bp_14[14];
	struct crc	*bp_18, **bp_42;
	struct symbol	*bp_34, *bp_38;

	bp_42 = &crc_root;
	if( getsymbol( d_00ee ) != NULL && getsymbol( bp_14 ) != NULL ) {
		if( ( bp_34 = s_8A85_0414( d_00ee ) ) != NULL ) {
			if( bp_34->s_flags & 0x8000 ) {
				if( ( bp_38 = s_8A85_0414( bp_14 ) ) != NULL ) {
					if( bp_38->s_flags & 0x8000 ) {
						if( ( bp_38->s_value - 4L ) > bp_34->s_value ) {
							while( *bp_42 != NULL )
								bp_42 = &(*bp_42)->cr_next;
							if( ( bp_18 = calloc( 1, sizeof( struct crc ) ) ) == NULL )
								error( "error calloc',27h,'ing %ld bytes for new crc", sizeof( struct crc ) );
							*bp_42 = bp_18;
							DEBUG_PRN( "Address %ld: allocated %d bytes (do_crc)\n", bp_18, sizeof( struct crc ) );
							bp_18->cr_start = bp_34->s_value;
							bp_18->cr_offset = bp_38->s_value - 4L;
							bp_18->cr_type = bp6;
						} else {
							sprintf( errmsgbuf, "Start (%5lX) >= End (%5lX)", bp_34->s_value, bp_38->s_value );
							errmsg( errmsgbuf );
						}
					} else {
						sprintf( errmsgbuf, "Symbol %s not resolved.", bp_14 );
						errmsg( errmsgbuf );
					}
				} else {
					sprintf( errmsgbuf, "Symbol %s not defined.", bp_14 );
					errmsg( errmsgbuf );
				}
			} else {
				sprintf( errmsgbuf, "Symbol %s not resolved.", d_00ee );
				errmsg( errmsgbuf );
			}
		} else {
			sprintf( errmsgbuf, "Symbol %s not defined.", d_00ee );
			errmsg( errmsgbuf );
		}
	}
}

/*
 * parameter 0
 *
 *  CRc      <Start> <Offset> <Reads/Halfsector> <Sectors>
 *     (Lewis type CRC - 16 nibble reads)
 *
 * parameter 1
 *
 *  KCrc     <Start> <Offset> <Reads/Halfsector> <Sectors>
 *     (KERMIT type CRC - 16 nibble reads)
 *
 * parameter 2
 *
 *  ECrc     <Start> <End>
 *     (Ernie type CRC)
 */
void  do_crc( int crc_type ) /* 2EC4 */
{
	long	crc_start, crc_reads, crc_sectors, crc_offset;
	struct crc	*bp_4, **bp_20;

	bp_20 = &crc_root;
	if( ( crc_start = getaddress() ) != -1L &&
		 ( crc_offset = getaddress() ) != -1L &&
		 ( crc_reads = getaddress() ) != -1L &&
		 ( crc_sectors = getaddress() ) != -1L ) {
		while( *bp_20 != NULL )
			bp_20 = &(*bp_20)->cr_next;
		if( ( bp_4 = calloc( 1, sizeof( struct crc ) ) ) == NULL )
			error( "error calloc'ing %ld bytes for new crc", sizeof( struct crc ) );
		*bp_20 = bp_4;
		DEBUG_PRN( "Address %ld: allocated %d bytes (do_crc)\n", bp_4, sizeof( struct crc ) );
		bp_4->cr_start = crc_start;
		bp_4->cr_offset = crc_offset;
		bp_4->cr_reads = crc_reads;
		bp_4->cr_sectors = crc_sectors;
		bp_4->cr_type = crc_type;
	}
}

void  set_debugfile( char *token ) /* 3080 */
{
	setname( &dbgname, token );
	if( dbgfile != NULL && dbgfile != stderr )
		fclose( dbgfile );
	if( strcmp( dbgname, "-" ) ) {
		if( ( dbgfile = fopen( dbgname, "wt" ) ) == NULL )
			error( "error creating %s for writing", dbgname );
	} else {
		setname( &dbgname, "[stderr]" );
		dbgfile = stderr;
	}
	setbuf( dbgfile, NULL );
	debug = TRUE;
}

/*
 *  DBug	<file name>
 */
void  do_dbug( void ) /* 3184 */
{
	char	*token;

	if( ( token = gettoken() ) == NULL )
		debug = 0;
	else
		set_debugfile( token );
}

/*
 *  DEfine   <Symbol name> <Hex addr>
 */
void  do_define( void ) /* 31BA */
{
	long	bp_4;
	struct symbol	*bp_8;

	if( getsymbol( d_00ee ) != NULL ) {
		if( ( bp_4 = getaddress() ) == -1L )
			return;
		bp_8 = s_8A85_0462( d_00ee );
		if( s_83EA_29DA() || !( bp_8->s_flags & 0x8000 ) ) {
			bp_8->s_value = bp_4;
			bp_8->s_flags |= 0x8000;
			bp_8->s_flags &= 0xBFFF;
			bp_8->s_module = &d_0372;
			return;
		}
		if( bp_8->s_value != bp_4 || bp_8->s_flags & 0x4000 ) {
			sprintf( errmsgbuf, "Symbol %s is already defined", bp_8->s_name );
			errmsg( errmsgbuf );
		}
	} else
		errmsg( "Missing symbol for DEfine" );
}

/*
 *  ENd      [<End Option>] - end commands and execute linking
 *     Either of:
 *       REsolveReferences (default) - resolve all references
 *       NOresolveReferences - don't fill in references
 */
void  do_end( void ) /* 32A6 */
{
	int	cmdvalue;

	reading = FALSE;
	no_resolve = FALSE;
	if( ( cmdvalue = getcommand() ) != 0 ) {
		if( cmdvalue == CMD_END_NORESOLVE ) {
			no_resolve = TRUE;
			return;
		}
		if( cmdvalue != CMD_END_RESOLVE )
			errmsg( "Unrecognized ENd option" );
	}
}

/*
 *  FLag     <FlagName>
 */
void  do_flag( void ) /* 32F0 */
{
	int	n;

	if( ( n = getcommand() ) != 0 )
		if( !setflag( n ) )
			errmsg( "Too many flags" );
}

/*
 *  FOrce
 */
void  do_force( void ) /* 3324 */
{
	d_05ea = TRUE;
}

/*
 *  HElp
 */
void  do_help( void ) /* 3332 */
{
	do_error( listver );
	sprintf( errmsgbuf, "  Maximum output module address is %.5lX", 0x100000L );
	do_error( errmsgbuf );
	do_error( "Legal Commands are:" );
	do_error( "  ABort" );
	do_error( "  ANnotate <Comment>" );
	do_error( "  BOundary <Hex addr> - Start next module on multiple of <addr>" );
	do_error( "  CHeck    <Symbol name> <Hex addr> <Hex addr>" );
	do_error( "  CKsum    <Start symbol> <End symbol>" );
	do_error( "     (Kermit type CRC--CRC value stored in 4 nibs before End)" );
	do_error( "  CRc      <Start> <Offset> <Reads/Halfsector> <Sectors>" );
	do_error( "     (Lewis type CRC - 16 nibble reads)" );
	do_error( "  DBug     <file name>" );
	do_error( "  DEfine   <Symbol name> <Hex addr>" );
	do_error( "  ENd      [<End Option>] - end commands and execute linking" );
	do_error( "     Either of:" );
	do_error( "       REsolveReferences (default) - resolve all references" );
	do_error( "       NOresolveReferences - don't fill in references" );
	do_error( "  ECrc     <Start> <End>" );
	do_error( "     (Ernie type CRC)" );
	do_error( "  ERror    <Text>" );
	do_error( "  FLag     <FlagName>" );
	do_error( "  FOrce" );
	do_error( "  HElp" );
	do_error( "  ID       <Hex number>" );
	do_error( "  IF [NOt] <FlagName> <Command...>" );
	do_error( "  KCrc     <Start> <Offset> <Reads/Halfsector> <Sectors>" );
	do_error( "     (KERMIT type CRC - 16 nibble reads)" );
	do_error( "  LIst     <List Option list>" );
	do_error( "     Any of:" );
	do_error( "       ALl" );
	do_error( "       COde" );
	do_error( "       SYmbols" );
	do_error( "       LXref" );
	do_error( "       REferences" );
	do_error( "       SHortenableReferences" );
	do_error( "       SUmmary" );
	do_error( "       UNresolvedRefs" );
	do_error( "       XRef" );
	do_error( "  LLu      <File spec>" );
	do_error( "  LOcate   <Hex addr>" );
	do_error( "  NExt - print out address at which next module would be loaded" );
	do_error( "  OPtion   <Option list>" );
	do_error( "     Any of:" );
	do_error( "       COde only" );
	do_error( "       SYmbols only" );
	do_error( "       RElocatable (default) - code and symbols" );
	do_error( "       NRefs (for simulator) - code and symbol names only" );
	do_error( "  OUtput   <File spec> - specify object file" );
	do_error( "  PAtch    <Hex Addr> <nibble>..." );
	do_error( "  PUrge    <File spec>" );
	do_error( "  RAnge    <SymbolName> <Hex Addr> <Hex Addr> - check symbol in range" );
	do_error( "  RElocate <File spec> - incorporate code at NExt address" );
	do_error( "  SEarch   <File spec> - read symbols only" );
	do_error( "  SOftkeys <File spec>" );
	do_error( "  STart    <Hex addr> - set first address for output module" );
	do_error( "  SUppress <List Option list>" );
	do_error( "  TItle    <Title text>" );
	do_error( "  TStamp   <Symbol name> [<date string>]" );
	do_error( "  TRansfer <File spec> - read commands from a file" );
	do_error( "  UDefine  <Symbol name> - unresolve symbol" );
	do_error( "  UFlag    <flagname> - unset flagname" );
	do_error( "  UNresolved <Symbol name> - set fill value for unresolved references" );
	do_error( "  URelocate <File spec> - duplicate symbols take precedence" );
	do_error( "  USearch   <File spec> - duplicate symbols take precedence" );
	do_error( "  ZEro     <Hex addr> <Hex addr> - NOT NEEDED!!!!!!" );
	do_error( "  /Abort" );
	do_error( "  /End" );
	do_error( "  ??" );
	do_error( "  ** <Comment>" );
}

/*
 *  ID       <Hex number>
 */
void  do_id( void ) /* 368C */
{
	if( ( romid = getaddress() ) < 0 )
		romid = 0L;
}

/*
 *  IF [NOt] <FlagName> <Command...>
 */
void  do_if( void ) /* 36AC */
{
	int bp_2 = FALSE, cmd, bp_6 = FALSE;

	while( ( cmd = getcommand() ) == CMD_IF_NOT )
		bp_2 = !bp_2;
	bp_6 = getflag( cmd );
	if( bp_2 )
		bp_6 = !bp_6;
	if( bp_6 )
		parse_line( getcommand() );
}

int  s_83EA_370E( void ) /* 370E */
{
	int	flags = LIST_NONE, cmd;

	while( ( cmd = getcommand() ) != 0 ) {
		switch( cmd ) {
			case CMD_LIST_SYMBOLS:
				flags |= LIST_SYMBOLS;
				break;
			case CMD_LIST_LXREF:
				flags |= LIST_LXREF;
				break;
			case CMD_LIST_REFERENCES:
				flags |= LIST_REFERENCES;
				break;
			case CMD_LIST_SHORTNREF:
				flags |= LIST_SHORTNREF;
				break;
			case CMD_LIST_SUMMARY:
				flags |= LIST_SUMMARY;
				break;
			case CMD_LIST_CODE:
				flags |= LIST_CODE;
				break;
			case CMD_LIST_UNRESOLVED:
				flags |= LIST_UNRESOLVED;
				break;
			case CMD_LIST_XREF:
				flags |= LIST_XREF;
				break;
			case CMD_LIST_ALL:
				flags |= LIST_ALL;
				break;
			default:
				errmsg( "Unrecognized listing type" );
				break;
		}
	}
	return flags;
}

/*
 *  LIst     <List Option list>
 *     Any of:
 *       ALl
 *       COde
 *       SYmbols
 *       LXref
 *       REferences
 *       SHortenableReferences
 *       SUmmary
 *       UNresolvedRefs
 *       XRef
 */
void  do_list( void ) /* 37A8 */
{
	list_mode |= s_83EA_370E();
}

/*
 *  LLu      <File spec>
 */
void  do_llu( void ) /* 37B8 */
{
	char	*token;

	if( listing )
		fclose( listfile );
	setname( &listname, ( ( token = gettoken() ) == NULL ) ? "-" : token );
	if( strcmp( listname, "-" ) ) {
		if( ( listfile = fopen( listname, "wt" ) ) == NULL )
			error( "unable to open %s as list file", listname );
	} else {
		listfile = stdout;
		setname( &listname, "[stdout]" );
	}
	DEBUG_PRN( "List file name set to '%s'\n", listname );
}

/*
 *  LOcate   <Hex addr>
 */
void  do_locate( void ) /* 3918 */
{
	if( ( d_032e = getaddress() ) < 0L )
		d_032e = codeaddr;
	d_0602 = TRUE;
}

/*
 *  NExt - print out address at which next module would be loaded
 */
void  do_next( void ) /* 394C */
{
	sprintf( errmsgbuf, "Next available location is %.5lX", codeaddr );
	do_error( errmsgbuf );
}

/*
 *  OPtion   <Option list>
 *     Any of:
 *       COde only
 *       SYmbols only
 *       RElocatable (default) - code and symbols
 *       NRefs (for simulator) - code and symbol names only
 */
void  do_option( void ) /* 3980 */
{
	if( !s_83EA_29DA() && option_flags != 0 )
		errmsg( "Duplicate OPtion command" );
	else
		switch( getcommand() ) {
			case CMD_OPT_CODE:
				option_flags = 1;
				return;
			case CMD_OPT_SYMBOLS:
				option_flags = 2;
				return;
			case CMD_OPT_RELOCATABLE:
				option_flags = 7;
				return;
			case CMD_OPT_NREFS:
				option_flags = 3;
				return;
			default:
				errmsg( "Unrecognized OPtion" );
		}
}

/*
 *  OUtput   <File spec> - specify object file
 */
void  do_output( void ) /* 39E6 */
{
	char	*token;

	if( ( token = gettoken() ) != NULL )
		setname( &objname, token );
	else
		errmsg( "Missing object file name" );
}

/*
 *  PAtch    <Hex Addr> <nibble>...
 */
void  do_patch( void ) /* 3A26 */
{
	char	bp_2;
	int	bp_4 = 0;
	char	bp_22[17];
	long	bp_26;

	if( ( bp_26 = getaddress() ) >= 0 ) {
		bp_4 = 0;
		do {
			bp_22[bp_4] = hextable[getnibble( bp_4 + bp_26 )];
			++bp_4;
		} while( bp_4 < 16 );
		bp_22[16] = '\0';
	}
	sprintf( errmsgbuf, "%.5lX - %s", bp_26, bp_22 );
	do_error( errmsgbuf );
	skipwhite();
	bp_2 = get_chr();
	while( char_table[(int)bp_2] & CHR_HEX ) {
		if( bp_26 >= 0x100000L ) {
			sprintf( errmsgbuf, "PAtch address too large (%.5lX)", bp_26 );
			errmsg( errmsgbuf );
			return;
		}
		fillnibble( bp_26++, chartohex( bp_2 ) );
		bp_2 = get_chr();
	}
}

/*
 *  PUrge    <File spec>
 */
void  do_purge( void ) /* 3B18 */
{
	char	*token;

	if( ( token = gettoken() ) != NULL ) {
		if( unlink( token ) == EOF )
			errmsg( "Unable to purge file" );
	} else
		errmsg( "Missing purge file name" );
}

/*
 *  RAnge    <SymbolName> <Hex Addr> <Hex Addr> - check symbol in range
 */
void  do_range( void ) /* 3B56 */
{
	struct symbol	*bp_4;
	long	bp_8, bp_12;

	if( getsymbol( d_00ee ) != NULL &&
		 ( bp_12 = getaddress() ) != -1 &&
		 ( bp_8 = getaddress() ) != -1 ) {
		if( ( bp_4 = s_8A85_0414( d_00ee ) ) != NULL ) {
			if( bp_4->s_flags & 0x8000 ) {
				if( bp_4->s_value < bp_12 || bp_4->s_value > bp_8 ) {
					sprintf( errmsgbuf,
								"Symbol %s at %s not in range [%.5lX...%.5lX]",
								bp_4->s_name,
								s_83EA_0E3E( bp_4->s_value, bp_4->s_module ),
/* ???
 * here is one PUSH extra,
 * but I don't know where it belongs ?
 * ( bp_4->s_value - bp_4->s_module->length )
 */
								bp_12, bp_8 );
					errmsg( errmsgbuf );
				}
			} else {
				sprintf( errmsgbuf, "Unable to check range of symbol %s (not resolved)", bp_4->s_name );
				errmsg( errmsgbuf );
			}
		} else {
			sprintf( errmsgbuf, "Unable to check range of symbol %s", d_00ee );
			errmsg( errmsgbuf );
		}
	}
}

union	SFC	*s_83EA_3CA2( int bp6 ) /* 3CA2 */
{
	char	*bp_4;

	if( bp6 < 0 ) {
		d_3b3a = 0;
		d_5b3e = 0;
		d_3b3c = 0;
		return NULL;
	}
	d_3b3a += bp6;
	d_3b3c += d_3b3a / 14;
	d_3b3a = d_3b3a % 14;
	while( d_3b3c >= d_5b3e ) {
		d_3b3c -= d_5b3e;
		if( ( d_5b3e = fread( d_3b3e, 256, 32, modulefile ) ) == 0 ) {
			sprintf( errmsgbuf, "Problem reading symbol block in file (%s)", nameptr );
			errmsg( errmsgbuf );
			return NULL;
		}
		if( strncmp( d_3b3e, "Symb", 4 ) ) {
			sprintf( errmsgbuf, "Corrupt symbol table in file (%s)", nameptr );
			errmsg( errmsgbuf );
			return NULL;
		}
	}
	bp_4 = ( 18 * d_3b3a ) + ( d_3b3c * 256 ) + d_3b3e + 4;
	++d_3b3a;
	return (union SFC *)bp_4;
}

void  read_symbols( int bp6, int bp8, int bp10 ) /* 3DAC */
{
	int	bp_6, bp_18;
	unsigned	int	bp_8 = 0;
	struct symbol	*bp_12 = NULL;
	union	SFC	*bp_4, *bp_16;

	s_83EA_3CA2( -1 );
	while( bp6-- > 0 ) {
		if( ( bp_16 = s_83EA_3CA2( bp_8 ) ) != NULL ) {
			bp_8 = get_u16( (unsigned char *)bp_16->s.os_fillcount );
			bp_18 = ( bp_8 & SYM_UNDEFINED ) ? TRUE : FALSE; /* cmp ax,1; sbb ax,ax; inc ax */
			bp_6 = ( bp_8 & SYM_RELOCATABLE ) ? TRUE : FALSE; /* cmp ax,1; sbb ax,ax; inc ax */
			bp_8 &= SYM_REFSMASK;
			if( !bp8 || bp_18 ) {
				bp_12 = s_8A85_0462( bp_16->s.os_name );
				if( bp_18 ) {
					if( ( ( bp_12->s_flags & SYM_UNDEFINED ) != 0 &&
						 ( bp_12->s_flags & SYM_RELOCATABLE ) != 0 ) ||
						 bp_12->s_value == get_32( (unsigned char *)bp_16->s.os_value ) ) {
						sprintf( errmsgbuf,
									"Duplicate entry %-12s in modules %s and %s",
									bp_12->s_name,
									s_83EA_0DB4( bp_12->s_module->filename ),
									s_83EA_0DB4( module_last->filename ) );
						errmsg( errmsgbuf );
					}
					if( ( bp_12->s_flags & SYM_UNDEFINED ) == 0 || bp10 ) {
						bp_12->s_value = get_32( (unsigned char *)bp_16->s.os_value );
						bp_12->s_flags |= SYM_UNDEFINED;
						if( bp_6 ) {
							bp_12->s_value += module_last->address;
							bp_12->s_flags |= SYM_RELOCATABLE;
						} else
							bp_12->s_flags &= ~SYM_RELOCATABLE;
						bp_12->s_module = module_last;
					}
				}
				if( !bp8 ) {
					while( bp_8-- != 0 )
						if( ( bp_4 = s_83EA_3CA2( 0 ) ) != NULL )
							addfillref( bp_12, bp_4 );
					++bp_8;
				}
			}
		}
	}
}

struct module	*newmodule( void ) /* 3F9C */
{
	struct module	**dummy;

	dummy = &module_root;
	while( *dummy != NULL )
		dummy = &(*dummy)->next;
	if( ( *dummy = (struct module *)calloc( 1, sizeof( struct module ) ) ) == NULL )
		error( "error calloc'ing %ld bytes for new module", (long)sizeof( struct module ) );
	DEBUG_PRN( "Address %ld: allocated %d bytes (newmodule)\n", *dummy, sizeof( struct module ) );
	module_last = *dummy;
	module_last->error = TRUE;
	return module_last;
}

void  s_83EA_40C4( int bp6, int bp8 ) /* 40C4 */
{
	char	*token, *strp;
	int	bp_8, bp_14;
	long	bp_274;
	unsigned int	bp_6;
	unsigned long	bp_282, bp_286;
	struct module	*bp_4, *bp_278;
	union {
		char		buffer[256];
		struct	SATURN	saturn;
	} headerin; /* bp_270 */

	if( ( token = gettoken() ) == NULL ) {
		errmsg( "Missing file specifier" );
		return;
	}
	setname( &modulename, token );
	if( ( modulefile = openfile( modulename, "SLOAD_OBJECT", "", "rb", NULL ) ) == NULL ) {
		sprintf( errmsgbuf, "Problem reading file %s", modulename );
		errmsg( errmsgbuf );
		return;
	}
	bp_278 = newmodule();
	if( full_filenames )
		bp_278->filename = nameptr;
	else {
		strp = nameptr;
		while( strlen( strp ) != strcspn( strp, "\\/" ) )
			strp += strcspn( strp, "\\/" ) + 1;
		bp_278->filename = strp;
	}
	if( fread( headerin.buffer, 256, 1, modulefile ) == 1 &&
		 !strncmp( headerin.saturn.id, "Saturn", 6 ) ) {
		bp_6 = get_u16( (unsigned char *)headerin.saturn.symbols );
		bp_286 = get_u32( (unsigned char *)headerin.saturn.nibbles );
		bp_282 = get_u32( (unsigned char *)headerin.saturn.start );
		bp_14 = ( bp_286 != 0L && !bp6 ) ? TRUE : FALSE;
		bp_8 = !(*headerin.saturn.absolute); /* cmp ax,1; sbb ax,ax; neg ax */
		if( bp_8 ) {
			bp_274 = ( d_0602 ) ? d_032e : codeaddr;
		} else {
			bp_274 = bp_282;
			if( d_0602 ) {
				if( s_83EA_29DA() == 0 && d_032e != bp_282 ) {
					sprintf( errmsgbuf, "Unable to locate as specified.  File=%s", s_83EA_0DB4( bp_278->filename ) );
					errmsg( errmsgbuf );
					fclose( modulefile );
					return;
				}
				bp_274 = d_032e;
			}
		}
		bp_278->address = bp_274;
		bp_278->length = ( bp_14 ) ? bp_286 : 0L;
		strncpy( bp_278->title, headerin.saturn.title, 40 );
		strncpy( bp_278->date, headerin.saturn.date, 26 );
		d_0602 = FALSE;
		if( bp_14 && bp_274 < codestart ) {
			sprintf( errmsgbuf, "Start address error.  File=%s", s_83EA_0DB4( bp_278->filename ) );
			errmsg( errmsgbuf );
			fclose( modulefile );
			return;
		}
		if( bp_14 && !bp_8 )
			absolute = FALSE;
		if( bp_14 ) {
			bp_4 = module_root;
			while( bp_4 != bp_278 ) {
				if( check_boundaries( bp_4->address, bp_4->length, bp_278->address, bp_278->length ) != 0 ) {
					sprintf( errmsgbuf, "Warning:  Modules %s and %s are not disjoint", s_83EA_0DB4( bp_4->filename ), s_83EA_0DB4( bp_278->filename ) );
					errmsg( errmsgbuf );
				}
				bp_4 = bp_4->next;
			}
			if( ( bp_274 + bp_286 - 1 ) >= 0x100000L ) {
				sprintf( errmsgbuf, "No room for code.  File=%s", s_83EA_0DB4( bp_278->filename ) );
				errmsg( errmsgbuf );
				fclose( modulefile );
				return;
			}
			if( ( bp_274 + bp_286 ) > codeaddr )
				codeaddr = bp_274 + bp_286;
			d_05f4 = TRUE;
		}
		if( !bp6 ) {
			if( fseek( modulefile, 256L, SEEK_SET ) == 0 ) {
				read_code( bp_274, bp_286 );
			} else {
				sprintf( errmsgbuf, "Unable to seek to start of code.  File=%s", s_83EA_0DB4( bp_278->filename ) );
				errmsg( errmsgbuf );
				fclose( modulefile );
				return;
			}
		}
		if( fseek( modulefile, (long)( ( ( ( bp_286 + 511 ) / 512 ) + 1 ) * 256 ), SEEK_SET ) == 0 ) {
			read_symbols( bp_6, bp6, bp8 );
			bp_278->error = FALSE;
		} else {
			sprintf( errmsgbuf, "Unable to seek to start of symbols.  File=%s", s_83EA_0DB4( bp_278->filename ) );
			errmsg( errmsgbuf );
		}
	} else {
		sprintf( errmsgbuf, "File is not a Saturn object code file.  File=%s", s_83EA_0DB4( bp_278->filename ) );
		errmsg( errmsgbuf );
	}
	fclose( modulefile );
}

/*
 *  RElocate <File spec> - incorporate code at NExt address
 */
void  do_relocate( void ) /* 45AE */
{
	s_83EA_40C4( FALSE, update );
}

/*
 *  SEarch   <File spec> - read symbols only
 */
void  do_search( void ) /* 45C4 */
{
	s_83EA_40C4( TRUE, update );
}

/*
 *  URelocate <File spec> - duplicate symbols take precedence
 */
void  do_urelocate( void ) /* 45DC */
{
	s_83EA_40C4( FALSE, TRUE );
}

/*
 *  USearch   <File spec> - duplicate symbols take precedence
 */
void  do_usearch( void ) /* 45F2 */
{
	s_83EA_40C4( TRUE, TRUE );
}

/*
 *  SOftkeys <File spec>
 */
void  do_softkeys( void ) /* 4606 */
{
	if( s_83EA_29DA() || softkeys[0] == '\0' )
		strncpy( softkeys, ptr, 20 );
	else
		errmsg( "Duplicate SOftkeys command" );
}

/*
 *  STart    <Hex addr> - set first address for output module
 */
void  do_start( void ) /* 4652 */
{
	long	bp_4;

	if( s_83EA_29DA() || !d_05f4 ) {
		if( ( bp_4 = getaddress() ) >= 0 )
			codestart = bp_4;
		codeaddr = codestart;
		absolute = FALSE;
		d_05f4 = TRUE;
	} else
		errmsg( "Duplicate STart command" );
}

/*
 *  SUppress <List Option list>
 */
void  do_suppress( void ) /* 46B0 */
{
	list_mode &= ~s_83EA_370E();
}

/*
 *  TItle    <Title text>
 */
void  do_title( void ) /* 46C2 */
{
	if( s_83EA_29DA() != 0 || *title == '\0' )
		strncpy( title, ptr, 40 );
	else
		errmsg( "Duplicate TItle command" );
}

/*
 *  TRansfer <File spec> - read commands from a file
 */
void  do_transfer( void ) /* 470E */
{
	char	*token;

	if( ( token = gettoken() ) != NULL && strcmp( token, "-" ) ) {
		setname( &inname, token );
		if( ( infile = freopen( inname, "rt", stdin ) ) == NULL )
			error( "Unable to open %s as input file", inname );
	} else {
		setname( &inname, "[stdin]" );
		infile = stdin;
	}
	DEBUG_PRN( "Input file name set to '%s'\n", inname );
	if( prompt_mode ) {
		if( promptfile != stdout ) {
			fclose( promptfile );
			DEBUG_PRN( "Closing prompt file\n" );
		}
		prompt_mode = FALSE;
	}
	if( isatty( fileno( stdin ) ) != 0 ) {
		prompt_mode = TRUE;
		openprompt();
	}
	d_2be7 = 0;
}

/*
 *  TStamp   <Symbol name> [<date string>]
 */
void  do_tstamp( void ) /* 490A */
{
	if( !getsymbol( d_0332 ) )
		errmsg( "Missing symbol for TStamp" );
}

/*
 *  UDefine  <Symbol name> - unresolve symbol
 */
void  do_udefine( void ) /* 493A */
{
	struct symbol	*bp_4;

	if( getsymbol( d_00ee ) )
		if( ( bp_4 = s_8A85_0414( d_00ee ) ) != NULL )
			bp_4->s_flags &= 0x7FFF;
	else
		errmsg( "Missing symbol for UDefine" );
}

/*
 *  UFlag    <flagname> - unset flagname
 */
void  do_uflag( void ) /* 4990 */
{
	int	n;

	if( ( n = getcommand() ) != 0 )
		if( !clearflag( n ) )
			errmsg( "No flags" );
}

/*
 *  UNresolved <Symbol name> - set fill value for unresolved references
 */
void  do_unresolved( void ) /* 49C4 */
{
	d_05fa = FALSE;
	if( !getsymbol( d_021e ) )
		errmsg( "Missing symbol for UNresolved" );
	else
		d_05fa = TRUE;
}

/*
 *  ZEro     <Hex addr> <Hex addr>
 */
void  do_zero( void ) /* 49FA */
{
	long	zero_start, zero_end;

	if( ( zero_start = getaddress() ) >= 0L )
		if( ( zero_end = getaddress() ) >= 0L )
			while( zero_start <= zero_end )
				fillnibble( zero_start++, 0 );
}

void	parse_line( int cmdvalue ) /* 4A54 */
{
	switch( cmdvalue ) {

		case CMD_ABORT:
			/*
			 *  ABort
			 */
			do_abort();
			break;

		case CMD_ANNOTATE:
			/*
			 *  ANnotate <Comment>
			 */
			do_annotate();
			break;

		case CMD_BOUNDARY:
			/*
			 *  BOundary <Hex addr> - Start next module on multiple of <addr>
			 */
			do_boundary();
			break;

		case CMD_CHECK:
			/*
			 *  CHeck    <Symbol name> <Hex addr> <Hex addr>
			 */
			do_check();
			break;

		case CMD_CKSUM:
			/*
			 *  CKsum    <Start symbol> <End symbol>
			 *     (Kermit type CRC--CRC value stored in 4 nibs before End)
			 */
			do_cksum( 3 );
			break;

		case CMD_CRC:
			/*
			 *  CRc      <Start> <Offset> <Reads/Halfsector> <Sectors>
			 *     (Lewis type CRC - 16 nibble reads)
			 */
			do_crc( 0 );
			break;

		case CMD_DBUG:
			/*
			 *  DBug	<file name>
			 */
			do_dbug();
			break;

		case CMD_DEFINE:
			/*
			 *  DEfine   <Symbol name> <Hex addr>
			 */
			do_define();
			break;

		case CMD_END:
			/*
			 *  ENd      [<End Option>] - end commands and execute linking
			 *     Either of:
			 *       REsolveReferences (default) - resolve all references
			 *       NOresolveReferences - don't fill in references
			 */
			do_end();
			break;

		case CMD_ECRC:
			/*
			 *  ECrc     <Start> <End>
			 *     (Ernie type CRC)
			 */
			do_crc( 2 );
			break;

		case CMD_ERROR:
			/*
			 *  ERror    <Text>
			 */
			sprintf( errmsgbuf, "ERror: %s", line );
			do_error( errmsgbuf );
			break;

		case CMD_FLAG:
			/*
			 *  FLag     <FlagName>
			 */
			do_flag();
			break;

		case CMD_FORCE:
			/*
			 *  FOrce
			 */
			do_force();
			break;

		case CMD_HELP:
			/*
			 *  HElp
			 */
			do_help();
			break;

		case CMD_ID:
			/*
			 *  ID       <Hex number>
			 */
			do_id();
			break;

		case CMD_IF:
			/*
			 *  IF [NOt] <FlagName> <Command...>
			 */
			do_if();
			break;

		case CMD_KCRC:
			/*
			 *  KCrc     <Start> <Offset> <Reads/Halfsector> <Sectors>
			 *     (KERMIT type CRC - 16 nibble reads)
			 */
			do_crc( 1 );
			break;

		case CMD_LIST:
			/*
			 *  LIst     <List Option list>
			 *     Any of:
			 *       ALl
			 *       COde
			 *       SYmbols
			 *       LXref
			 *       REferences
			 *       SHortenableReferences
			 *       SUmmary
			 *       UNresolvedRefs
			 *       XRef
			 */
			do_list();
			break;

		case CMD_LLU:
			/*
			 *  LLu      <File spec>
			 */
			do_llu();
			break;

		case CMD_LOCATE:
			/*
			 *  LOcate   <Hex addr>
			 */
			do_locate();
			break;

		case CMD_NEXT:
			/*
			 *  NExt - print out address at which next module would be loaded
			 */
			do_next();
			break;

		case CMD_OPTION:
			/*
			 *  OPtion   <Option list>
			 *     Any of:
			 *       COde only
			 *       SYmbols only
			 *       RElocatable (default) - code and symbols
			 *       NRefs (for simulator) - code and symbol names only
			 */
			do_option();
			break;

		case CMD_OUTPUT:
			/*
			 *  OUtput   <File spec> - specify object file
			 */
			do_output();
			break;

		case CMD_PATCH:
			/*
			 *  PAtch    <Hex Addr> <nibble>...
			 */
			do_patch();
			break;

		case CMD_PURGE:
			/*
			 *  PUrge    <File spec>
			 */
			do_purge();
			break;

		case CMD_RANGE:
			/*
			 *  RAnge    <SymbolName> <Hex Addr> <Hex Addr> - check symbol in range
			 */
			do_range();
			break;

		case CMD_RELOCATE:
			/*
			 *  RElocate <File spec> - incorporate code at NExt address
			 */
			do_relocate();
			break;

		case CMD_SEARCH:
			/*
			 *  SEarch   <File spec> - read symbols only
			 */
			do_search();
			break;

		case CMD_SOFTKEYS:
			/*
			 *  SOftkeys <File spec>
			 */
			do_softkeys();
			break;

		case CMD_START:
			/*
			 *  STart    <Hex addr> - set first address for output module
			 */
			do_start();
			break;

		case CMD_SUPPRESS:
			/*
			 *  SUppress <List Option list>
			 */
			do_suppress();
			break;

		case CMD_TITLE:
			/*
			 *  TItle    <Title text>
			 */
			do_title();
			break;

		case CMD_TRANSFER:
			/*
			 *  TRansfer <File spec> - read commands from a file
			 */
			do_transfer();
			break;

		case CMD_TSTAMP:
			/*
			 *  TStamp   <Symbol name> [<date string>]
			 */
			do_tstamp();
			break;

		case CMD_UDEFINE:
			/*
			 *  UDefine  <Symbol name> - unresolve symbol
			 */
			do_udefine();
			break;

		case CMD_UFLAG:
			/*
			 *  UFlag    <flagname> - unset flagname
			 */
			do_uflag();
			break;

		case CMD_UNRESOLVED:
			/*
			 *  UNresolved <Symbol name> - set fill value for unresolved references
			 */
			do_unresolved();
			break;

		case CMD_URELOCATE:
			/*
			 *  URelocate <File spec> - duplicate symbols take precedence
			 */
			do_urelocate();
			break;

		case CMD_USEARCH:
			/*
			 *  USearch   <File spec> - duplicate symbols take precedence
			 */
			do_usearch();
			break;

		case CMD_ZERO:
			/*
			 *  ZEro     <Hex addr> <Hex addr>
			 */
			do_zero();
			break;

		case 0x2F41:
			/*
			 *  /Abort
			 */
			do_abort();
			break;

		case 0x2F45:
			/*
			 *  /End
			 */
			do_end();
			break;

		case 0x3F3F:
			/*
			 *  ??
			 */
			do_help();
			break;

		case CMD_COMMENT:
		case 0x0:
			/*
			 *  ** <Comment>
			 */
			do_comment();
			return;

		default:
/* original
 *			errmsg( "Unrecognized command" );
 */
			sprintf( errmsgbuf, "Unrecognized command : '%s'", line );
			errmsg( errmsgbuf );
			return;

	}
}

void  init_sload( void ) /* 4D2E */
{
	char	*current_date, *promptptr;
	time_t	clock;

	if( time( &clock ) == -1 )
		error( "%s", "unable to read current time" );
	current_date = ctime( &clock );
	d_0348 = strncpy( date, current_date, 26 );
	date[24] = '\0';
	s_8A85_000A();
	strncpy( d_0332, "@TIME@", 13 );
	strncpy( d_021e, "", 13 );
	setname( &inname, "-" );
/*	DEBUG_PRN( "Input file name set to '%s'\n", inname ); this is redundant */
	infile = stdin;
	setname( &listname, "-" );
/*	DEBUG_PRN( "List file name set to '%s'\n", listname ); this is redundant */
	listfile = stdout;
	setname( &msgname, "-" );
/*	DEBUG_PRN( "Message file name set to '%s'\n", msgname ); this is redundant */
	msgfile = stderr;
	prompt_string = "/SLOAD:  ";
	if( ( promptptr = getenv( "SLOAD_PROMPT" ) ) != NULL ) {
		if( ( prompt_string = calloc( 1, strlen( promptptr ) + 1 ) ) == NULL )
			error( "error calloc'ing memory for prompt string (%s)", promptptr );
/*		DEBUG_PRN( "Address %ld: allocated %d bytes (prompt)\n", prompt_string, strlen( promptptr ) + 1 );
 this is redundant
 */
		strcpy( prompt_string, promptptr );
	}
	d_0372.filename = "Defined";
	d_0372.address = 0L;
	d_0372.length = 0L;
	strcpy( d_0372.title, "" );
	strcpy( d_0372.date, date );
	d_0372.error = FALSE;
	d_0372.next = NULL;
#if	defined(__FLAT_MEM__)
	memory = getblock();
#endif

}

void	get_options( int argc, char **argv, char **env ) /* 517E */
{
	char	i;

	while( ( i = getopt( argc, argv, "D:d:ef:Hhl:M:m:Nno:p:Puvw:" ) ) != (char)(-1) )
		switch( i ) {
			case 'D':
				/* <symb>[=<addr>]
				 * Defines <symb> to be a value (default=1).
				 * For example, -D~Fred=1000 would set the value for symbol
				 * Fred to 1000 decimal.
				 */
				s_8A85_04DE( optarg );
				break;
			case 'd':
				/* <filename>
				 * Defines new debug filename
				 */
#if	defined(__DEBUG__)
				set_debugfile( optarg );
#else
				errmsg( "Not compiled with DEBUG on !" );
#endif
				break;
			case 'e':
				d_0622 = TRUE;
				break;
			case 'f':
				/* Allows setting conditional load flags
				 * (only the first two letters of each flag are significant)
				 */
				ptr = optarg;
				if( setflag( getcommand() ) == 0 )
					errmsg( "Too many flags" );
				break;
			case 'H':
				/* Force SLOAD to code only, with no symbols or references. */
				codeonly = TRUE;
				hexcode = FALSE;
				break;
			case 'h':
				codeonly = TRUE;
				hexcode = TRUE;
				break;
			case 'l':
				/* A listing file is specified. If this option is not used,
				 * then stdout will be the listing file
				 */
				setname( &listname, optarg );
				DEBUG_PRN( "List file name set to '%s'\n", optarg );
				listing = TRUE;
				break;
			case 'M':
			case 'm':
				/* Specifies a file other than stdout for messages
				 * (not the listing file)
				 */
				setname( &msgname, optarg );
				DEBUG_PRN( "Message file name set to '%s'\n", optarg );
				msg_append = ( i == 'M' ) ? TRUE : FALSE;
				break;
			case 'N':
				listing = FALSE;
				break;
			case 'n':
				full_filenames = TRUE;
				break;
			case 'o':
				/* Write the object file to <outfile> */
				setname( &objname, optarg );
				DEBUG_PRN( "Object file name set to '%s'\n", optarg );
				break;
			case 'p':
				/* Specifies pagelength for the listing file (default 60) */
				if( sscanf( optarg, "%ld", &pagelength ) != 1 ) {
					sprintf( errmsgbuf, "Non-numeric page length '%s'", optarg );
					errmsg( errmsgbuf );
					pagelength = 60; /* dword */
				}
				if( pagelength < 4L ) {
					sprintf( errmsgbuf, "Page length must be at least 4 (%ld)", pagelength );
					errmsg( errmsgbuf );
					pagelength = 60; /* dword */
				}
				break;
			case 'P':
				prompt_mode = TRUE;
				break;
			case 'u':
				update = TRUE;
				break;
			case 'v':
				usage( TRUE );
				exit( 0 );
				break;
			case 'w':
				/* Specifies pagewidth for the listing file (default 80) */
				if( sscanf( optarg, "%ld", &pagewidth ) != 1 || pagewidth < 30L ) {
					sprintf( errmsgbuf, "Invalid page width '%s' (minimum width is 30)", optarg );
					errmsg( errmsgbuf );
					pagewidth = 80; /* dword */
				}
				break;
			case '?':
			default:
				++errorcount;
				break;
		}
	if( ( optind + 1 ) == argc ) {
		setname( &inname, argv[optind] );
		DEBUG_PRN( "Input file name set to '%s'\n", argv[optind] );
	}
}

void	open_files( void ) /* 573A */
{
	if( strcmp( inname, "-" ) ) {
		/* beware of that 'stdin' stuff !! */
		if( ( infile = freopen( inname, "rt", stdin ) ) == NULL )
			error( "unable to open %s as input file", inname );
	} else {
		setname( &inname, "[stdin]" );
		DEBUG_PRN( "Input file name set to '%s'\n", inname );
	}
	if( strcmp( msgname, "-" ) ) {
		if( ( msgfile = fopen( msgname, ( msg_append ) ? "at" : "wt" ) ) == NULL )
			error( "unable to open %s as message file", msgname );
	} else {
		msgfile = stderr;
		setname( &msgname, "[stderr]" );
		DEBUG_PRN( "Message file name set to '%s'\n", msgname );
	}
	if( listing ) {
		if( strcmp( listname, "-" ) ) {
			if( ( listfile = fopen( listname, "wt" ) ) == NULL )
				error( "unable to open %s as list file", listname );
		} else {
			listfile = stdout;
			setname( &listname, "[stdout]" );
			DEBUG_PRN( "List file name set to '%s'\n", listname );
		}
	}
	if( isatty( fileno( stdin ) ) != 0 )
		prompt_mode = TRUE;
	d_2be7 = 0;
	if( prompt_mode )
		openprompt();
}

void	main( int argc, char **argv, char **env ) /* 5A92 */
{
	progname = argv[0];
	init_sload();
	get_options( argc, argv, env );
	if( errorcount != 0 ) { /* dword */
		usage( FALSE );
		exit( -1 );
	}
	open_files();
	while( reading ) {
		if( prompt_mode )
			write_prompt( prompt_string );
		if(( ptr = getline( line, stdin )) != NULL ) {
			parse_line( getcommand() );
		} else {
			ptr = "";
			do_end();
		}
	}
	if( aborted ) {
		sprintf( errmsgbuf, "Saturn Loader Aborted (%ld)", (long)aborterr );
		do_error( errmsgbuf );
		exit( aborterr );
	}
	if( !no_resolve )
		resolve_symbols();
	resolve_cksums( cksum_root );
	resolve_crcs( crc_root );
	if( objname != NULL )
		write_objfile();
	write_listing();
	sprintf( errmsgbuf, "%sEnd of Saturn Loader Execution", prompt_string );
	do_error( errmsgbuf );
	exit( errorcount );
}


#if	0

FILE	*d_0000; 		/* promptfile */
long	d_0004; 			/* lineofpage */
char	d_0008[20];		/* softkeys */
struct module	*d_001c;
FILE	*d_0020; 		/* infile */
int	d_0024;
int	d_0026[100];	/* user_flags */
char	d_00ee[14];
char	d_00fc[256];	/* line */
char	*d_01fc; 		/* dbgname */
FILE	*d_0200; 		/* dgbfile */
time_t	d_0204; 		/* clock */
int	d_0208; 			/* symbolcount */
int	d_020a; 			/* refcount */
FILE	*d_020c; 		/* listfile */
FILE	*d_0210;			/* objfile */
/* d_0214, 0x04 */
int	d_0218; 			/* column */
char	*d_021a;			/* progname */
char	d_021e[14];
int	d_022c; 			/* oldcolumn */
char	d_022e[256]; 	/* errmsgbuf */
long	d_032e;
char	d_0332[14];
char	*d_0340;			/* nameptr */
char	*d_0344; 		/* ptr */
char	*d_0348;
char	d_034c[26]; 	/* date */
FILE	*d_0366;			/* modulefile */
char	*d_036a; 		/* oldptr */
FILE	*d_036e; 		/* msgfile */
struct module	d_0372;
/* d_03c6, 0x02 */
struct module	*d_03c8;
char	d_03cc[504];	/* title */
char	*d_05c4; 		/* inname */
char	*d_05c8;			/* objname */
char	*d_05cc; 		/* msgname */
char	*d_05d0; 		/* listname */
char	*d_05d4; 		/* promptname */
char	*d_05d8;			/* modulename */
int	d_05dc = TRUE;		/* reading */
int	d_05de = FALSE; 	/* aborted */
int	d_05e0;				/* aborterr */
int	d_05e2 = FALSE; 	/* prompt_mode */
char	*d_05e4; 			/* prompt_string */
int	d_05e8 = FALSE; 	/* msg_append */
int	d_05ea = FALSE;
long	d_05ec = 0L;		/* codeaddr */
long	d_05f0 = 0L;		/* codestart */
int	d_05f4 = FALSE;
long	d_05f6 = 0L;
int	d_05fa = FALSE;
int	d_05fc = LIST_DEFAULT;	/* list_mode */
long	d_05fe = 0L;		/* romid */
int	d_0602 = FALSE;
int	d_0604 = FALSE; 	/* no_resolve */
long	d_0606 = 0L; 		/* errorcount */
struct	module	*d_060a = NULL;	/* module_last */
struct	module	*d_060e = NULL;	/* module_root */
struct	crc	*d_0612 = NULL;		/* crc_root */
struct	checksum	*d_0616 = NULL;	/* cksum_root */
long	d_061a = 0L;
long	d_061e = 0L;
int	d_0622 = FALSE;
long	d_0624 = 60; 		/* pagelength */
long	d_0628 = 80; 		/* pagewidth */
int	d_062c = TRUE; 	/* listing */
int	d_062e = FALSE;	/* codeonly */
int	d_0630 = FALSE;	/* hexcode */
int	d_0632 = 0;			/* option_flags */
int	d_0634 = TRUE;		/* absolute */
int	d_0636 = FALSE;	/* update */
char	d_0638[] = "SLOAD  Ver. %I%, %G%    ";	/* version */
char	d_0652[] = "@(#) Saturn Loader, Ver. %I%, %G%"; /* listversion */
char	*d_0674 = d_0652 + 5;
char	d_0678[] = "0123456789ABCDEF"; /* hextable */
int	d_0688;
int	d_068a = FALSE;	/* debug */

#endif
