/*
 * write.c - SATURN loader, output routines
 * (c) 1995 Mario Mikocevic Mozgy (Mario.Mikocevic@fer.hr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef	HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef	HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif

#include "struct.h"
#include "write.h"

extern	FILE	*objfile;
extern	char	*objname;
extern	char	date[], title[], version[];
extern	int	option_flags, codeonly, absolute;
extern	int	symbolcount, refcount;
extern	long	codestart, codeaddr, romid;

int	write_count = -1;
char	write_buffer[256] = "Symb";


void	write_header( void )
{
	char	buffer[256];
	int	symcnt, objcount, blocks, refcnt;
	long	code_len;

	blocks = 1;
	symcnt = symbolcount;
	refcnt = refcount;
	objcount = 0;
	if( !codeonly ) {
		if( ( option_flags & OPT_CODE ) != 0 ) {
			code_len = codeaddr - codestart;
			blocks += ( ( code_len + 511L ) >> 9 ); /* '>> 9' == '/ 512' */
		} else
			code_len = 0L;
		if( ( option_flags & OPT_SYMBOLS ) != 0 )
			objcount += symbolcount;
		else
			symcnt = 0;
		if( ( option_flags & OPT_REFS ) != 0 )
			objcount += refcount;
		else
			refcnt = 0;
		blocks += ( objcount + 13 ) / 14;
		memcpy( headerout.saturn.id, "Saturn", 6 );
		set_16( (unsigned char *)headerout.saturn.length, blocks );
		set_32( (unsigned char *)headerout.saturn.nibbles, code_len );
		set_16( (unsigned char *)headerout.saturn.symbols, symcnt );
		set_16( (unsigned char *)headerout.saturn.references, refcnt );
		set_32( (unsigned char *)headerout.saturn.start, (unsigned long)codestart );
		set_8( (unsigned char *)headerout.saturn.absolute, !absolute ); /* cmp absolute,1; sbb ax,ax; neg ax; sub ah,ah; */
		set_8( (unsigned char *)headerout.saturn.filler, '\0' );
		sprintf( buffer, "%-26s", date );
		memcpy( headerout.saturn.date, buffer, 26 );
		sprintf( buffer, "%-40s", title );
		memcpy( headerout.saturn.title, buffer, 40 );
		memset( headerout.saturn.softkeys, ' ', 20 );
		sprintf( buffer, "%-26s", version );
		memcpy( headerout.saturn.version, buffer, 26 );
		set_32( (unsigned char *)headerout.saturn.romid, romid );
		memset( headerout.saturn.reserved, 0, 118 ); /* 118 = 256 - 138 */
		rewind( objfile );
		if( fwrite( headerout.buffer, 256, 1, objfile ) != 1 )
			error( "WR_HD: error writing to %s", objname );
	}
}

union SFC	*next_object( void )
{
	if( ++write_count >= 14 ) {
		if( fwrite( write_buffer, 256, 1, objfile ) != 1 )
			error( "NX_OB: error writing to %s", objname );
		write_count = 0;
	}
	return (union SFC *)( write_buffer + write_count * 18 + 4 );
}

void	flush_objects( void )
{
	if( write_count >= 0 ) {
		if( fwrite( write_buffer, 256, 1, objfile ) != 1 )
			error( "FL_OB: error writing to %s", objname );
	}
}


void	write_one_symbol( struct symbol *symbp )
{
	unsigned int	fcount;
	union SFC	*ofp, *osp;
	int	i, len;
	long	value;
	struct fillref	*frp;

	osp = next_object();
	frp = symbp->s_fref_first;
	len = strlen( symbp->s_name );
	i = 0;
	while( i < len ) {
		osp->s.os_name[i] = symbp->s_name[i];
		++i;
	}
	while( i < 12 )
		osp->s.os_name[i++] = ' ';
	fcount = symbp->s_flags;
	value = symbp->s_value;
	if( ( fcount & SYM_RELOCATABLE ) != 0 )
		value -= codestart;
	if( ( option_flags & OPT_REFS ) == 0 ) {
		fcount &= SYM_TYPEMASK;
		frp = NULL;
	}
	set_32( (unsigned char *)osp->s.os_value, value );
	set_16( (unsigned char *)osp->s.os_fillcount, fcount );
	fcount &= SYM_REFSMASK;
	while( frp != NULL ) {
		ofp = next_object();
		set_32( (unsigned char *)ofp->f.of_address, frp->f_address - codestart );
		set_32( (unsigned char *)ofp->f.of_bias, frp->f_bias );
		set_8( (unsigned char *)ofp->f.of_class, frp->f_class );
		set_8( (unsigned char *)ofp->f.of_subclass, frp->f_subclass );
		set_16( (unsigned char *)ofp->f.of_length, frp->f_length );
		frp = frp->f_next;
		--fcount;
	}
	if( fcount != 0 )
		error( "WR_SY: reference count != list length:  %s", symbp->s_name );
}

void	write_symbols( void )
{
	if( !codeonly ) {
		do_all_symbols( write_one_symbol );
		flush_objects();
	}
}

void	write_objfile( void )
{
	if( ( objfile = openfile( objname, "SLOAD_OUTPUT", "", "wb", NULL ) ) == NULL )
		error( "WR_OU: unable to open %s as object file", objname );
	DEBUG_PRN( "Object file name set to '%s'\n", objname );
	if( option_flags == OPT_NONE )
		option_flags = OPT_DEFAULT;
	write_header();
	if( option_flags & OPT_CODE )
		write_code( codestart, codeaddr - codestart );
	if( option_flags & OPT_SYMBOLS )
		write_symbols();
}
